/*
 * UAR - UNIX AppleTalk Router
 * 
 * Copyright (c) 1993, The University of Melbourne.
 *
 * All Rights Reserved.
 *
 * UAR may NOT be publicly redistributed (for example via
 * anonymous FTP), sold, or the source used for any other
 * purpose without the permission of the copyright owner.
 * 
 * This software is supplied "as is" without express or
 * implied warranty.
 * 
 * djh@munnari.OZ.AU
 *
 * $Author: djh $
 * $Revision: 1.1.1.2 $
 *
 */

#include "uar.h"

extern int debug;
extern int numintfc;
extern u_char baddr[];
extern u_char maddr[];
extern char *stats_msg[];
extern u_long uar_stats[];
extern u_long pkts_dropped;
extern struct iflist iflist[];

int aarp_do_dump = 0;

struct aarpq *aarpQueue;
struct aarp *aarpTab[AARPTABSIZ];

/*
 * Initialize AARP data structures
 *
 */

void
aarp_initialize()
{
	int i;

	aarpQueue = AARPQNULL;

	for (i = 0; i < AARPTABSIZ; i++)
	  aarpTab[i] = AARPNULL;

	return;
}

/*
 * process an incoming AARP packet
 *
 */

void
aarp(ifn, buf, len)
int ifn;
u_char *buf;
int len;
{
	int new;
	struct pd pd;
	u_short function;
	register u_char *pkt, *addr;
	struct aarp *aarp_new(), *aarp;
	struct aarp *aarp_delete(), *aarp_find();
	void aarp_ifnode_inuse(), aarp_mnode_inuse();
	void aarp_respond(), aarp_dequeue();
	void aarp_insert(), aarp_glean();
	void dumppacket();

	incr(uarAARPPacketsIn);

	if (debug & AARPDEBUG) {
	  fprintf(stderr, "AARP: %-5s READ  %3d bytes\n",
	    iflist[ifn].if_name, len);
	  if (debug & PKTDEBUG)
	    dumppacket(buf, len);
	}

	switch (iflist[ifn].phase) {
	  case PHASE1:
	    pkt = buf+14;
	    break;
	  case PHASE2:
	    pkt = buf+22;
	    break;
	  default:
	    drop(uarUnknownPhase);
	    break;
	}

	addr = buf+6;	/* ethernet source	*/

	if (bcmp((char *)iflist[ifn].eaddr, (char *)addr,  6) == 0
	 || bcmp((char *)iflist[ifn].eaddr, (char *)(pkt+8), 6) == 0)
	  drop(uarIgnoreSelfSend);

	if (pkt[1] != 0x01 || pkt[2] != 0x80 || pkt[3] != 0x9b)
	  drop(uarBadPacketHdr);
	
	pd.dstNet = (pkt[25] << 8) | pkt[26];
	pd.srcNet = (pkt[15] << 8) | pkt[16];

	if (iflist[ifn].phase == PHASE1 && pd.dstNet != 0)
	  drop(uarBadPacketHdr);

	if (pd.dstNet == 0x0000)
	  pd.dstNet = iflist[ifn].net;

	if (pd.srcNet == 0x0000)
	  pd.srcNet = iflist[ifn].net;

	pd.dstNode = pkt[27];
	pd.srcNode = pkt[17];
	pd.ddpType = 0;
	pd.addr = NULL;

	if (pd.dstNode == 0x00 || pd.dstNode == BCAST)
	  drop(uarBadPacketHdr);

	function = (pkt[6] << 8) | pkt[7];

	switch (function) {
	  case AARPProbe:
	    /* pick new node if we're also probing */
	    if (iflist[ifn].net == pd.dstNet) {
	      if (iflist[ifn].node == pd.dstNode
	       && iflist[ifn].state == ADDR_PENDING)
	        aarp_ifnode_inuse(ifn);
	      if (iflist[ifn].mnode[pd.dstNode].state == MNODE_PENDING)
	        aarp_mnode_inuse(ifn, pd.dstNode);
	    }
	    /* we are required to "age out" this entry */
	    if ((aarp = aarp_find(pd.dstNet, pd.dstNode)) != AARPNULL)
	      (void)aarp_delete(aarp);
	    /* fall through */
	  case AARPReq:
	    /* defend address if one of ours */
	    if (iflist[ifn].net == pd.dstNet) {
	      if (iflist[ifn].node == pd.dstNode
	       && iflist[ifn].state == ADDR_VALID)
	        aarp_respond(ifn, &pd, addr);
	      if (iflist[ifn].mnode[pd.dstNode].state == MNODE_VALID)
	        aarp_respond(ifn, &pd, addr);
	    }
	    if (function == AARPReq)
	      aarp_glean(ifn, &pd, addr);
	    break;
	  case AARPResp:
	    if (iflist[ifn].net == pd.srcNet) {
	      if (iflist[ifn].node == pd.srcNode
	       && iflist[ifn].state == ADDR_PENDING)
	        aarp_ifnode_inuse(ifn);
	      if (iflist[ifn].mnode[pd.srcNode].state == MNODE_PENDING)
	        aarp_mnode_inuse(ifn, pd.srcNode);
	    }
	    /* response to probe or request */
	    aarp_dequeue(pd.srcNet, pd.srcNode);

	    if (new = ((aarp = aarp_find(pd.srcNet, pd.srcNode)) == AARPNULL))
	      if ((aarp = aarp_new()) == AARPNULL)
	        drop(uarMemoryError);

	    aarp->net = pd.srcNet;
	    aarp->node = pd.srcNode;
	    aarp->eaddr[0] = pkt[8];
	    aarp->eaddr[1] = pkt[9];
	    aarp->eaddr[2] = pkt[10];
	    aarp->eaddr[3] = pkt[11];
	    aarp->eaddr[4] = pkt[12];
	    aarp->eaddr[5] = pkt[13];
	    aarp->next = AARPNULL;

	    if (new)
	      aarp_insert(aarp);

	    break;
	  default:
	    drop(uarBadPacketHdr);
	    break;
	}

	return;
}

/*
 * putative router interface node unavailable
 *
 */

void
aarp_ifnode_inuse(ifn)
int ifn;
{
	void log(), cleanup();

	if (--iflist[ifn].node == 0)
	  iflist[ifn].node = (iflist[ifn].phase == PHASE2) ? 0xfd : 0xfe;

	if (iflist[ifn].node == iflist[ifn].hint) {
	  if (++iflist[ifn].net > iflist[ifn].net_hi) {
	    log("PANIC: no free nodes for router, giving up.");
	    cleanup();
	  }
	}

	iflist[ifn].count = AARPCount;

	return;
}

/*
 * router multinode support, a pending node number is unavailable
 *
 */

void
aarp_mnode_inuse(ifn, node)
int ifn;
u_char node;
{
	u_char client;
	void log(), aarp_enqueue();

	client = iflist[ifn].mnode[node].client;

	iflist[ifn].mnode[node].state = MNODE_UNUSED;
	iflist[ifn].mnode[node].client = MNODE_UNUSED;

	/*
	 * pick another node number
	 * (on the same net)
	 *
	 */
	for ( ; ; ) {
	  if (--node == 0) {
	    log("OOPS: no more free multinodes, continuing.");
	    return;
	  }
	  if (iflist[ifn].mnode[node].state == MNODE_UNUSED
	   && iflist[ifn].node != node)
	    break;
	}

	iflist[ifn].mnode[node].state = MNODE_PENDING;
	iflist[ifn].mnode[node].client = client;
	aarp_enqueue(ifn, iflist[ifn].net, node, AARPProbe);

	return;
}

/*
 * update AARP table with info from incoming packet
 *
 */

void
aarp_glean(ifn, pd, addr)
int ifn;
struct pd *pd;
u_char *addr;
{
	struct aarp *aarp;
	struct aarp *aarp_find();
	struct aarp *aarp_new();
	void aarp_insert();

	/*
	 * don't glean from NBP (too bogus)
	 *
	 */
	if (pd->ddpType == NBP)
	  return;

	/*
	 * glean only from locally sourced packets
	 *
	 */
	switch (iflist[ifn].phase) {
	  case PHASE1:
	    if (pd->srcNet == 0x0000)
	      break;
	    if (pd->srcNet != iflist[ifn].net)
	      return; /* drop, not local */
	    break;
	  case PHASE2:
	    if (pd->srcNet >= P2STARTUP)
	      break;
	    if (!(pd->srcNet >= iflist[ifn].net_lo
	       && pd->srcNet <= iflist[ifn].net_hi))
	      return; /* drop, not local */
	    break;
	  default:
	    drop(uarUnknownPhase);
	    break;
	}

	/*
	 * update existing or add new AARP entry
	 *
	 */
	if ((aarp = aarp_find(pd->srcNet, pd->srcNode)) == AARPNULL) {
	  if ((aarp = aarp_new()) == AARPNULL) {
	    incr(uarMemoryError);
	    return;
	  }
	  aarp->net = pd->srcNet;
	  aarp->node = pd->srcNode;
	  aarp->next = AARPNULL;
	  aarp_insert(aarp);
	}

	aarp->eaddr[0] = addr[0];
	aarp->eaddr[1] = addr[1];
	aarp->eaddr[2] = addr[2];
	aarp->eaddr[3] = addr[3];
	aarp->eaddr[4] = addr[4];
	aarp->eaddr[5] = addr[5];

	if (debug & AARPDEBUG) {
	  fprintf(stderr, "aarp: %-5s GLEAN %3d.%-3d   %3d   ",
	    iflist[ifn].if_name, pd->srcNet >> 8,
	    pd->srcNet & 0xff, pd->srcNode);
	  fprintf(stderr, "%02x:%02x:%02x:%02x:%02x:%02x\n",
	    addr[0], addr[1], addr[2], addr[3], addr[4], addr[5]);
	}

	return;
}

/*
 * send an AARP response to addr
 *
 */

void
aarp_respond(ifn, pd, addr)
int ifn;
struct pd *pd;
u_char *addr;
{
	u_char buf[128];
	register u_char *pkt;
	void aarp_write();

	switch (iflist[ifn].phase) {
	  case PHASE1:
	    pkt = buf+14;		/* ethernet header data	*/
	    pkt[14] = 0;		/* must be zero		*/
	    pkt[15] = 0;		/* must be zero		*/
	    pkt[16] = 0;		/* must be zero		*/
	    pkt[17] = pd->dstNode;	/* src protocol address	*/
	    pkt[24] = 0;		/* must be zero		*/
	    pkt[25] = 0;		/* must be zero		*/
	    pkt[26] = 0;		/* must be zero		*/
	    pkt[27] = pd->srcNode;	/* dst protocol address	*/
	    break;
	  case PHASE2:
	    pkt = buf+22;		/* ethernet header data	*/
	    pkt[14] = 0;		/* must be zero		*/
	    pkt[15] = pd->dstNet >> 8;	/* src network		*/
	    pkt[16] = pd->dstNet & 0xff;/* src network		*/
	    pkt[17] = pd->dstNode;	/* src protocol address	*/
	    pkt[24] = 0;		/* must be zero		*/
	    pkt[25] = pd->srcNet >> 8;	/* dst network		*/
	    pkt[26] = pd->srcNet & 0xff;/* dst network		*/
	    pkt[27] = pd->srcNode;	/* dst protocol address	*/
	    break;
	  default:
	    return;
	    break;
	}

	bcopy((char *)addr, (char *)buf,      6);
	bcopy((char *)addr, (char *)(pkt+18), 6);

	if (debug & AARPDEBUG) {
	  fprintf(stderr,
	    "AARP: %-5s RESPOND to %02x:%02x:%02x:%02x:%02x:%02x\n",
	    iflist[ifn].if_name, addr[0], addr[1],
	    addr[2], addr[3], addr[4], addr[5]);
	}

	aarp_write(ifn, buf, AARPResp);

	return;
}

/*
 * send an AARP Request/Probe
 *
 */

void
aarp_lookup(ifn, dstNet, dstNode, function)
int ifn;
u_char dstNode;
u_short dstNet;
u_short function;
{
	register u_char *pkt;
	u_char srcNode, buf[128], *addr;
	u_short srcNet;
	void aarp_write();

	srcNet  = (function == AARPProbe) ? dstNet  : iflist[ifn].net;
	srcNode = (function == AARPProbe) ? dstNode : iflist[ifn].node;

	switch (iflist[ifn].phase) {
	  case PHASE1:
	    pkt = buf+14;		/* ethernet header data	*/
	    pkt[14] = 0;		/* must be zero		*/
	    pkt[15] = 0;		/* must be zero		*/
	    pkt[16] = 0;		/* must be zero		*/
	    pkt[17] = srcNode;		/* src protocol address	*/
	    pkt[24] = 0;		/* must be zero		*/
	    pkt[25] = 0;		/* must be zero		*/
	    pkt[26] = 0;		/* must be zero		*/
	    pkt[27] = dstNode;		/* dst protocol address	*/
	    addr = baddr;		/* broadcast enet addr	*/
	    break;
	  case PHASE2:
	    pkt = buf+22;		/* ethernet header data	*/
	    pkt[14] = 0;		/* must be zero		*/
	    pkt[15] = srcNet >> 8;	/* src network 		*/
	    pkt[16] = srcNet & 0xff;	/* src network		*/
	    pkt[17] = srcNode;		/* src protocol address	*/
	    pkt[24] = 0;		/* must be zero		*/
	    pkt[25] = dstNet >> 8;	/* dst network		*/
	    pkt[26] = dstNet & 0xff;	/* dst network		*/
	    pkt[27] = dstNode;		/* dst protocol address	*/
	    addr = maddr;		/* multicast enet addr	*/
	    break;
	  default:
	    return;
	    break;
	}

	bcopy((char *)addr, (char *)buf, 6);	/* ethernet destination	*/
	bzero((char *)(pkt+18), 6);		/* must be zero		*/

	if (debug & AARPDEBUG)
	  fprintf(stderr, "AARP: %-5s %s for %3d.%-3d %3d\n",
	    iflist[ifn].if_name, (function == AARPProbe) ? "PROBE" : "REQUEST",
	    pkt[25], pkt[26], pkt[27]);

	aarp_write(ifn, buf, function);

	return;
}

/*
 * Write an AARP packet
 * Assumes both the EtherNet and AARP destination Hardware Address
 * and the Desired/Destination/Tentative protocol address are set.
 *
 */

void
aarp_write(ifn, buf, function)
int ifn;
u_char *buf;
u_short function;
{
	int fd;
	short len;
	register u_char *pkt;
	void dumppacket();

	switch (iflist[ifn].phase) {
	  case PHASE1:
	    pkt = buf+14;		/* ethernet header data	*/
	    buf[12] = 0x80;		/* ethernet protocol	*/
	    buf[13] = 0xf3;		/* ethernet protocol	*/
	    len = 42;
	    break;
	  case PHASE2:
	    pkt = buf+22;		/* ethernet header data	*/
	    buf[12] = 0x00;		/* 802.3 header length	*/
	    buf[13] = 0x24;		/* 802.3 header length	*/
	    buf[14] = 0xaa;		/* 802.2 header		*/
	    buf[15] = 0xaa;		/* 802.2 header		*/
	    buf[16] = 0x03;		/* 802.2 header		*/
	    buf[17] = 0x00;		/* SNAP prot discrim	*/
	    buf[18] = 0x00;		/* SNAP prot discrim	*/
	    buf[19] = 0x00;		/* SNAP prot discrim	*/
	    buf[20] = 0x80;		/* SNAP prot discrim	*/
	    buf[21] = 0xf3;		/* SNAP prot discrim	*/
	    len = 50;
	    break;
	  default:
	    drop(uarUnknownPhase);
	    break;
	}

	/* AARP packet	*/

	pkt[0] = 0x00;			/* hardware type	*/
	pkt[1] = 0x01;			/* hardware type	*/
	pkt[2] = 0x80;			/* protocol type	*/
	pkt[3] = 0x9b;			/* protocol type	*/
	pkt[4] = 0x06;			/* h/w address length	*/
	pkt[5] = 0x04;			/* prot address length	*/
	pkt[6] = function >> 8;		/* AARP function	*/
	pkt[7] = function & 0xff;	/* AARP function	*/

	bcopy((char *)iflist[ifn].eaddr, (char *)(buf+6), 6);
	bcopy((char *)iflist[ifn].eaddr, (char *)(pkt+8), 6);

	if (debug & AARPDEBUG) {
	  fprintf(stderr, "AARP: %-5s WRITE %3d bytes\n",
	    iflist[ifn].if_name, len);
	  if (debug & PKTDEBUG)
	    dumppacket(buf, len);
	}

	incr(uarAARPPacketsOut);

	fd = (iflist[ifn].aarpfd >= 0) ? iflist[ifn].aarpfd : iflist[ifn].etfd;

	if (write_eth(fd, buf, len) != len)
	  if (debug) perror("write_eth()");

	return;
}

/*
 * the AARP process
 *
 */

void
aarp_process()
{
	u_char node;
	int ifn, client;
	struct aarpq *p, *q, *t;
	void aarp_lookup(), aarp_enqueue();
	void zone_getnetinfo(), cap_start();
	void uar_register(), log();

	p = aarpQueue;
	q = AARPQNULL;

	/*
	 * check AARP queue for pending probes or requests
	 *
	 */
	while (p != AARPQNULL) {
	  if (iflist[p->intfc].state == ADDR_VALID) {
	    if (p->count-- > 0) {
	      aarp_lookup(p->intfc, p->net, p->node, p->function);
	    } else {
	    /*
	     * probe unchallenged or no response to request
	     *
	     */
	      if (p->function == AARPProbe
	       && iflist[p->intfc].net == p->net
	       && iflist[p->intfc].mnode[p->node].state == MNODE_PENDING) {
	      /*
	       * multinode probe, start service
	       *
	       */
	        iflist[p->intfc].mnode[p->node].state = MNODE_VALID;
	        switch (iflist[p->intfc].mnode[p->node].client) {
	          case MNODE_CAP:
	            cap_start(p->intfc, p->node);
	            break;
	          default:
	            break;
	        }
	      }
	      /*
	       * dequeue the request
	       *
	       */
	      if (q == AARPQNULL)
	        aarpQueue = p->next;
	      else
	        q->next = p->next;
	      t = p->next;
	      free((char *)p);
	      p = t;
	      continue;
	    }
	  }
	  q = p;
	  p = p->next;
	}

	/*
	 * probe for interface addresses
	 * start probes for multinodes when interface address valid
	 *
	 */
	for (ifn = 0; ifn < numintfc; ifn++) {
	  if (iflist[ifn].state == ADDR_PENDING) {
	    if (iflist[ifn].count-- > 0) {
	      aarp_lookup(ifn, iflist[ifn].net, iflist[ifn].node, AARPProbe);
	    } else {
	    /*
	     * adopt this address, register NBP name
	     *
	     */
	      iflist[ifn].state = ADDR_VALID;
	      if (iflist[ifn].phase == PHASE2)
	        zone_getnetinfo(ifn);
	      uar_register(ifn);
	      /*
	       * (re)start multinode clients
	       *
	       */
	      bzero((char *)iflist[ifn].mnode, sizeof(iflist[ifn].mnode));
	      iflist[ifn].mnode[iflist[ifn].node].state = MNODE_ROUTER;
	      if (iflist[ifn].clients) {
	        node = iflist[ifn].node;
	        client = MNODE_MIN;
	        for ( ; ; ) {
	          if (--node == 0) {
	            log("OOPS: no more free multinodes, continuing.");
	            break;
	          }
	          if (iflist[ifn].clients & client) {
	            if (iflist[ifn].mnode[node].state == MNODE_UNUSED) {
	              iflist[ifn].mnode[node].state = MNODE_PENDING;
	              iflist[ifn].mnode[node].client = (u_char)client;
	              aarp_enqueue(ifn, iflist[ifn].net, node, AARPProbe);
	            } else continue;
	          } else node++;
	          if ((client <<= 1) > MNODE_MAX)
	            break;
	        }
	      }
	    }
	  }
	}

	return;
}

/*
 * queue an AARP probe or lookup, ignore if already queued
 *
 */

void
aarp_enqueue(ifn, net, node, function)
int ifn;
u_short net;
u_char node;
u_short function;
{
	struct aarpq *aarpq, *p, *q;

	p = aarpQueue;
	q = AARPQNULL;

	while (p != AARPQNULL) {
	  if (p->net == net
	   && p->node == node)
	    return;
	  if (p->node > node)
	    break;
	  q = p;
	  p = p->next;
	}

	aarpq = (struct aarpq *)malloc(sizeof(struct aarpq));

	if (aarpq == AARPQNULL) {
	  incr(uarMemoryError);
	  return;
	}

	aarpq->net = net;
	aarpq->node = node;
	aarpq->function = function;
	aarpq->count = AARPCount;
	aarpq->intfc = ifn;
	aarpq->next = AARPQNULL;

	if (q == AARPQNULL)
	  aarpQueue = aarpq;
	else
	  q->next = aarpq;

	aarpq->next = p;

	return;
}

/*
 * dequeue an AARP probe or lookup
 *
 */

void
aarp_dequeue(net, node)
u_short net;
u_char node;
{
	struct aarpq *p, *q;

	p = aarpQueue;
	q = AARPQNULL;

	while (p != AARPQNULL) {
	  if (p->net == net
	   && p->node == node)
	    break;
	  q = p;
	  p = p->next;
	}

	if (p == AARPQNULL)
	  return;

	if (q == AARPQNULL)
	  aarpQueue = p->next;
	else
	  q->next = p->next;

	free((char *)p);

	return;
}

/*
 * find an entry in the AARP table.
 *
 */

struct aarp *
aarp_find(net, node)
u_short net;
u_char node;
{
	struct aarp *aarp;

	incr(aarpLookups);

	aarp = aarpTab[node];

	while (aarp != AARPNULL) {
	  if (aarp->net == net
	   && aarp->node == node) {
	    incr(aarpHits);
	    return(aarp);
	  }
	  aarp = aarp->next;
	}

	incr(uarAARPCacheFail);

	return(AARPNULL);
}

/*
 * insert AARP tuple in linked list
 *
 */

void
aarp_insert(aarp)
struct aarp *aarp;
{
	int idx;
	struct aarp *p, *q;

	idx = aarp->node;

	p = aarpTab[idx];
	q = AARPNULL;

	while (p != AARPNULL) {
	  if (p->net >= aarp->net)
	    break;
	  q = p;
	  p = p->next;
	}

	if (q == AARPNULL)
	  aarpTab[idx] = aarp;
	else
	  q->next = aarp;

	aarp->next = p;

	return;
}

/*
 * delete AARP tuple from linked list
 *
 */

struct aarp *
aarp_delete(aarp)
struct aarp *aarp;
{
	int idx;
	struct aarp *next, *p, *q;
	void aarp_free();

	idx = aarp->node;

	p = aarpTab[idx];
	q = AARPNULL;

	while (p != AARPNULL) {
	  if (p == aarp)
	    break;
	  q = p;
	  p = p->next;
	}

	if (p == AARPNULL)
	  return(p);

	next = p->next;

	if (q == AARPNULL)
	  aarpTab[idx] = next;
	else
	  q->next = next;

	aarp_free(aarp);

	return(next);
}

/*
 * dump the AARP table
 *
 */

void
aarp_dump()
{
	int i, count;
	FILE *r, *fopen();
	struct aarp *aarp;

	if ((r = fopen(DUMP_AARP, "w+")) == NULL) {
	  incr(uarBadDumpFile);
	  return;
	}

	fprintf(r, "UAR AARP Table:\n\n");
	fprintf(r, "  net   node  eaddr\n");

	for (i = 0, count = 0; i < AARPTABSIZ; i++) {
	  aarp = aarpTab[i];
	  while (aarp != AARPNULL) {
	    fprintf(r, "%3d.%-3d  %3d  %02x:%02x:%02x:%02x:%02x:%02x\n",
	      aarp->net >> 8, aarp->net & 0xff, aarp->node,
	      aarp->eaddr[0], aarp->eaddr[1], aarp->eaddr[2],
	      aarp->eaddr[3], aarp->eaddr[4], aarp->eaddr[5]);
	    aarp = aarp->next;
	    count++;
	  }
	}
	fprintf(r, "\n\n%d AARP entries\n", count);

	(void)fclose(r);

	if (debug & AARPDEBUG)
	  fprintf(stderr, "AARP: dumped AARP table, %d entries\n", count);

	return;
}

/*
 * create a new AARP element
 *
 */

struct aarp *
aarp_new()
{
	struct aarp *aarp;

	aarp = (struct aarp *)malloc(sizeof(struct aarp));

	if (aarp == AARPNULL)
	  incr(uarMemoryError);

	return(aarp);
}

/*
 * free memory occupied by AARP element
 *
 */

void
aarp_free(aarp)
struct aarp *aarp;
{
	if (debug & AARPDEBUG)
	  fprintf(stderr, "AARP: FREE net %3d.%-3d node %3d\n",
	    aarp->net >> 8, aarp->net & 0xff, aarp->node);

	free((char *)aarp);

	return;
}
