/*
 * UAR - UNIX AppleTalk Router
 * 
 * Copyright (c) 1993, The University of Melbourne.
 *
 * All Rights Reserved.
 *
 * UAR may NOT be publicly redistributed (for example via
 * anonymous FTP), sold, or the source used for any other
 * purpose without the permission of the copyright owner.
 * 
 * This software is supplied "as is" without express or
 * implied warranty.
 * 
 * djh@munnari.OZ.AU
 *
 * $Author: djh $
 * $Revision: 1.3 $
 *
 */

#include "uar.h"

extern int debug;
extern int capfd;
extern int checksum;
extern u_long uar_stats[];
extern u_long pkts_dropped;
extern struct iflist iflist[];
extern char *stats_msg[];
extern char *thisZone;
extern u_char buf[];

static int portrange;
static u_char capNode;
static u_short capNet;
static struct sockaddr_in lsin;

/*
 * open CAP UDP loopback port, set socket options and check NIC port range
 *
 */

void
cap_initialize(ifn)
int ifn;
{
	int len;
	unsigned long inet_addr();
	struct zone *zone, *zone_make();
	struct servent *s, *getservbyname();
	void exit();

	if (capfd != -1) {
	  fprintf(stderr,
	    " CAP: %-5s services available on one interface only\n",
	    iflist[ifn].if_name);
	  exit(1);
	}

	if ((capfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
	  perror("socket()");
	  exit(1);
	}

	lsin.sin_family = AF_INET;
	lsin.sin_port = htons(CAP_PORT);
	lsin.sin_addr.s_addr = (unsigned long)inet_addr("127.0.0.1");
	if (bind(capfd, (struct sockaddr *)&lsin, sizeof(lsin)) < 0) {
	  perror("bind()");
	  exit(1);
	}

	len = SKTBUFSIZ;
	if (setsockopt(capfd,SOL_SOCKET,SO_RCVBUF,(char *)&len,sizeof(int))<0){
	  perror("SO_RCVBUF");
	  exit(1);
	}

	len = SKTBUFSIZ;
	if (setsockopt(capfd,SOL_SOCKET,SO_SNDBUF,(char *)&len,sizeof(int))<0){
	  perror("SO_SNDBUF");
	  exit(1);
	}

	s = getservbyname("at-rtmp", "udp");
	portrange = (s == NULL) ? 768 : ntohs(s->s_port)-1;

	capNet = 0;
	capNode = 0;

	if (thisZone != NULL)
	  if ((zone = zone_make((u_char *)thisZone)) != ZONENULL)
	    iflist[ifn].ourzone = zone;

	iflist[ifn].clients |= MNODE_CAP;

	if (debug & CAPDEBUG)
	  fprintf(stderr, " CAP: %-5s initialised CAP services\n",
	    iflist[ifn].if_name);

	return;
}

/*
 * start CAP services
 *
 */

void
cap_start(ifn, node)
int ifn;
u_char node;
{
	char *ctime();
	char *zone_name();
	FILE *r, *fopen();
	time_t now, time();

	capNode = node;
	capNet = iflist[ifn].net;

	/*
	 * create a suitable /etc/etalk.local file
	 *
	 */
	if ((r = fopen(ETH_DATABASE, "w")) != NULL) {
	  (void)time(&now);
	  fprintf(r, "#\n# EtherTalk dynamic configuration data\n#\n");
	  fprintf(r, "# Last update:\t%s#\n", ctime(&now));
	  fprintf(r, "# Generated by UAR\n#\n");
	  fprintf(r, "%-12s\t\"%s\"\n", ETH_INTERFACE, iflist[ifn].if_name);
	  if (iflist[ifn].phase == PHASE2) {
	    fprintf(r, "%-12s\t%3d.%-3d\n", ETH_NET_START,
	      iflist[ifn].net_lo >> 8, iflist[ifn].net_lo & 0xff);
	    fprintf(r, "%-12s\t%3d.%-3d\n", ETH_NET_END,
	      iflist[ifn].net_hi >> 8, iflist[ifn].net_hi & 0xff);
	  }
	  fprintf(r, "%-12s\t%3d.%-3d\n", ETH_THIS_NET,
	    capNet >> 8, capNet & 0xff);
	  fprintf(r, "%-12s\t%3d\n", ETH_THIS_NODE, node);
	  fprintf(r, "%-12s\t\"%s\"\n", ETH_THIS_ZONE, zone_name(ifn));
	  fprintf(r, "%-12s\t%3d.%-3d\n", ETH_BRIDGE_NET,
	    iflist[ifn].net >> 8, iflist[ifn].net & 0xff);
	  fprintf(r, "%-12s\t%3d\n", ETH_BRIDGE_NODE, iflist[ifn].node);
	  fprintf(r, "%-12s\t127.0.0.1\n", ETH_BRIDGE_IP);
	  fprintf(r, "%-12s\t%3d.%-3d\n", ETH_NIS_NET,
	    capNet >> 8, capNet & 0xff);
	  fprintf(r, "%-12s\t%3d\n", ETH_NIS_NODE, node);
	  (void)fflush(r);
	  (void)fclose(r);
	}

	if (debug & CAPDEBUG)
	  fprintf(stderr, " CAP: %-5s starting %3d.%-3d %3d\n",
	    iflist[ifn].if_name, capNet >> 8, capNet & 0xff, capNode);

	return;
}

/*
 * read from the CAP loopback port
 *
 */

void
readcap(fd)
int fd;
{
	int cc;
	void cap();

	if ((cc = read(fd, (char *)(buf+HDRSIZ), BUFSIZE-HDRSIZ)) < 0) {
	  if (debug) perror("readcap()");
	  return;
	}

	if (cc == 0)
	  return;

	cap(buf+HDRSIZ, cc);

	return;
}

/*
 * packet to CAP services from network via router (interface 'ifn')
 *
 */

void
cap_route(ifn, pd)
int ifn;
struct pd *pd;
{
	short len;
	u_char *pkt, *hdr;
	u_short cap_udpport();
	u_short cksum, chkSum();
	void dumppacket();

	if (capfd < 0 || capNet == 0 || capNode == 0)
	  return;

	pd->lapType = LDDP;

	pkt = pd->data-13;
	len = pd->dataLen+13;

	/*
	 * construct the DDP header
	 *
	 */
	pkt[0]  = (pd->hop << 2) | ((len >> 8) & 0x03);
	pkt[1]  = len & 0xff;
	pkt[4]  = pd->dstNet >> 8;
	pkt[5]  = pd->dstNet & 0xff;
	pkt[6]  = pd->srcNet >> 8;
	pkt[7]  = pd->srcNet & 0xff;
	pkt[8]  = pd->dstNode;
	pkt[9]  = pd->srcNode;
	pkt[10] = pd->dstSkt;
	pkt[11] = pd->srcSkt;
	pkt[12] = pd->ddpType;
	cksum   = (checksum) ? chkSum(pkt+4, len-4) : 0x0000;
	pkt[2]  = cksum >> 8;
	pkt[3]  = cksum & 0xff;

	/*
	 * construct the IPTalk header
	 *
	 */
	len += 3;
	hdr = pkt-3;

	hdr[0] = 0xfa;	/* magic */
	hdr[1] = 0xce;	/* magic */
	hdr[2] = pd->lapType;

	/*
	 * route to local CAP process via UDP port
	 *
	 */
	lsin.sin_port = htons(cap_udpport(pd->dstSkt));

	if (debug & CAPDEBUG) {
	  fprintf(stderr, " CAP: %-5s WRITE %3d bytes\n", "lo0", len);
	  if (debug & PKTDEBUG)
	    dumppacket(hdr, len);
	}

	incr(uarCAPPacketsOut);

	if (sendto(capfd, (char *)hdr, len, 0,
	 (struct sockaddr *)&lsin, sizeof(struct sockaddr)) < 0)
	  if (debug) perror("sendto()");

	return;
}

/*
 * map AppleTalk destination socket to UDP port
 *
 */

u_short
cap_udpport(skt)
u_char skt;
{
	return((skt >= 64) ? (skt+16384) : (skt+portrange));
}

/*
 * packet from CAP services via UDP loopback
 *
 */

void
cap(buf, len)
u_char *buf;
int len;
{
	u_char *pkt;
	struct pd pd;
	short pktlen;
	int ddp_for_gw();
	int ddp_for_mn();
	u_short cksum, chkSum();
	struct rtmp *rtmp, *rtmp_route();
	void ddp_gw(), ddp_mn(), ddp_route(), dumppacket();

	if (capfd < 0 || capNet == 0 || capNode == 0)
	  return;

	incr(uarCAPPacketsIn);

	if (debug & CAPDEBUG) {
	  fprintf(stderr, " CAP: %-5s READ  %3d bytes\n", "lo0", len);
	  if (debug & PKTDEBUG)
	    dumppacket(buf, len);
	}

	pd.lapType = buf[2];

	pkt = buf+3;

	pd.hop = ((pkt[0] & 0x3c) >> 2);
	pktlen = ((pkt[0] & 0x03) << 8) | pkt[1];

	if (pktlen > len)
	  drop(ddpTooShortErrors);

	switch (pd.lapType) {
	  case SDDP: /* short DDP */
	    pd.dstNet  = capNet;
	    pd.srcNet  = capNet;
	    pd.dstSkt  = pkt[2];
	    pd.srcSkt  = pkt[3];
	    pd.ddpType = pkt[4];
	    pd.dataLen = pktlen-5;
	    pd.data    = pkt+5;
	    break;
	  case LDDP: /* long DDP  */
	    cksum      = (pkt[2] << 8) | pkt[3];
	    pd.dstNet  = (pkt[4] << 8) | pkt[5];
	    pd.srcNet  = (pkt[6] << 8) | pkt[7];
	    pd.dstNode = pkt[8];
	    pd.srcNode = pkt[9];
	    pd.dstSkt  = pkt[10];
	    pd.srcSkt  = pkt[11];
	    pd.ddpType = pkt[12];
	    pd.dataLen = pktlen-13;
	    pd.data    = pkt+13;
	    if (checksum && cksum)
	      if (cksum != chkSum(pkt+4, pktlen-4))
	        drop(ddpChecksumErrors);
	    break;
	  default:
	    drop(uarBadLapType);
	    break;
	}

	if (pd.dataLen > MAXDDPLEN)
	  drop(ddpTooLongErrors);

	pd.addr = NULL;

	/*
	 * search for routing table entry
	 *
	 */
	if ((rtmp = rtmp_route(pd.dstNet)) == RTMPNULL)
	  drop(ddpOutNoRoutes);

	/*
	 * route if not local
	 *
	 */
	if (rtmp->dist != 0) {
	  if (pd.hop++ >= 15)
	    drop(ddpHopCountErrors);
	  incr(ddpForwRequests);
	  ddp_route(rtmp->intfc, &pd, rtmp);
	  return;
	}

	/*
	 * DDP dst node == 0 ?
	 *
	 */
	if (pd.dstNode == 0x00) {
	  ddp_gw(rtmp->intfc, &pd);
	  return;
	}

	/*
	 * packet for gateway on rtmp->intfc ?
	 *
	 */
	if (ddp_for_gw(rtmp->intfc, pd.dstNet, pd.dstNode)) {
	  ddp_gw(rtmp->intfc, &pd);
	  if (pd.dstNode != BCAST)
	    return;
	}

	/*
	 * packet for multinode client on rtmp->intfc ?
	 *
	 */
	if (ddp_for_mn(rtmp->intfc, pd.dstNet, pd.dstNode)) {
	  ddp_mn(rtmp->intfc, &pd);
	  if (pd.dstNode != BCAST)
	    return;
	}

	if (pd.hop++ >= 15)
	  drop(ddpHopCountErrors);

	incr(ddpForwRequests);

	ddp_route(rtmp->intfc, &pd, rtmp);

	return;
}
