/*
 * UAR - UNIX AppleTalk Router
 * 
 * Copyright (c) 1993, The University of Melbourne.
 *
 * All Rights Reserved.
 *
 * UAR may NOT be publicly redistributed (for example via
 * anonymous FTP), sold, or the source used for any other
 * purpose without the permission of the copyright owner.
 * 
 * This software is supplied "as is" without express or
 * implied warranty.
 * 
 * djh@munnari.OZ.AU
 *
 * $Author: djh $
 * $Revision: 1.1.1.4 $
 *
 */

#include "uar.h"

extern int debug;
extern int checksum;
extern u_char baddr[];
extern u_char maddr[];
extern u_long uar_stats[];
extern u_long pkts_dropped;
extern struct iflist iflist[];
extern struct RDS RDS[];
extern char *stats_msg[];
extern struct timeval tn;
extern u_char buf[];

/*
 * read an ELAP packet from the net
 *
 */

void
readeth(ifn, fd)
int ifn, fd;
{
	int cc, i;
	void ddp();

	if ((cc = read_eth(fd, buf+HDRSIZ, BUFSIZE-HDRSIZ)) < 0) {
	  if (debug) perror("readeth()");
	  return;
	}

	if (cc == 0)
	  return;

	if (iflist[ifn].state == ADDR_PENDING)
	  return; /* not init'd yet */

	for (i = 0; i < NUMRDS; i++) {
	  if (RDS[i].dataLen == 0)
	    break;
	  ddp(ifn, RDS[i].dataPtr, RDS[i].dataLen);
	}

	return;
}

/*
 * read an AARP packet
 *
 */

void
readaarp(ifn, fd)
int ifn, fd;
{
	int cc, i;
	void aarp();

	if ((cc = read_eth(fd, buf+HDRSIZ, BUFSIZE-HDRSIZ)) < 0) {
	  if (debug) perror("readaarp()");
	  return;
	}

	if (cc == 0)
	  return;

	for (i = 0; i < NUMRDS; i++) {
	  if (RDS[i].dataLen == 0)
	    break;
	  aarp(ifn, RDS[i].dataPtr, RDS[i].dataLen);
	}

	return;
}

/*
 * Sony NEWS allows /dev/ether? to be opened once only.
 * readdev() switches incoming packets to the appropriate input routine.
 *
 * Yasuaki Honda <honda@csl.sony.co.jp>
 *
 */

void
readdev(ifn, fd)
int ifn, fd;
{
	int cc, i;
	u_char *pkt;
	void ddp(), aarp();

	if ((cc = read_eth(fd, buf+HDRSIZ, BUFSIZE-HDRSIZ)) < 0) {
	  if (debug) perror("readdev()");
	  return;
	}

	if (cc == 0)
	  return;

	for (i = 0; i < NUMRDS; i++) {
	  if (RDS[i].dataLen == 0)
	    break;

	  pkt = RDS[i].dataPtr;
	  cc = RDS[i].dataLen;

	  if (iflist[ifn].phase == PHASE1) {
	    if (pkt[12] == 0x80) {
	      switch (pkt[13]) {
	        case 0xf3:
	          aarp(ifn, pkt, cc);
	          break;
	        case 0x9b:
	          if (iflist[ifn].state == ADDR_PENDING)
	            return; /* not init'd yet */
	          ddp(ifn, pkt, cc);
	          break;
	        default:
	          break;
	      }
	    }
	  } else if (iflist[ifn].phase == PHASE2) {
	    if ((pkt[12]*256+pkt[13] < 1500)
	     && (pkt[14]==0xAA) && (pkt[15]==0xAA) && (pkt[16]==0x03)) {
	      /* IEEE802.3 and SNAP and control = 0x03 */
	      if ((pkt[17]==0x08) && (pkt[18]==0x00) && (pkt[19]==0x07)
	       && (pkt[20]==0x80) && (pkt[21]==0x9B)) {
	        if (iflist[ifn].state == ADDR_PENDING)
	          return; /* not init'd yet */
	        ddp(ifn, pkt, cc);
	      } else {
	        if ((pkt[17]==0x00) && (pkt[18]==0x00) && (pkt[19]==0x00)
	         && (pkt[20]==0x80) && (pkt[21]==0xF3)) {
	          aarp(ifn, pkt, cc);
	        }
	      }
	    }
	  }
	}

	return;
}

/*
 * process a DDP packet received on interface 'ifn'
 *
 */

void
ddp(ifn, buf, len)
int ifn;
u_char *buf;
int len;
{
	register u_char *pkt, *addr, *ethType;
	struct rtmp *rtmp, *rtmp_route();
	u_short cksum, chkSum();
	void dumppacket();
	void aarp_glean();
	int ddp_for_gw();
	int ddp_for_mn();
	int ddp_is_gw();
	void ddp_route();
	void ddp_gw();
	void ddp_mn();
	struct pd pd;
	short pktlen;

	incr(ddpInReceives);

	if (debug & ETHDEBUG) {
	  fprintf(stderr, "ETH%d: %-5s READ  %3d bytes\n",
	    iflist[ifn].phase, iflist[ifn].if_name, len);
	  if (debug & PKTDEBUG)
	    dumppacket(buf, len);
	}

	addr = buf+6; /* ethernet source */

	/*
	 * check if we can see our own transmissions
	 *
	 */
	if (iflist[ifn].itype == ITYPE_ETHERTALK)
	  if (bcmp((char *)iflist[ifn].eaddr, (char *)addr, 6) == 0)
	    drop(uarIgnoreSelfSend);

	switch (iflist[ifn].phase) {
	  case PHASE1:
	    pd.dstNode = buf[14];
	    pd.srcNode = buf[15];
	    pd.lapType = buf[16];
	    ethType = buf+12;
	    pkt = buf+17;
	    len = len-17;
	    break;
	  case PHASE2:
	    pd.lapType = LDDP;
	    ethType = buf+20;
	    pkt = buf+22;
	    len = len-22;
	    break;
	  default:
	    drop(uarUnknownPhase);
	    break;
	}

	if (ethType[0] != 0x80 || ethType[1] != 0x9b)
	  drop(uarBadEtherType);

	pd.hop = ((pkt[0] & 0x3c) >> 2);
	pktlen = ((pkt[0] & 0x03) << 8) | pkt[1];

	if (pktlen > len)
	  drop(ddpTooShortErrors);

	switch (pd.lapType) {
	  case SDDP: /* short DDP */
	    pd.dstNet  = iflist[ifn].net;
	    pd.srcNet  = iflist[ifn].net;
	    pd.dstSkt  = pkt[2];
	    pd.srcSkt  = pkt[3];
	    pd.ddpType = pkt[4];
	    pd.dataLen = pktlen-5;
	    pd.data    = pkt+5;
	    break;
	  case LDDP: /* long DDP  */
	    cksum      = (pkt[2] << 8) | pkt[3];
	    pd.dstNet  = (pkt[4] << 8) | pkt[5];
	    pd.srcNet  = (pkt[6] << 8) | pkt[7];
	    pd.dstNode = pkt[8];
	    pd.srcNode = pkt[9];
	    pd.dstSkt  = pkt[10];
	    pd.srcSkt  = pkt[11];
	    pd.ddpType = pkt[12];
	    pd.dataLen = pktlen-13;
	    pd.data    = pkt+13;
	    if (checksum && cksum)
	      if (cksum != chkSum(pkt+4, pktlen-4))
	        drop(ddpChecksumErrors);
	    break;
	  default:
	    drop(uarBadLapType);
	    break;
	}

	if (pd.dataLen > MAXDDPLEN)
	  drop(ddpTooLongErrors);

	pd.addr = NULL;

	/*
	 * check if we can see our own transmissions
	 *
	 */
	if (ddp_is_gw(ifn, pd.srcNet, pd.srcNode))
	  drop(uarIgnoreSelfSend);

	/*
	 * glean ethernet hardware addresses
	 *
	 */
	if (iflist[ifn].itype == ITYPE_ETHERTALK)
	  aarp_glean(ifn, &pd, addr);

	/*
	 * Datagram routing algorithm for a router.
	 * Inside AppleTalk, 2nd Edition, page 5-26 (P5-21)
	 *
	 */

	/*
	 * packet aimed at gateway ?
	 *
	 */
	if (ddp_for_gw(ifn, pd.dstNet, pd.dstNode)) {
	  ddp_gw(ifn, &pd);
	  if (pd.dstNode != BCAST)
	    return;
	}

	/*
	 * packet for EtherTalk multinode client on ifn ?
	 *
	 */
	if (ddp_for_mn(ifn, pd.dstNet, pd.dstNode)) {
	  ddp_mn(ifn, &pd);
	  if (pd.dstNode != BCAST)
	    return;
	}

	/*
	 * check if network specific broadcast
	 *
	 */
	if (pd.dstNet >= iflist[ifn].net_lo
	 && pd.dstNet <= iflist[ifn].net_hi
	 && pd.dstNet != iflist[ifn].net
	 && pd.dstNode == BCAST)
	  drop(ddpBroadcastErrors);

	/*
	 * squelch local broadcasts
	 *
	 */
	if (ddp_netmatch(iflist[ifn].net, pd.dstNet))
	  if (pd.dstNode == BCAST)
	    return;

	/*
	 * not for anyone on the incoming interface,
	 * search for routing table entry
	 *
	 */
	if ((rtmp = rtmp_route(pd.dstNet)) == RTMPNULL)
	  drop(ddpOutNoRoutes);

	/*
	 * route if destination net
	 * not local to router
	 *
	 */
	if (rtmp->dist != 0) {
	  if (pd.hop++ >= 15)
	    drop(ddpHopCountErrors);
	  incr(ddpForwRequests);
	  ddp_route(ifn, &pd, rtmp);
	  return;
	}

	/*
	 * DDP dst node == 0 ?
	 *
	 */
	if (pd.dstNode == 0x00) {
	  ddp_gw(rtmp->intfc, &pd);
	  return;
	}

	/*
	 * packet for gw on interface rtmp->intfc ?
	 *
	 */
	if (ddp_for_gw(rtmp->intfc, pd.dstNet, pd.dstNode)) {
	  ddp_gw(rtmp->intfc, &pd);
	  if (pd.dstNode != BCAST)
	    return;
	}

	/*
	 * check if packet for EtherTalk multinode client on rtmp->intfc
	 *
	 */
	if (ddp_for_mn(rtmp->intfc, pd.dstNet, pd.dstNode)) {
	  ddp_mn(rtmp->intfc, &pd);
	  if (pd.dstNode != BCAST)
	    return;
	}

	/*
	 * otherwise forward it
	 *
	 */
	if (pd.hop++ >= 15)
	  drop(ddpHopCountErrors);

	incr(ddpForwRequests);

	ddp_route(ifn, &pd, rtmp);

	return;
}

/*
 * return true if address belongs to the gateway on interface ifn
 *
 */

int
ddp_for_gw(ifn, net, node)
int ifn;
u_short net;
u_char node;
{
	int ddp_netmatch();

	if (ddp_netmatch(iflist[ifn].net, net))
	  if (iflist[ifn].node == node || node == BCAST)
	    return(1);

	return(0);
}

/*
 * return true if the node number is for an EtherTalk multinode client
 *
 */

int
ddp_for_mn(ifn, net, node)
int ifn;
u_short net;
u_char node;
{
	int ddp_netmatch();

	if (ddp_netmatch(iflist[ifn].net, net))
	  if (iflist[ifn].itype == ITYPE_ETHERTALK)
	    if (iflist[ifn].mnode[node].state == MNODE_VALID || node == BCAST)
	      return(1);

	return(0);
}

/*
 * return true if valid interface node
 *
 */

int
ddp_is_gw(ifn, net, node)
int ifn;
u_short net;
u_char node;
{
	if (iflist[ifn].net == net)
	  if (iflist[ifn].node == node
	   && iflist[ifn].state == ADDR_VALID)
	    return(1);

	return(0);
}

/*
 * return true if valid multinode node
 *
 */

int
ddp_is_mn(ifn, net, node)
int ifn;
u_short net;
u_char node;
{
	if (iflist[ifn].net == net)
	  if (iflist[ifn].mnode[node].state == MNODE_VALID)
	    return(1);

	return(0);
}

/*
 * check if nets match according to the DDP rules
 * Net A matches net B if A==B or A==0 or B==0
 *
 * "Network number equivalence" P4-21 (P4-15)
 *
 */

int
ddp_netmatch(A, B)
u_short A, B;
{
	return((A == B) || (A == 0) || (B == 0));
}

/*
 * packet for router on interface 'ifn'
 *
 */

void
ddp_gw(ifn, pd)
int ifn;
struct pd *pd;
{
	void echo(), nbp();
	void rtmpr(), rtmpd();
	void zip(), zip_atp();
	void timesrvr_atp();

	incr(ddpInLocalDatagrams);

	if (pd->dstNet == 0x0000)
	  pd->dstNet = iflist[ifn].net;

	switch (pd->ddpType) {
	  case RTMP:
	    incr(rtmpInDataPkts);
	    rtmpd(ifn, pd);
	    break;
	  case RTMPR:
	    incr(rtmpInRequestPkts);
	    rtmpr(ifn, pd);
	    break;
	  case NBP:
	    nbp(ifn, pd);
	    break;
	  case AEP:
	    echo(ifn, pd);
	    break;
	  case ZIP:
	    zip(ifn, pd);
	    break;
	  case ATP:
	    incr(atpInPkts);
	    if (pd->dstSkt == ZIP) {
	      zip_atp(ifn, pd);
	      break;
	    }
	    if (pd->dstSkt == TMLDSKT) {
	      timesrvr_atp(ifn, pd);
	      break;
	    }
	    /* fall thro' */
	  case ADSP:
	  case SNMP:
	  default:
	    incr(ddpNoProtocolHandlers);
	    break;
	}

	return;
}

/*
 * packet destined for EtherTalk multinode clients on interface 'ifn'
 *
 */

void
ddp_mn(ifn, pd)
int ifn;
struct pd *pd;
{
	int i;
	void cap_route();

	incr(uarMNodePacketsIn);

	if (pd->dstNet == 0x0000)
	  pd->dstNet = iflist[ifn].net;

	if (pd->dstNode == BCAST) {
	  for (i = 1; i < MAXNODES-1; i++) {
	    if (iflist[ifn].mnode[i].state == MNODE_VALID) {
	      switch (iflist[ifn].mnode[i].client) {
	        case MNODE_CAP:
	          cap_route(ifn, pd);
	          break;
	        default:
	          break;
	      }
	    }
	  }
	} else {
	  if (iflist[ifn].mnode[pd->dstNode].state == MNODE_VALID) {
	    switch (iflist[ifn].mnode[pd->dstNode].client) {
	      case MNODE_CAP:
	        cap_route(ifn, pd);
	        break;
	      default:
	        break;
	    }
	    return;
	  }
	}
}

/*
 * handle AppleTalk Echo Protocol
 *
 */

void
echo(ifn, pd)
int ifn;
struct pd *pd;
{
	short len;
	u_char *pkt;
	struct pd opd;
	struct rtmp *rtmp_route();
	void ddp_route();

	pkt = pd->data;
	len = pd->dataLen;

	incr((pkt[0] == 0x01) ? atechoRequests : atechoReplies);

	if (debug & ECHODEBUG)
	  fprintf(stderr, "ECHO: %-5s Receive Echo %sPacket\n",
	    iflist[ifn].if_name, (pkt[0] == 0x01) ? "Request " : "");

	if (len == 0)
	  drop(ddpTooShortErrors);
	  
	if (len > MAXDDPLEN)
	  drop(ddpTooLongErrors);

	if (pd->dstSkt != AEP)
	  drop(uarNoSocketHandler);

	if (pkt[0] != 0x01)
	  drop(uarBadPacketHdr);

	/* ECHO Response */

	pkt[0] = 0x02;

	opd.lapType = LDDP;
	opd.ddpType = AEP;
	opd.dstNet = pd->srcNet;
	opd.srcNet = iflist[ifn].net;
	opd.dstNode = pd->srcNode;
	opd.srcNode = iflist[ifn].node;
	opd.dstSkt = pd->srcSkt;
	opd.srcSkt = AEP;
	opd.dataLen = pd->dataLen;
	opd.data = pd->data;
	opd.addr = NULL;
	opd.hop  = 0;

	if (debug & ECHODEBUG)
	  fprintf(stderr, "ECHO: %-5s Sending Echo Response Packet\n",
	    iflist[ifn].if_name);

	incr(atechoReplies);
	incr(ddpOutRequests);

	ddp_route(ifn, &opd, rtmp_route(opd.dstNet));

	return;
}

/*
 * handle timeserver requests
 *
 */

#define isleap(yr) ((yr) % 4 == 0 && ((yr) % 100 != 0 || (yr) % 400 == 0))

void
timesrvr_atp(ifn, pd)
int ifn;
struct pd *pd;
{
	short len;
	time_t now;
	u_char *pkt;
	struct pd opd;
	u_int mact, diff, usrbytes;
	struct tm gmt, local, *localtime(), *gmtime();
	struct rtmp *rtmp_route();
	void dumppacket();
	void ddp_route();

	if (iflist[ifn].if_timesrvr[0] == '\0')
	  drop(uarNoSocketHandler);

	pkt = pd->data;
	len = pd->dataLen;

	if (len == 0)
	  drop(ddpTooShortErrors);

	if (len > MAXDDPLEN)
	  drop(ddpTooLongErrors);

	if (pkt[0] == TREL)
	  return;

	if (pkt[0] != (TREQ|XO) || pkt[1] != 0x01)
	  drop(uarBadPacketHdr);

	if (debug & TMLDDEBUG) {
	  fprintf(stderr, "TMLD: %-5s Receive Timelord Request\n",
	    iflist[ifn].if_name);
	  if (debug & PKTDEBUG)
	    dumppacket(pd->data, pd->dataLen);
	}

	usrbytes = (pkt[4] << 24) | (pkt[5] << 16) | (pkt[6] << 8) | pkt[7];

	switch (usrbytes) {
	  case TMLDGETTIME:
	    now = (time_t)tn.tv_sec;
	    gmt = *gmtime((time_t *)&now);
	    local = *localtime((time_t *)&now);
	    diff = gmt.tm_year - local.tm_year;
	    diff *= 365;
	    if (gmt.tm_year > local.tm_year) {
	      if (isleap(local.tm_year))
		diff++;
	    } else {
	      if (local.tm_year > gmt.tm_year) {
	        if (isleap(gmt.tm_year))
		  diff--;
	      }
	    }
	    diff += gmt.tm_yday - local.tm_yday;
	    diff *= 24;
	    diff += gmt.tm_hour - local.tm_hour;
	    diff *= 60;
	    diff += gmt.tm_min - local.tm_min;
	    diff *= 60;
	    diff += gmt.tm_sec - local.tm_sec;
	    mact = now - diff + TMLDOFFSET;
	    pkt[8]  = (mact >> 24) & 0xff;
	    pkt[9]  = (mact >> 16) & 0xff;
	    pkt[10] = (mact >>  8) & 0xff;
	    pkt[11] = (mact & 0xff);
	    usrbytes = TMLDENONE;
	    len = 12;
	    break;
	  default:
	    strcpycp("Bad Request", (char *)(pkt+8));
	    usrbytes = TMLDENOPERM;
	    len = 8+pkt[8];
	    break;
	}

	pkt[0] = TRESP | EOM;
	pkt[1] = 0x00;

	pkt[4] = (usrbytes >> 24) & 0xff;
	pkt[5] = (usrbytes >> 16) & 0xff;
	pkt[6] = (usrbytes >>  8) & 0xff;
	pkt[7] = (usrbytes & 0xff);

	opd.lapType = LDDP;
	opd.ddpType = ATP;
	opd.dstNet  = pd->srcNet;
	opd.srcNet  = iflist[ifn].net;
	opd.dstNode = pd->srcNode;
	opd.srcNode = iflist[ifn].node;
	opd.dstSkt  = pd->srcSkt;
	opd.srcSkt  = TMLDSKT;
	opd.dataLen = len;
	opd.data = pkt;
	opd.addr = NULL;
	opd.hop  = 0;

	if (debug & TMLDDEBUG) {
	  fprintf(stderr, "TMLD: %-5s Sending Timelord Reply %s",
	    iflist[ifn].if_name, ctime(&now));
	  if (debug & PKTDEBUG)
	    dumppacket(opd.data, opd.dataLen);
	}

	incr(ddpOutRequests);

	ddp_route(ifn, &opd, rtmp_route(opd.dstNet));

	return;
}

/*
 * send packet via route specified in rtmp entry
 * ('ifn' is the entry or originating port for the packet)
 *
 */

void
ddp_route(ifn, pd, rtmp)
int ifn;
struct pd *pd;
struct rtmp *rtmp;
{
	short len;
	u_short dstNet;
	u_char dstNode;
	u_short cksum, chkSum();
	u_char *addr, *pkt, *hdr;
	struct aarp *aarp, *aarp_find();
	struct rtmp rtmpn;
	int ddp_for_mn();
	int nbp_filter();
	void aarp_enqueue();
	void dumppacket();
	void ddp_mn();

	/*
	 * check routing, handle Phase 2 startup range
	 *
	 */
	if (rtmp == RTMPNULL) {
	/* no route known to destination */
	  if (iflist[ifn].phase == PHASE2
	   && pd->dstNet >= P2STARTUP) {
	  /* startup network range, route locally */
	    rtmpn.dist = 0;
	    rtmpn.intfc = ifn;
	    rtmpn.rnet = 0;
	    rtmpn.rnode = 0;
	    rtmp = &rtmpn;
	  } else
	    drop(ddpOutNoRoutes);
	}

	if (iflist[rtmp->intfc].state != ADDR_VALID)
	  return;

	if (debug & ROUTEDEBUG) {
	  fprintf(stderr, "ETH%d: %-5s ROUTE %3d.%-3d   %3d",
	    iflist[rtmp->intfc].phase, iflist[rtmp->intfc].if_name,
	    pd->dstNet >> 8, pd->dstNet & 0xff, pd->dstNode);
	  if (rtmp->dist != 0)
	    fprintf(stderr, "   via %3d.%-3d   %3d\n",
	      rtmp->rnet >> 8, rtmp->rnet & 0xff, rtmp->rnode);
	  else
	    fprintf(stderr, "   (local)\n");
	}

	/*
	 * ensure only long DDP sent for Phase 2
	 *
	 */
	if (iflist[rtmp->intfc].phase == PHASE2)
	  pd->lapType = LDDP;

	/*
	 * if local Phase 2 broadcast, make
	 * network-wide (maybe zone-specific)
	 *
	 */
	if (iflist[rtmp->intfc].phase == PHASE2
	 && pd->dstNode == BCAST
	 && rtmp->dist == 0)
	  dstNet = 0x0000;
	else
	  dstNet = pd->dstNet;

	/*
	 * check if packet for EtherTalk multinode client on rtmp->intfc
	 *
	 */
	if (ddp_for_mn(rtmp->intfc, dstNet, pd->dstNode)) {
	  ddp_mn(rtmp->intfc, pd);
	  if (pd->dstNode != BCAST)
	    return;
	}

	/*
	 * construct the DDP header
	 *
	 */
	switch (pd->lapType) {
	  case SDDP:
	    incr(ddpOutShorts);
	    pkt     = pd->data-5;
	    len     = pd->dataLen+5;
	    pkt[0]  = (len >> 8) & 0x03;
	    pkt[1]  = len & 0xff;
	    pkt[2]  = pd->dstSkt;
	    pkt[3]  = pd->srcSkt;
	    pkt[4]  = pd->ddpType;
	    break;
	  case LDDP:
	    incr(ddpOutLongs);
	    pkt     = pd->data-13;
	    len     = pd->dataLen+13;
	    pkt[0]  = (pd->hop << 2) | ((len >> 8) & 0x03);
	    pkt[1]  = len & 0xff;
	    pkt[4]  = dstNet >> 8;
	    pkt[5]  = dstNet & 0xff;
	    pkt[6]  = pd->srcNet >> 8;
	    pkt[7]  = pd->srcNet & 0xff;
	    pkt[8]  = pd->dstNode;
	    pkt[9]  = pd->srcNode;
	    pkt[10] = pd->dstSkt;
	    pkt[11] = pd->srcSkt;
	    pkt[12] = pd->ddpType;
	    cksum   = (checksum) ? chkSum(pkt+4, len-4) : 0x0000;
	    pkt[2]  = cksum >> 8;
	    pkt[3]  = cksum & 0xff;
	    break;
	  default:
	    drop(uarBadLapType);
	    break;
	}

	addr = NULL;

	/*
	 * determine destination ethernet address (and
	 * local destination node for Phase 1 LAP header)
	 *
	 */
	if (pd->dstNode == BCAST
	 && rtmp->dist == 0) {
	  dstNode = BCAST;
	  dstNet  = iflist[rtmp->intfc].net;
	  switch (iflist[rtmp->intfc].itype) {
	    case ITYPE_ETHERTALK:
	      addr = (iflist[rtmp->intfc].phase == PHASE2) ?
	        ((pd->addr) ? pd->addr : maddr) : baddr;
	      break;
	    case ITYPE_TNNL:
	      break;
	    case ITYPE_ARNS:
	      break;
	  }
	} else {
	  dstNode = (rtmp->rnode == 0) ? pd->dstNode : rtmp->rnode;
	  dstNet  = (rtmp->rnet  == 0) ? pd->dstNet  : rtmp->rnet;
	  switch (iflist[rtmp->intfc].itype) {
	    case ITYPE_ETHERTALK:
	      if ((aarp = aarp_find(dstNet, dstNode)) == AARPNULL) {
	        aarp_enqueue(rtmp->intfc, dstNet, dstNode, AARPReq);
	        return;
	      }
	      addr = aarp->eaddr;
	      break;
	    case ITYPE_TNNL:
	      break;
	    case ITYPE_ARNS:
	      break;
	  }
	}

	/*
	 * if for "internal" interface, write DDP packet & extended LAP hdr
	 *
	 */
	if (iflist[rtmp->intfc].itype != ITYPE_ETHERTALK) {
	  len += 4;
	  hdr = pkt-4;
	  hdr[0] = 0x00;
	  hdr[1] = dstNode;
	  hdr[2] = iflist[rtmp->intfc].node;
	  hdr[3] = pd->lapType;
	  if (iflist[rtmp->intfc].write != NULL)
	    (*(iflist[rtmp->intfc].write))(rtmp->intfc, hdr, len);
	  return;
	}

	if (addr == NULL)
	  return;

	/*
	 * construct the EtherTalk header
	 *
	 */
	switch (iflist[rtmp->intfc].phase) {
	  case PHASE1:
	    hdr = pkt-17;
	    hdr[12] = 0x80;
	    hdr[13] = 0x9b;
	    hdr[14] = dstNode;
	    hdr[15] = iflist[rtmp->intfc].node;
	    hdr[16] = pd->lapType;
	    len += 17;
	    break;
	  case PHASE2:
	    len += 8;
	    hdr = pkt-22;
	    hdr[12] = len >> 8;		/* 802.3 length field	*/
	    hdr[13] = len & 0xff;	/* 802.3 length field	*/
	    hdr[14] = 0xaa;		/* 802.2 hdr dst SAP	*/
	    hdr[15] = 0xaa;		/* 802.2 hdr src SAP	*/
	    hdr[16] = 0x03;		/* 802.2 hdr ctrl byte	*/
	    hdr[17] = 0x08;		/* SNAP prot. discrim	*/
	    hdr[18] = 0x00;		/* SNAP prot. discrim	*/
	    hdr[19] = 0x07;		/* SNAP prot. discrim	*/
	    hdr[20] = 0x80;		/* SNAP prot. discrim	*/
	    hdr[21] = 0x9b;		/* SNAP prot. discrim	*/
	    len += 14;
	    break;
	  default:
	    drop(uarUnknownPhase);
	    break;
	}

	bcopy((char *)addr, (char *)hdr, 6);
	bcopy((char *)iflist[rtmp->intfc].eaddr, (char *)(hdr+6), 6);

	if (debug & ETHDEBUG) {
	  fprintf(stderr, "ETH%d: %-5s WRITE %3d bytes\n",
	    iflist[rtmp->intfc].phase, iflist[rtmp->intfc].if_name, len);
	  if (debug & PKTDEBUG)
	    dumppacket(hdr, len);
	}

	if (write_eth(iflist[rtmp->intfc].etfd, hdr, len) != len)
	  if (debug) perror("write_eth()");

	return;
}

/*
 * DDP checksum calculator
 *
 */

u_short
chkSum(pkt, len)
register u_char *pkt;
register int len;
{
	register int sum = 0;

	while (len-- > 0) {
	  sum = ((sum&0xffff)+*pkt++) << 1;
	  if (sum&0x10000)
	    sum++;
	}
	sum &= 0xffff;
	return((sum == 0) ? 0xffff : sum);
}
