/*
 * UAR - UNIX AppleTalk Router
 * 
 * Copyright (c) 1993, The University of Melbourne.
 *
 * All Rights Reserved.
 *
 * UAR may NOT be publicly redistributed (for example via
 * anonymous FTP), sold, or the source used for any other
 * purpose without the permission of the copyright owner.
 * 
 * This software is supplied "as is" without express or
 * implied warranty.
 * 
 * djh@munnari.OZ.AU
 *
 * $Author: djh $
 * $Revision: 1.1.1.1 $
 *
 */

#include "uar.h"

extern int debug;
extern u_char baddr[];
extern u_char maddr[];
extern u_long uar_stats[];
extern u_long pkts_dropped;
extern struct iflist iflist[];
extern struct timeval tn;
extern char *stats_msg[];

int nbpOutID;
int nbp_do_dump = 0;

struct nbpq *nbpQueue;
struct nbptyp *nbpCache;

/*
 * Initialize NBP data structures
 *
 */

void
nbp_initialize()
{
	nbpQueue = NBPQNULL;
	nbpCache = NBPTYPNULL;

	nbpOutID = 1;

	return;
}

/*
 * process an incoming NBP packet
 *
 */

void
nbp(ifn, pd)
int ifn;
struct pd *pd;
{
	short len;
	u_char *pkt;
	char *nbp_type();
	void nbp_brreq(), nbp_lkupreq();
	void nbp_fwdreq(), nbp_lkupreply();

	pkt = pd->data;
	len = pd->dataLen;

	if (debug & NBPDEBUG)
	  fprintf(stderr, " NBP: %-5s Receive %s\n",
	    iflist[ifn].if_name, nbp_type(pkt[0]));

	if (len == 0)
	  drop(ddpTooShortErrors);
	  
	if (len > MAXDDPLEN)
	  drop(ddpTooLongErrors);

	if (pd->dstSkt != NBP)
	  drop(uarNoSocketHandler);

	switch (pkt[0] >> 4) {
	  case NBPBRRQ:
	    incr(nbpInBroadcastRequests);
	    nbp_brreq(ifn, pd);
	    break;
	  case NBPLKUPREQ:
	    incr(nbpInLookUpRequests);
	    nbp_lkupreq(ifn, pd);
	    break;
	  case NBPLKUPREPLY:
	    incr(nbpInLookUpReplies);
	    nbp_lkupreply(ifn, pd);
	    break;
	  case NBPFWDREQ:
	    incr(nbpInForwardRequests);
	    nbp_fwdreq(ifn, pd);
	    break;
	  default:
	    drop(nbpInErrors);
	    break;
	}

	return;
}

/*
 * The NBP process in the requesting node prepares an NBP Broadcast
 * Request (BrRq packet) and sends it to the NIS of A-ROUTER. The
 * NBP process in the router, in cooperation with the NBP process
 * in the other routers of the Internet, arranges to convert the
 * BrRq into one Forward Request (FwdReq) packet for each network
 * that contains nodes in the target zone of the lookup request.
 *
 * Inside AppleTalk, 2nd Edition, page 7-10
 *
 */

void
nbp_brreq(ifn, pd)
int ifn;
struct pd *pd;
{
	short len, tuples, zlen;
	u_char *addr, *zone_mcast();
	u_char *pkt, *ent, *obj, *typ, *zon;
	struct rtmp *rtmp, *rtmp_route();
	struct zone *zone, *zone_find();
	struct zit *zit;
	struct net *net;
	u_short srcNet;
	void nbp_send();
	int nbp_invalid();
	int nbp_match_tuple();
	void nbp_add_lookup();
	void dumppacket();

	pkt = pd->data;
	len = pd->dataLen;

	if ((tuples = pkt[0] & 0x0f) != 0x01)
	/* not a single tuple */
	  drop(nbpInErrors);

	ent = pkt+NBPHDRSIZ;
	obj = ent+ENTHDRSIZ;
	typ = obj+*obj+1;
	zon = typ+*typ+1;

	if (nbp_invalid(obj, typ, zon, pkt, len))
	  drop(nbpInErrors);

	zone = ZONENULL;

	if (zon[0] == 0x00 || (zon[0] == 0x01 && zon[1] == '*')) {
	  srcNet = (ent[0] << 8) | ent[1];

	  if ((rtmp = rtmp_route(srcNet)) == RTMPNULL)
	    drop(uarRTMPCacheFail);

	  if (rtmp->zquantity > 1)
	    drop(nbpInErrors);

	  if ((zit = rtmp->zitlist) == ZITNULL)
	    drop(nbpInErrors);

	  if ((zone = zit->zone) == ZONENULL)
	    drop(nbpInErrors);

	  len -= zon[0];
	  bcopy((char *)zone->name, (char *)zon, (int)(zone->name[0]+1));
	  len += zone->name[0];
	  zon = zone->name;
	}

	zlen = *zon++;
	pd->dataLen = len;

	if (zone == ZONENULL)
	  if ((zone = zone_find(zon, zlen)) == ZONENULL)
	    drop(uarNoDefaultZone);

	pd->srcNet = iflist[ifn].net;
	pd->srcNode = iflist[ifn].node;
	pd->srcSkt = NBP;
	pd->dstSkt = NBP;

	addr = zone_mcast(zon, zlen);
	net = zone->nets;

	while (net != NETNULL) {
	  if ((rtmp = rtmp_route(net->net_lo)) != RTMPNULL) {
	    if (rtmp->dist == 0 || iflist[rtmp->intfc].phase == PHASE1) {
	      pkt[0] = (NBPLKUPREQ << 4) | tuples;
	      pd->addr = (iflist[rtmp->intfc].phase == PHASE2) ? addr : NULL;
	      pd->dstNode = BCAST;
	    } else { /* Phase 2, not local */
	      pkt[0] = (NBPFWDREQ << 4) | tuples;
	      pd->addr = NULL;
	      pd->dstNode = 0x00;
	    }
	    pd->dstNet = net->net_lo;
	    nbp_send(ifn, pd, rtmp);
	  }
	  net = net->next;
	}

	return;
}

/*
 * Each of these FwdReq packets is then sent to the NIS in any
 * router directly connected to the corresponding network (these
 * packets are addressed to DDP node address zero). When this router
 * receives the FwdReq, it converts the FwdReq to a LkUp packet and
 * broadcasts it to the NIS in all nodes in the target zone on the
 * destination network.
 *
 * Inside AppleTalk, 2nd Edition, page 7-10
 *
 */

void
nbp_fwdreq(ifn, pd)
int ifn;
struct pd *pd;
{
	short len, tuples, zlen;
	u_char *pkt, *ent, *obj, *typ, *zon, *zone_mcast();
	struct rtmp *rtmp, *rtmp_route();
	int nbp_invalid();
	void nbp_add_lookup();
	void dumppacket();
	void nbp_send();

	pkt = pd->data;
	len = pd->dataLen;

	if (iflist[ifn].phase == PHASE1)
	  drop(nbpInErrors);

	if (pd->dstNode != 0x00)
	  drop(nbpInErrors);

	if ((tuples = pkt[0] & 0x0f) != 0x01)
	/* not a single tuple */
	  drop(nbpInErrors);

	ent = pkt+NBPHDRSIZ;
	obj = ent+ENTHDRSIZ;
	typ = obj+*obj+1;
	zon = typ+*typ+1;

	if (nbp_invalid(obj, typ, zon, pkt, len))
	  drop(nbpInErrors);

	if (zon[0] == 0x00 || (zon[0] == 0x01 && zon[1] == '*'))
	  drop(nbpInErrors);

	if ((rtmp = rtmp_route(pd->dstNet)) == RTMPNULL)
	  drop(uarRTMPCacheFail);

	zlen = *zon++;

	if (rtmp->dist == 0) {
	  pkt[0] = (NBPLKUPREQ << 4) | tuples;
	  pd->addr = zone_mcast(zon, zlen);
	  pd->dstNode = BCAST;
	}

	pd->srcNet = iflist[ifn].net;
	pd->srcNode = iflist[ifn].node;
	pd->srcSkt = NBP;
	pd->dstSkt = NBP;

	nbp_send(ifn, pd, rtmp);

	return;
}

/*
 * process an incoming NBP lookup reply
 * (to our NBP registration attempt)
 *
 */

void
nbp_lkupreply(ifn, pd)
int ifn;
struct pd *pd;
{
	register u_char *ent, *obj, *typ, *zon, *nxt;
	u_char object[NBPSTR], type[NBPSTR], zone[NBPSTR];
	u_char *pkt, nbpID;
	short len, tuples;
	int nbp_invalid();
	void nbp_dequeue();
	void strcpypc();

	pkt = pd->data;
	len = pd->dataLen;

	tuples = pkt[0] & 0x0f;
	nbpID  = pkt[1];

	ent = pkt+NBPHDRSIZ;

	while (tuples-- > 0) {

	  obj = ent+ENTHDRSIZ;
	  typ = obj+*obj+1;
	  zon = typ+*typ+1;
	  nxt = zon+*zon+1;

	  if (nbp_invalid(obj, typ, zon, pkt, len))
	    drop(nbpInErrors);

	  strcpypc((char *)obj, (char *)object);
	  strcpypc((char *)typ, (char *)type);
	  strcpypc((char *)zon, (char *)zone);

	  if (debug & NBPDEBUG)
	    fprintf(stderr, " NBP: %-5s Reply tuple %s:%s@%s\n",
	      iflist[ifn].if_name, object, type, zone);

	  /*
	   * "soft" dequeue of NBP lookup
	   * (ie: modify object and try again)
	   *
	   */
	  nbp_dequeue(ent, zon-ent, nbpID, 0);

	  ent = nxt;
	}

	return;
}

/*
 * process an incoming NBP lookup packet
 *
 */

void
nbp_lkupreq(ifn, pd)
int ifn;
struct pd *pd;
{
	short len;
	register u_char *pkt, *ent, *obj, *typ, *zon, nbpID;
	u_char object[NBPSTR], type[NBPSTR], zone[NBPSTR];
	int scmp, objWild, typWild, strcmpci(), nbp_wild();
	int nbp_match(), nbp_zonmatch(), nbp_invalid();
	int ddp_is_gw(), ddp_is_mn();
	struct nbpobj *r, *s, *t;
	struct nbptyp *p, *q;
	u_short entNet;
	u_char entNode;
	void nbp_reply();
	void strcpypc();

	if (nbpCache == NBPTYPNULL)
	/* no entries */
	  return;

	pkt = pd->data;
	len = pd->dataLen;

	if ((pkt[0] & 0x0f) != 0x01)
	/* not a single tuple */
	  drop(nbpInErrors);

	nbpID = pkt[1];

	ent = pkt+NBPHDRSIZ;
	obj = ent+ENTHDRSIZ;
	typ = obj+*obj+1;
	zon = typ+*typ+1;

	if (nbp_invalid(obj, typ, zon, pkt, len))
	  drop(nbpInErrors);

	strcpypc((char *)obj, (char *)object);
	strcpypc((char *)typ, (char *)type);
	strcpypc((char *)zon, (char *)zone);

	if (debug & NBPDEBUG)
	  fprintf(stderr, " NBP: %-5s Look Up for %s:%s@%s\n",
	    iflist[ifn].if_name, object, type, zone);

	objWild = nbp_wild(object); /* wild card ? */
	typWild = nbp_wild(type);   /* wild card ? */

	p = nbpCache;
	q = NBPTYPNULL;

	/*
	 * search cache for matching entries, if there are wild
	 * cards in object or type, search is effectively linear
	 *
	 */
	while (p != NBPTYPNULL) {
loop:	  scmp = -1;
	  if (!typWild) {
	    if ((scmp = strcmpci(p->nbptyp+1, type)) < 0) {
	      q = p;
	      p = p->next;
	      continue;
	    }
	    if (scmp > 0)
	      break;
	  }
	  if (scmp == 0 || nbp_match(type, p->nbptyp+1)) {
	    /*
	     * found the NBP type, now look for object
	     *
	     */
	    r = p->list;
	    s = NBPOBJNULL;

	    while (r != NBPOBJNULL) {
	      scmp = -1;
	      if (!objWild) {
	        if ((scmp = strcmpci(r->nbpobj+1, object)) < 0) {
	          s = r;
	          r = r->next;
	          continue;
	        }
	        if (scmp > 0)
	          break;
	      }
	      if (scmp == 0 || nbp_match(object, r->nbpobj+1)) {
	      /*
	       * check to see if the entry is valid for the interface
	       *
	       */
	        entNet = (r->nbpent[0] << 8) | r->nbpent[1];
	        entNode = r->nbpent[2];
	        if (!((ddp_is_gw(r->intfc, entNet, entNode))
	         || (ddp_is_mn(r->intfc, entNet, entNode)))
	         || (r->when != 0 && tn.tv_sec > r->when+NBPCACHETIMEOUT)) {
	          /*
	           * not connected/expired, delete NBP cache entry
	           *
	           */
	          if (debug & NBPDEBUG)
	            fprintf(stderr," NBP: %-5s expired, deleting %s:%s@%s\n",
	              iflist[r->intfc].if_name, r->nbpobj+1,
	              p->nbptyp+1, r->zone+1);

	          if (s == NBPOBJNULL)
	            p->list = r->next;
	          else
	            s->next = r->next;
	          t = r->next;
	          free((char *)r);
	          r = t;

	          if (p->list == NBPOBJNULL) {
	            /*
	             * no objects left for this type
	             *
	             */
	            if (debug & NBPDEBUG)
	              fprintf(stderr," NBP: %-5s empty, deleting type %s\n",
	                iflist[ifn].if_name, p->nbptyp+1);

	            if (q == NBPTYPNULL)
	              nbpCache = p->next;
	            else
	              q->next = p->next;
	            free((char *)p);

	            /*
	             * start again
	             *
	             */
	            q = NBPTYPNULL;

	            if ((p = nbpCache) == NBPTYPNULL)
	              return;

	            goto loop;
	          }
	          continue;
	        }
	        /*
	         * send a single packet NBP reply per match
	         *
	         */
	        if (nbp_zonmatch(zone, r->zone+1))
	          nbp_reply(ifn, ent, r->nbpent, p->nbptyp, r->zone, nbpID);
	      }
	      s = r;
	      r = r->next;
	    }
	  }
	  q = p;
	  p = p->next;
	}

	return;
}

/*
 * reply to the originator of the NBP request
 * (address in req) with the NBP data in rep.
 *
 */

void
nbp_reply(ifn, req, rep, typ, zon, nbpID)
int ifn;
u_char *req, *rep, *typ, *zon, nbpID;
{
	short len, nbp_pack();
	u_char *pkt, *obj, buf[160];
	struct rtmp *rtmp_route();
	struct pd opd;
	int nbp_invalid();
	void nbp_send();

	/*
	 * drop if request from same net/node
	 *
	 */
	if (req[0] == rep[0]
	 && req[1] == rep[1]
	 && req[2] == rep[2])
	  return;

	obj = rep+ENTHDRSIZ;

	if (nbp_invalid(obj, typ, zon, (u_char *)NULL, 0))
	  return;

	len = 0;

	opd.lapType = LDDP;
	opd.ddpType = NBP;
	opd.dstNet = (req[0] << 8) | req[1];
	opd.srcNet = (rep[0] << 8) | rep[1];
	opd.dstNode = req[2];
	opd.srcNode = rep[2];
	opd.dstSkt = req[3];
	opd.srcSkt = NBP;
	opd.data = buf+HDRSIZ+LDDPHDRSIZ;
	opd.dataLen = len;
	opd.addr = NULL;
	opd.hop = 0;

	pkt = opd.data;

	pkt[0] = (NBPLKUPREPLY << 4) | 0x01;	/* LkUp Reply	*/
	pkt[1] = nbpID;				/* lookup ID	*/

	len += NBPHDRSIZ;
	pkt += NBPHDRSIZ;

	bcopy((char *)rep, (char *)pkt, ENTHDRSIZ);

	pkt += ENTHDRSIZ;
	len += ENTHDRSIZ;

	len += nbp_pack(pkt, obj, typ, zon);

	opd.dataLen = len;

	nbp_send(ifn, &opd, rtmp_route(opd.dstNet));

	return;
}

/*
 * send an NBP packet, possibly to a multicast zone address
 *
 */

void
nbp_send(ifn, pd, rtmp)
int ifn;
struct pd *pd;
struct rtmp *rtmp;
{
	char *nbp_type();
	int intfc, ddp_for_gw();
	void ddp_gw(), ddp_route();

	pd->lapType = LDDP;
	pd->hop = 0;

	intfc = (rtmp == NULL) ? ifn : rtmp->intfc;

	if (debug & NBPDEBUG) {
	  fprintf(stderr, " NBP: %-5s Sending %s to %3d.%-3d %3d",
	    iflist[intfc].if_name, nbp_type(pd->data[0]),
	    pd->dstNet >> 8, pd->dstNet & 0xff, pd->dstNode);
	  if (pd->addr != NULL)
	    fprintf(stderr, " %02x:%02x:%02x:%02x:%02x:%02x",
	      pd->addr[0], pd->addr[1], pd->addr[2],
	      pd->addr[3], pd->addr[4], pd->addr[5]);
	  fprintf(stderr, "\n");
	}

	/*
	 * also send a copy of LKUPs and LKUPREPLYs to gw
	 *
	 */
	if ((pd->data[0] >> 4) == NBPLKUPREQ
	 || (pd->data[0] >> 4) == NBPLKUPREPLY) {
	  if (ddp_for_gw(intfc, pd->dstNet, pd->dstNode)) {
	    ddp_gw(intfc, pd);
	    if (pd->dstNode != BCAST)
	      return;
	  }
	}

	incr(ddpOutRequests);

	ddp_route(ifn, pd, rtmp);

	return;
}

/*
 * return a string describing the NBP packet type
 *
 */

char *
nbp_type(fn)
u_char fn;
{
	switch (fn >> 4) {
	  case NBPBRRQ:
	    return("Broadcast Request");
	    break;
	  case NBPLKUPREQ:
	    return("Look Up");
	    break;
	  case NBPLKUPREPLY:
	    return("LookUp Reply");
	    break;
	  case NBPFWDREQ:
	    return("Forward Request");
	    break;
	  default:
	    break;
	}

	return("Unknown");
}

/*
 * the NBP process
 *
 * send NBP LKUPs for pending NBP registrations
 * (use our node in entity so we get any responses)
 *
 */

void
nbp_process()
{
	short zlen;
	u_char skt, nod, *zon;
	struct nbpq *p, *q, *t;
	struct zone *zone, *zone_default();
	void nbp_brreq(), nbp_insert();

	p = nbpQueue;
	q = NBPQNULL;

	while (p != NBPQNULL) {
	  if (iflist[p->intfc].state == ADDR_VALID) {
	    if (p->zonlen != 0) { /* have default zone */
	      if (p->timer > 0 && ++p->period >= p->timer) {
	        p->period = 0;
	        if (p->count-- > 0) {
	          /* send an NBP lookup */
	          nod = p->opd.data[4];
	          p->opd.data[4] = iflist[p->intfc].node;
	          skt = p->opd.data[5];
	          p->opd.data[5] = NBP;
	          nbp_brreq(p->intfc, &p->opd);
	          p->opd.dataLen = p->dataLen;
	          p->opd.data[4] = nod;
	          p->opd.data[5] = skt;
	        } else {
	          /* finished, add NBP data to cache (no timeout) */
	          nbp_insert(p->intfc, p->opd.data, p->dataLen, 1);
	          /* dequeue the request */
	          if (q == NBPQNULL)
	            nbpQueue = p->next;
	          else
	            q->next = p->next;
	          t = p->next;
	          free((char *)p);
	          p = t;
	          continue;
	        }
	      }
	    } else { /* try to find out the default zone */
	      if ((zone = zone_default(p->intfc)) != ZONENULL) {
	        if ((zlen = zone->name[0]) <= MAXNBP) {
	          p->zonlen = zlen+1;
	          zon = p->opd.data + p->dataLen;
	          bcopy((char *)zone->name, (char *)zon, p->zonlen);
	          p->dataLen += p->zonlen;
	          p->opd.dataLen = p->dataLen;
	          continue;
	        }
	      }
	    }
	  }
	  q = p;
	  p = p->next;
	}

	return;
}

/*
 * register an NBP entity
 *
 */

void
nbp_register(ifn, obj, typ, zon, addr, count, timer)
int ifn;
u_char *obj, *typ, *zon;
struct ataddr *addr;
short count, timer;
{
	short len, nbp_pack();
	u_char *pkt, data[160];
	u_char object[NBPSTR], type[NBPSTR], zone[NBPSTR];
	int nbp_invalid(), nbp_wild();
	void nbp_enqueue(), strcpypc();

	if (nbp_invalid(obj, typ, zon, (u_char *)NULL, 0))
	  return;

	strcpypc((char *)obj, (char *)object);
	strcpypc((char *)typ, (char *)type);
	strcpypc((char *)zon, (char *)zone);

	if (nbp_wild(object))
	  return;

	if (nbp_wild(type))
	  return;

	if (!(zone[0] == 0x00 || (zone[0] == '*' && zone[1] == 0x00)))
	  return;

	if (debug & NBPDEBUG)
	  fprintf(stderr, " NBP: %-5s Registering %s:%s@%s\n",
	    iflist[ifn].if_name, object, type, zone);

	data[0] = (NBPBRRQ << 4) | 0x01;
	data[1] = (nbpOutID++) & 0xff;
	data[2] = addr->net >> 8;
	data[3] = addr->net & 0xff;
	data[4] = addr->node;
	data[5] = addr->skt;
	data[6] = 0x01;

	pkt = data+NBPHDRSIZ+ENTHDRSIZ;
	len = NBPHDRSIZ+ENTHDRSIZ;

	len += nbp_pack(pkt, obj, typ, (u_char *)NULL); /* no zone */

	nbp_enqueue(ifn, data, len, count, timer);

	return;
}

/*
 * mark a node's NBP cache entry for deletion
 *
 */

void
nbp_delete(ifn, net, node, skt)
int ifn;
u_short net;
u_char node, skt;
{
	register struct nbptyp *p;
	register struct nbpobj *r;

	p = nbpCache;

	while (p != NBPTYPNULL) {
	  r = p->list;
	  while (r != NBPOBJNULL) {
	    if (ifn == r->intfc
	     && node == r->nbpent[2]
	     && (skt == r->nbpent[3] || skt == 0x00)
	     && net == ((r->nbpent[0] << 8) | r->nbpent[1])) {
	      r->when = 1;
	    }
	    r = r->next;
	  }
	  p = p->next;
	}

	return;
}

/*
 * queue an NBP lookup for the NBP process
 *
 */

void
nbp_enqueue(ifn, data, len, count, timer)
int ifn;
u_char *data;
short len, count, timer;
{
	short siz;
	u_char *ent;
	struct nbpq *nbpq;

	if (len > (sizeof(nbpq->data)-HDRSIZ-LDDPHDRSIZ))
	  return;

	nbpq = nbpQueue;

	/*
	 * check if ent/obj/typ already queued
	 *
	 */
	while (nbpq != NBPQNULL) {
	  ent = nbpq->opd.data+NBPHDRSIZ;
	  siz = nbpq->dataLen-nbpq->zonlen;
	  if (len == siz) {
	    if (bcmp((char *)(data+NBPHDRSIZ), (char *)ent, siz-NBPHDRSIZ) == 0)
	    /* same ent/obj/typ, ignore */
	      return;
	    if (bcmp((char *)(data+NBPHDRSIZ+ENTHDRSIZ),
	      (char *)(ent+ENTHDRSIZ), siz-NBPHDRSIZ-ENTHDRSIZ) == 0) {
	    /* same obj/typ, update ent, reset timer */
	      bcopy((char *)(data+NBPHDRSIZ), (char *)ent, ENTHDRSIZ);
	      nbpq->timer = timer;
	      nbpq->count = count;
	      return;
	    }
	  }
	  nbpq = nbpq->next;
	}

	/*
	 * create a new queue element
	 *
	 */
	if ((nbpq = (struct nbpq *)malloc(sizeof(struct nbpq))) == NBPQNULL) {
	  incr(uarMemoryError);
	  return;
	}

	nbpq->intfc = ifn;
	nbpq->dataLen = len;
	nbpq->timer = timer;
	nbpq->count = count;
	nbpq->period = 0;
	nbpq->zonlen = 0;
	nbpq->modifier = NULL;

	nbpq->opd.lapType = LDDP;
	nbpq->opd.ddpType = NBP;
	nbpq->opd.dstNet = iflist[ifn].net;
	nbpq->opd.srcNet = (data[2] << 8) | data[3];
	nbpq->opd.dstNode = iflist[ifn].node;
	nbpq->opd.srcNode = data[4];
	nbpq->opd.dstSkt = NBP;
	nbpq->opd.srcSkt = NBP;
	nbpq->opd.data = nbpq->data+HDRSIZ+LDDPHDRSIZ;
	nbpq->opd.dataLen = len;
	nbpq->opd.addr = NULL;
	nbpq->opd.hop = 0;

	bcopy((char *)data, (char *)nbpq->opd.data, (int)len);

	/*
	 * queue it
	 *
	 */
	nbpq->next = nbpQueue;
	nbpQueue = nbpq;

	return;
}

/*
 * dequeue pending NBP registration
 *
 * if a "hard" dequeue, remove an NBP lookup from the
 * queue otherwise modify object and reset counters.
 *
 * (data and len include entity, object and type)
 *
 */

void
nbp_dequeue(data, len, nbpID, hard)
u_char *data;
short len;
u_char nbpID;
int hard;
{
	short siz, off;
	u_char *qobj, *obj;
	struct nbpq *p, *q, *t;
	int nbp_modify();

	p = nbpQueue;
	q = NBPQNULL;

	off = NBPHDRSIZ+ENTHDRSIZ;
	siz = len-ENTHDRSIZ;
	obj = data+ENTHDRSIZ;

	while (p != NBPQNULL) {
	  if (p->opd.data[1] == nbpID
	   && (p->dataLen-p->zonlen-off) == siz) {
	    /* same NBP ID and obj/typ length */
	    qobj = p->opd.data+off; /* get queued object */
	    if (bcmp((char *)qobj, (char *)obj, siz) == 0) {
	      /* found same obj/typ fields */
	      if (!hard)
	        hard = nbp_modify(p);
	      if (hard) {
	        /* dequeue request */
	        if (q == NBPQNULL)
	          nbpQueue = p->next;
	        else
	          q->next = p->next;
	        t = p->next;
	        free((char *)p);
	        p = t;
	        continue;
	      }
	    }
	  }
	  q = p;
	  p = p->next;
	}

	return;
}

/*
 * modify NBP object for retry
 *
 */

int
nbp_modify(p)
struct nbpq *p;
{
	u_char *obj, *typ, *zon;

	p->count = 8;
	p->timer = 5;
	p->period = 0;

	obj = p->opd.data+NBPHDRSIZ+ENTHDRSIZ;

	/*
	 * add or increment object modifier
	 * (trailing -N where N is 0-9, A-Z)
	 *
	 */
	if (p->modifier == NULL) {
	  typ = obj+*obj+1;
	  zon = typ+*typ+1;
	  if (*obj < (MAXNBP-2))
	    bcopy((char *)typ, (char *)(typ+2), (int)(*typ+*zon+2));
	  else
	    typ -= 2;
	  typ[0] = '-';
	  typ[1] = '0';
	  p->modifier = typ+1;
	  if (*obj < (MAXNBP-2)) {
	    p->opd.dataLen += 2;
	    p->dataLen += 2;
	    *obj += 2;
	  }
	} else {
	  *p->modifier++;
	  if (*p->modifier > '9'
	   && *p->modifier < 'A')
	    *p->modifier = 'A';
	}

	return(*p->modifier > 'Z');
}

/*
 * insert NBP tuple(s) into the NBP cache
 *
 * The NBP cache is a linked list sorted by NBP type. Each cache
 * node contains a pointer to a linked list ordered by NBP object.
 *
 */

void
nbp_insert(ifn, pkt, len, perm)
int ifn;
u_char *pkt;
short len;
int perm;
{
	register struct nbpobj *r, *s, *t;
	register struct nbptyp *m, *p, *q;
	register u_char *ent, *obj, *typ, *zon, *nxt;
	u_char object[NBPSTR], type[NBPSTR], zone[NBPSTR];
	int scmp, tuples, strcmpci(), nbp_invalid();
	struct zone *z, *zone_default();
	struct nbpobj *newTuple();
	struct nbptyp *newCache();
	void strcpypc();

	tuples = pkt[0] & 0x0f;

	ent = pkt+NBPHDRSIZ;

	while (tuples-- > 0) {
	  obj = ent+ENTHDRSIZ;
	  typ = obj+*obj+1;
	  zon = typ+*typ+1;
	  nxt = zon+*zon+1;

	  if (nbp_invalid(obj, typ, zon, pkt, len))
	    return;

	  /*
	   * don't store "null" zone names if default known
	   *
	   */
	  if (zon[0] == 0x00 || (zon[0] == 0x01 && zon[1] == '*'))
	    if ((z = zone_default(ifn)) != NULL)
	      zon = z->name;

	  strcpypc((char *)obj, (char *)object);
	  strcpypc((char *)typ, (char *)type);
	  strcpypc((char *)zon, (char *)zone);

	  if (debug & NBPDEBUG)
	    fprintf(stderr, " NBP: %-5s inserting tuple %s:%s@%s\n",
	      iflist[ifn].if_name, object, type, zone);

	  /*
	   * create initial list entry
	   *
	   */
	  if (nbpCache == NBPTYPNULL) {
	    if (debug & NBPDEBUG)
	      fprintf(stderr, " NBP: initialising cache\n");
	    if ((r = newTuple(ifn,ent,obj,zon,NBPOBJNULL,perm)) == NBPOBJNULL)
	      return;
	    if ((nbpCache = newCache(typ, r, NBPTYPNULL)) == NBPTYPNULL)
	      free((char *)r);
	    return;
	  }

	  p = nbpCache;
	  q = NBPTYPNULL;

	  while (p != NBPTYPNULL) {
	    if ((scmp = strcmpci(p->nbptyp+1, type)) < 0) {
	      q = p;
	      p = p->next;
	      continue;
	    }
	    if (scmp > 0)
	      break;
	    if (scmp == 0) {
	      /*
	       * found NBP type
	       *
	       */
	      r = p->list;
	      s = NBPOBJNULL;

	      while (r != NBPOBJNULL) {
	        if ((scmp = strcmpci(r->nbpobj+1, object)) < 0) {
	  	  s = r;
	  	  r = r->next;
	  	  continue;
	        }
	        if (scmp > 0)
	  	  break;
	        if (scmp == 0) {
	          /*
	  	   * found NBP object
	  	   *
	  	   */
	  	  if (debug & NBPDEBUG)
	  	    fprintf(stderr, " NBP: updating %s:%s@%s\n",
	  	      object, type, zone);
	  	  bcopy((char *)ent, (char *)r->nbpent, ENTHDRSIZ);
	          r->when = (perm) ? 0 : tn.tv_sec;
	  	  r->intfc = ifn;
	  	  goto next;
	        }
	      }
	      /*
	       * object not found, insert it
	       *
	       */
	      if ((t = newTuple(ifn, ent, obj, zon, r, perm)) == NBPOBJNULL)
	        return;

	      if (s == NBPOBJNULL)
	        p->list = t;
	      else
	        s->next = t;

	      return;
	    }
	  }

	  /*
	   * type isn't in cache list, create a new entry
	   *
	   */
	  if ((r = newTuple(ifn,ent,obj,zon,NBPOBJNULL,perm)) == NBPOBJNULL)
	    return;

	  if ((m = newCache(typ, r, p)) == NBPTYPNULL) {
	    free((char *)r);
	    return;
	  }

	  /*
	   * link it into cache
	   *
	   */
	  if (q == NBPTYPNULL)
	    nbpCache = m;
	  else
	    q->next = m;

	  /*
	   * get next tuple
	   *
	   */
next:	  ent = nxt;
	}

	return;
}

/*
 * create a new tuple (NBP tuple without a type field)
 * (make the object field null terminated for convenience)
 *
 */

struct nbpobj *
newTuple(ifn, ent, obj, zon, next, perm)
int ifn;
register u_char *ent, *obj, *zon;
struct nbpobj *next;
int perm;
{
	register struct nbpobj *t;

	if ((t = (struct nbpobj *)malloc(sizeof(*t))) == NBPOBJNULL)
	  return(NBPOBJNULL);

	bcopy((char *)ent, (char *)t->nbpent, ENTHDRSIZ);
	bcopy((char *)obj, (char *)t->nbpobj, (int)(*obj+1));
	bcopy((char *)zon, (char *)t->zone, (int)(*zon+1));
	t->nbpobj[*obj+1] = '\0';
	t->zone[*zon+1] = '\0';
	t->when = (perm) ? 0 : tn.tv_sec;
	t->intfc = ifn;
	t->next = next;

	return(t);
}

/*
 * create a new cache entry (a list of ordered objects with NBP type)
 * (make the type field null terminated for convenience)
 *
 */

struct nbptyp *
newCache(typ, tuple, next)
register u_char *typ;
struct nbpobj *tuple;
struct nbptyp *next;
{
	register struct nbptyp *c;

	if ((c = (struct nbptyp *)malloc(sizeof(*c))) == NBPTYPNULL)
	  return(NBPTYPNULL);

	bcopy((char *)typ, (char *)c->nbptyp, (int)(*typ+1));
	c->nbptyp[*typ+1] = '\0';
	c->list = tuple;
	c->next = next;

	return(c);
}

/*
 * dump the NBP cache
 *
 */

void
nbp_dump()
{
	int i, j, count;
	char entity[100];
	FILE *r, *fopen();
	struct nbptyp *nbptyp;
	struct nbpobj *nbpobj;
	extern int numintfc;

	if ((r = fopen(DUMP_NBP, "w+")) == NULL) {
	  incr(uarBadDumpFile);
	  return;
	}

	fprintf(r, "UAR NBP Cache:\n\n");
	fprintf(r, "  net   node skt enum ifc  NBP entity\n");

	count = 0;
	nbptyp = nbpCache;

	while (nbptyp != NBPTYPNULL) {
	  nbpobj = nbptyp->list;
	  while (nbpobj != NBPOBJNULL) {
	    sprintf(entity, "%s:%s@%s", nbpobj->nbpobj+1,
	      nbptyp->nbptyp+1, nbpobj->zone+1);
	    fprintf(r, "%3d.%-3d  %3d %3d  %3d %2d%c  %s\n",
	      nbpobj->nbpent[0], nbpobj->nbpent[1],
	      nbpobj->nbpent[2], nbpobj->nbpent[3],
	      nbpobj->nbpent[4], nbpobj->intfc,
	      (nbpobj->when == 0) ? '*' : ' ',
	      entity);
	    count++;
	    nbpobj = nbpobj->next;
	  }
	  nbptyp = nbptyp->next;
	}
	fprintf(r, "\n\n%d NBP entries\n", count);

	(void)fclose(r);

	if (debug & NBPDEBUG)
	  fprintf(stderr, " NBP: dumped NBP cache, %d entries\n", count);

	return;
}

/*
 * pack object, type and zone (if any) into contiguous data
 *
 */

short
nbp_pack(pkt, obj, typ, zon)
register u_char *pkt, *obj, *typ, *zon;
{
	register int i;
	register short len;

	len = 0;

	i = *obj+1;
	bcopy((char *)obj, (char *)pkt, i);
	pkt += i;
	len += i;

	i = *typ+1;
	bcopy((char *)typ, (char *)pkt, i);
	pkt += i;
	len += i;

	if (zon == NULL)
	  return(len);

	i = *zon+1;
	bcopy((char *)zon, (char *)pkt, i);
	pkt += i;
	len += i;

	return(len);
}

/*
 * check (pascal) object, type and zone strings for sanity
 *
 */

int
nbp_invalid(obj, typ, zon, pkt, len)
register u_char *obj, *typ, *zon, *pkt;
register short len;
{
	if (*obj > MAXNBP)
	  return(1);

	if (*typ > MAXNBP)
	  return(1);

	if (*zon > MAXNBP)
	  return(1);

	if (pkt == NULL)
	  return(0);

	if ((zon+*zon+1) > pkt+len)
	  return(1);

	return(0);
}

/*
 * check for NBP obj/typ wildcards in string
 *
 */

int
nbp_wild(str)
register u_char *str;
{
	if (str[0] == nbpEquals && str[1] == 0x00)
	  return(1);

	if (index((char *)str, nbpApprox))
	  return(1);

	return(0);
}

/*
 * NBP object/type match
 * (taking wildcards in "pat" into account)
 *
 */

int
nbp_match(pat, thing)
register u_char *pat, *thing;
{
	register u_char *p;
	register short pl, tl, hl;
	int strcmpci(), strncmpci();

	if ((pat[0] == nbpEquals || pat[0] == nbpApprox) && pat[1] == 0x00)
	  return(1);

	if (p = (u_char *)index((char *)pat, nbpApprox)) {
	  if ((hl = p - pat)
	   && strncmpci(pat, thing, hl) != 0)
	    return(0);

	  pl = strlen((char *)pat) - 1;
	  tl = strlen((char *)thing);

	  if (tl < pl)
	    return(0);

	  if (hl < pl
	   && strcmpci((p+1), thing+(tl-(pl-hl))) != 0)
	    return(0);

	  return(1);
	}

	if (strcmpci(pat, thing) != 0)
	  return(0);

	return(1);
}

/*
 * check for matching zone names
 *
 */

int
nbp_zonmatch(zonea, zoneb)
register u_char *zonea, *zoneb;
{
	int strcmpci();

	if (zonea[0] == 0x00 || (zonea[0] == '*' && zonea[1] == 0x00))
	  return(1); /* match anything */

	if (strcmpci(zonea, zoneb) == 0)
	  return(1);

	return(0);
}

/*
 * copy a pascal str to a C string
 *
 */

void
strcpypc(str, string)
register char *str, *string;
{
	register int len;

	if ((len = *str) > MAXNBP)
	  len = MAXNBP;

	bcopy(str+1, string, len);

	string[len] = '\0';

	return;
}

/*
 * copy a C str to a pascal string
 *
 */

void
strcpycp(str, string)
register char *str, *string;
{
	register int len;

	if ((len = strlen(str)) > MAXNBP)
	  len = MAXNBP;

	bcopy(str, string+1, len);

	string[0] = len;

	return;
}

/*
 * case insensitive version of strcmp()
 *
 */

int
strcmpci(s, t)
register u_char *s, *t;
{
	register u_char c, d;
	u_char mactolower();

	if (debug & NBPDEBUG)
	  fprintf(stderr, " NBP: comparing :%s: to :%s:\n",
	    (char *)s, (char *)t);

	for ( ; ; ) {
	  c = mactolower(*s++);
	  d = mactolower(*t++);
	  if (c != d)
	    return(c-d);
	  if (c == '\0')
	    return(0);
	}
}

/*
 * case insensitive version of strncmp()
 *
 */

int
strncmpci(s, t, n)
register u_char *s, *t;
register int n;
{
	register u_char c, d;
	u_char mactolower();

	for ( ; n > 0 ; n--) {
	  c = mactolower(*s++);
	  d = mactolower(*t++);
	  if (c != d)
	    return(c-d);
	  if (c == '\0')
	    return(0);
	}

	return(0);
}

/*
 * map Macintosh uppercase characters to lowercase
 * (Inside AppleTalk, 2nd Edition, Appendix D)
 *
 */

u_char
mactolower(c)
register u_char c;
{
	if (!(c & 0x80)) {
	  if (isupper((int)c))
	    return(tolower((int)c));
	  else
	    return(c);
	}

	switch (c) {
	  case 0xcb:
	    return(0x88);
	    break;
	  case 0x80:
	    return(0x8a);
	    break;
	  case 0xcc:
	    return(0x8b);
	    break;
	  case 0x81:
	    return(0x8c);
	    break;
	  case 0x82:
	    return(0x8d);
	    break;
	  case 0x83:
	    return(0x8e);
	    break;
	  case 0x84:
	    return(0x96);
	    break;
	  case 0x85:
	    return(0x9a);
	    break;
	  case 0xcd:
	    return(0x9b);
	    break;
	  case 0x86:
	    return(0x9f);
	    break;
	  case 0xae:
	    return(0xbe);
	    break;
	  case 0xaf:
	    return(0xbf);
	    break;
	  case 0xce:
	    return(0xcf);
	    break;
	}
	return(c);
}

/*
 * map Macintosh lowercase characters to uppercase
 * (Inside AppleTalk, 2nd Edition, Appendix D)
 *
 *
 */

u_char
mactoupper(c)
register u_char c;
{
	if (!(c & 0x80)) {
	  if (islower((int)c))
	    return(toupper((int)c));
	  else
	    return(c);
	}

	switch (c) {
	  case 0x88:
	    return(0xcb);
	    break;
	  case 0x8a:
	    return(0x80);
	    break;
	  case 0x8b:
	    return(0xcc);
	    break;
	  case 0x8c:
	    return(0x81);
	    break;
	  case 0x8d:
	    return(0x82);
	    break;
	  case 0x8e:
	    return(0x83);
	    break;
	  case 0x96:
	    return(0x84);
	    break;
	  case 0x9a:
	    return(0x85);
	    break;
	  case 0x9b:
	    return(0xcd);
	    break;
	  case 0x9f:
	    return(0x86);
	    break;
	  case 0xbe:
	    return(0xae);
	    break;
	  case 0xbf:
	    return(0xaf);
	    break;
	  case 0xcf:
	    return(0xce);
	    break;
	}
	return(c);
}
