/*
 * General Purpose AppleTalk Packet Filter Interface
 *
 * Copyright (c) 1992-1995, The University of Melbourne.
 * All Rights Reserved.  Permission to redistribute or
 * use any part of this software for any purpose must
 * be obtained in writing from the copyright owner.
 *
 * This software is supplied "as is" without express
 * or implied warranty.
 *
 * djh@munnari.OZ.AU
 *
 * Supports:
 *	SunOS Network InterFace Tap (SNITPF)
 *	DEC ULTRIX Packet Filter (UPFILT)
 *	DEC OSF/Alpha Packet Filter (UPFILT)
 *	DEC Packet Filter via FDDI (UPFILT_FDDI)
 *	Stanford Ethernet Packet Filter (ENETPF)
 *	Solaris STREAMS Data Link Provider Interface (DLPIPF)
 *	SGI IRIX SNOOP PF for promisc, DRAIN PF for non-promisc
 *	Sony NEWS-OS 4.N rawether (RAWETH) [Phase 1 only for 4.0]
 *	Hewlett Packard HP-UX Link Level Access (LLAPF)
 *	Berkeley Packet Filter (BPFILT)
 *	AIX Ethernet Device (AIXETH)
 *	Linux SOCK_PACKET
 *	
 * $Author: djh $
 * $Revision: 1.20 $
 *
 */


#ifdef	SOLARIS
#define	DLPIPF
#define	MAPFUNCS
#endif	SOLARIS

#ifdef	sun
#ifndef	ENETPF
#ifndef	BPFILT
#ifndef	DLPIPF
#define	SNITPF
#endif	DLPIPF
#endif	BPFILT
#endif	ENETPF
#endif	sun

#ifdef	ultrix
#ifndef	BPFILT
#define	UPFILT
#endif	BPFILT
#endif	ultrix

#ifdef	__alpha
#define	UPFILT
#endif	__alpha

#ifdef	sony_news
#define	RAWETH
#endif	sony_news

#ifdef	AIX
#define	AIXETH
#endif	AIX

#ifdef	BPFILT
#define	NEEDSYSSTROPTS
#endif	BPFILT

#ifdef	__386BSD__
#define	BPFILT
#endif	__386BSD__

#ifdef	__NetBSD__
#define	BPFILT
#endif	__NetBSD__

#ifdef	FreeBSD
#define	BPFILT
#endif	FreeBSD

#ifdef	bsdi
#define	BPFILT
#endif	bsdi

#ifdef	linux
#endif	linux

/*
 * include header files
 *
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/file.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include <net/if.h>


#ifdef	SNITPF
#define	DEV_NIT	"/dev/nit"
#include <net/nit.h>
#include <net/nit_if.h>
#include <net/nit_pf.h>
#include <net/nit_buf.h>
#include <net/packetfilt.h>
#include <stropts.h>
#endif	SNITPF

#ifdef	ENETPF
#include <a.out.h>
#include <stropts.h>
#include <sys/stat.h>
#include <sys/errno.h>
#include <net/if_ieee802.h>
#include <net/enet.h>
#endif	ENETPF

#ifdef	BPFILT
#include <errno.h>
#include <net/bpf.h>
#ifdef	NEEDSYSSTROPTS
#include <sys/stropts.h>
#endif	NEEDSYSSTROPTS
#endif	BPFILT

#ifdef	UPFILT
#include <net/pfilt.h>
#ifdef	UPFILT_FDDI
#include <netinet/if_fddi.h>
#include <net/if_llc.h>
#endif	UPFILT_FDDI
#endif	UPFILT

#ifdef	DLPIPF
#define	MAXDLBUF	1024
#define	OFFADDR(s, n)   (u_char*)((char*)(s) + (int)(n))
#include <sys/fcntl.h>
#include <sys/sockio.h>
#include <sys/signal.h>
#include <sys/bufmod.h>
#include <sys/pfmod.h>
#include <sys/stropts.h>
#include <sys/ethernet.h>
#include <sys/dlpi.h>
#include <sys/pfmod.h>
#endif	DLPIPF

#ifdef	sgi
#include <net/raw.h>
#endif	sgi

#ifdef	hpux
#include <sys/fcntl.h>
#include <netio.h>
#endif	hpux

#ifdef	RAWETH
#include <newsif/if_rawether.h>
#ifdef	sony_phaseII
#define	IF_NAME "en0"
#endif	sony_phaseII
#endif	RAWETH

#ifdef	AIXETH
#include <sys/ciouser.h>
#include <sys/comio.h>
#include <sys/entuser.h>
#include <sys/devinfo.h>
#endif	AIXETH

#ifndef	SOLARIS
#include <netinet/in.h>
#include <netinet/if_ether.h>
#endif	SOLARIS

#ifdef	__NetBSD__
#include <net/if_dl.h>
#include <net/if_types.h>
#endif	__NetBSD__

#ifdef	bsdi
#include <kvm.h>
#include <net/if_dl.h>
#include <net/if_types.h>
#endif	bsdi

#ifdef	linux
#include <sys/errno.h>
#include <netinet/if_ether.h>
#endif	linux

#include <netdb.h>
#include <ctype.h>

/*
 * map compatible functions
 *
 */

#ifdef	MAPFUNCS
#define	bcopy(a,b,l)	memcpy((char *)(b),(char *)(a),(l))
#define	bcmp(a,b,l)	memcmp((char *)(a),(char *)(b),(l))
#define	bzero(a,l)	memset((char *)(a),0,(l))
#define	rindex(s,c)	strrchr((char *)(s),(c))
#define	index(s,c)	strchr((char *)(s),(c))
#endif	MAPFUNCS

/*
 * select common modules
 *
 */

#ifdef	SNITPF
#define	USE_PFILT
#define	USE_GIFADDR
#define	USE_SADDMULTI
#endif	SNITPF

#ifdef	ENETPF
#define	USE_ENET
#define	USE_PFILT
#define	USE_WRITEV
#define	USE_SADDMULTI
#endif	ENETPF

#ifdef	BPFILT
#define	USE_WRITEV
#define	USE_GIFADDR
#define	USE_SADDMULTI
#endif	BPFILT

#ifdef	UPFILT
#define	USE_ENET
#define	USE_PFILT
#define	USE_WRITEV
#define	USE_SADDMULTI
#endif	UPFILT

#ifdef	DLPIPF
#define	USE_PFILT
#define	USE_WRITEV
#endif	DLPIPF

#ifdef	sgi
#define	USE_WRITEV
#define	USE_GIFADDR
#define	USE_SADDMULTI
#endif	sgi

#ifdef	RAWETH
#endif	RAWETH

#ifdef	AIXETH
#endif	AIXETH

#ifdef	__386BSD__
#undef	USE_SADDMULTI
#endif	__386BSD__

#ifdef	__NetBSD__
#define	USE_SIOCGIFCONF
#undef	USE_GIFADDR
#endif	__NetBSD__

#ifdef	FreeBSD
#endif	FreeBSD

#ifdef	bsdi
#define	USE_SIOCGIFCONF
#undef	USE_GIFADDR
#endif	bsdi

#ifdef	linux
#define	USE_SADDMULTI
#endif	linux

/*
 * definitions
 *
 */

#define	READBUFSIZ	4096
#define	NUMRDS		32

struct RDS {
  u_short dataLen;
  u_char *dataPtr;
};

/*
 * variables
 *
 */

#ifdef	hpux
int ethtyp;
#endif	hpux

#ifdef	DLPIPF
struct strbuf dlpictl;
long dlpibuf[MAXDLBUF];
#endif	DLPIPF

#ifdef	BPFILT
u_int pf_bufsize = READBUFSIZ;
#endif	BPFILT

#ifdef	UPFILT
#ifdef	UPFILT_FDDI
int ethtyp[32];
int onfddi[32];
#endif	UPFILT_FDDI
#endif	UPFILT

#ifdef	linux
struct socklist {
  struct sockaddr sa;
} socklist[32 /* ZZ */];
#endif	linux

struct ifreq ifr;
extern int errno;
extern int promisc;

struct RDS RDS[NUMRDS];

/*
 * Open and initialize packet filter
 * for a particular protocol type.
 *
 */

int
init_eth(interface, mode, protocol, typ)
char *interface;
u_short protocol;
int typ, mode;
{
  int s;
  int ioarg;
  char device[64];
  unsigned long if_flags;

#ifdef	SNITPF
  struct strioctl si;

  if ((s = open(DEV_NIT, mode)) < 0) {
    perror(DEV_NIT);
    return(-1);
  }
#endif	SNITPF


#ifdef	ENETPF
  sprintf(device, "/dev/%s", interface);
  if ((s = open(device, mode)) < 0) {
    perror(device);
    return(-1);
  }
#endif	ENETPF


#ifdef	BPFILT
  int devno = 0;

  for (;;) {
    /* loop through, find available device */
    sprintf(device, "/dev/bpf%d", devno++);
    if ((s = open(device, mode)) >= 0)
      break;
    if (s < 0 && errno != EBUSY) {
      perror(device);
      return(-1);
    }
  }
#endif	BPFILT


#ifdef	UPFILT
  if ((s = pfopen(interface, mode)) < 0) {
    perror(interface);
    return(-1);
  }
#ifdef	UPFILT_FDDI
  { struct endevp devParams;

    if (ioctl(s, EIOCDEVP, &devParams) < 0) {
      perror(interface);
      return(-1);
    }
    onfddi[s] = (devParams.end_dev_type == ENDT_FDDI);
    ethtyp[s] = typ;
  }
#endif	UPFILT_FDDI
#endif	UPFILT


#ifdef	DLPIPF
  dl_attach_req_t attach_req;
  dl_bind_req_t bind_req;
  union DL_primitives *dlp;
  struct strioctl si;
  int devno = 0;
  char *p;

  sprintf(device, "/dev/%s", interface);
  for (p = device; *p != '\0'; p++) {
    if (*p >= '0' && *p <= '9') {
      devno = atoi(p);
      *p = '\0';
      break;
    }
  }

  if ((s = open(device, mode)) < 0) {
    perror(device);
    return(-1);
  }

  /* attach file descriptor to interface */
  attach_req.dl_primitive = DL_ATTACH_REQ;
  attach_req.dl_ppa = devno;

  if (doDLPI(s, (char *)&attach_req, sizeof(attach_req),
    sizeof(dl_ok_ack_t)) < 0)
      return(-1);

  dlp = (union DL_primitives *)dlpictl.buf;
  if (dlp->dl_primitive != (u_long)DL_OK_ACK) {
    fprintf(stderr, "DL_ATTACH_REQ: bad ack response\n");
    return(-1);
  }

  /* label file descriptor with ethernet type */
  bind_req.dl_primitive = DL_BIND_REQ;
  bind_req.dl_sap = (typ == 2) ? 16 : protocol;
  bind_req.dl_max_conind = 0;
  bind_req.dl_service_mode = DL_CLDLS;
  bind_req.dl_conn_mgmt = 0;
  bind_req.dl_xidtest_flg = 0;

  if (doDLPI(s, (char *)&bind_req, sizeof(bind_req),
    sizeof(dl_bind_ack_t)) < 0)
      return(-1);

  dlp = (union DL_primitives *)dlpictl.buf;
  if (dlp->dl_primitive != (u_long)DL_BIND_ACK) {
    fprintf(stderr, "DL_BIND_REQ: bad ack response\n");
    return(-1);
  }

  /* set raw mode */
  si.ic_cmd = DLIOCRAW;
  si.ic_timout = -1;
  si.ic_len = 0;
  si.ic_dp = NULL;
  if (ioctl(s, I_STR, &si) < 0) {
    perror("DLIOCRAW");
    return(-1);
  }

  /* set up messages */
  if (ioctl(s, I_SRDOPT, (char *)RMSGD) < 0) {
    perror("I_SRDOPT");
    return(-1);
  }
#endif	DLPIPF


#ifdef	sgi
  if ((s = rawopen(interface, protocol, typ)) < 0) {
    perror(interface);
    return(-1);
  }
#endif	sgi
  

#ifdef	hpux
  sprintf(device, "/dev/%s", interface);
  if ((s = open(device, mode)) < 0) {
    perror(device);
    return(-1);
  }
#endif	hpux


#ifdef	RAWETH
  if ((s = open(interface, O_RDWR)) < 0) {
    perror(interface);
    return(-1);
  }
#endif	RAWETH


#ifdef	AIXETH
  sprintf(device, "/dev/%s", interface);
  if ((s = open(device, O_RDWR)) < 0) {
    perror(interface);
    return(-1);
  }
#endif	AIXETH


#ifdef	linux
  { u_short prot;

    prot = ((typ == 2) ? htons(ETH_P_ALL) : htons(protocol));
    if ((s = socket(AF_INET, SOCK_PACKET, prot)) < 0) {
      perror(interface);
      return(-1);
    }
    /* ZZ */
    if (s >= 32) {
      fprintf(stderr, "OOPS: socket() returned %d\n", s);
      return(-1);
    }
    /* ZZ */
  }
#endif	linux

  /*
   * set filter for protocol and type (IPTalk, Phase 1/2)
   *
   */

  if (setup_pf(s, protocol, typ) < 0)
    return(-1);

  /*
   * set options, bind to underlying interface
   *
   */

  strncpy(ifr.ifr_name, interface, sizeof(ifr.ifr_name));

#ifdef	SNITPF
  si.ic_cmd = NIOCBIND;
  si.ic_timout = 10;
  si.ic_len = sizeof(ifr);
  si.ic_dp = (caddr_t)&ifr;
  if (ioctl(s, I_STR, (caddr_t)&si) < 0) {
    perror("I_STR");
    return(-1);
  }
  if (promisc) {
    if_flags = NI_PROMISC;
    si.ic_cmd = NIOCSFLAGS;
    si.ic_timout = 10;
    si.ic_len = sizeof(if_flags);
    si.ic_dp = (caddr_t)&if_flags;
    if (ioctl(s, I_STR, (caddr_t)&si) < 0) {
      perror("I_STR");
      return(-1);
    }
  }
  if (ioctl(s, I_SRDOPT, (char *)RMSGD) < 0) {
    perror("I_SRDOPT");
    return(-1);
  }
  if (ioctl(s, I_FLUSH, (char *)FLUSHR) < 0) {
    perror("I_FLUSH");
    return(-1);
  }
#endif	SNITPF


#ifdef	ENETPF
#endif	ENETPF


#ifdef	BPFILT
  { struct bpf_version vers;

    /* check the version number */
    if ((ioctl(s, BIOCVERSION, &vers)) < 0) {
      perror("BIOCVERSION");
      return(-1);
    }
    if (vers.bv_major != BPF_MAJOR_VERSION
     || vers.bv_minor < BPF_MINOR_VERSION) {
      fprintf(stderr, "Incompatible BPF version\n");
      return(-1);
    }
#ifdef	__386BSD__
    /* we have to read EXACTLY the buffer size */
    if ((ioctl(s, BIOCSBLEN, &pf_bufsize)) < 0) {
      perror("BIOCSBLEN");
      return(-1);
    }
    if (pf_bufsize != READBUFSIZ) {
      fprintf(stderr, "Can't set exact BPF buffer size\n");
      return(-1);
    }
#else	/* __386BSD__ */
    /* we have to read EXACTLY the buffer size */
    if ((ioctl(s, BIOCGBLEN, &pf_bufsize)) < 0) {
      perror("BIOCGBLEN");
      return(-1);
    }
    if (pf_bufsize > READBUFSIZ) {
      fprintf(stderr, "Large BPF buffers in use (%d bytes)\n", pf_bufsize);
      return(-1);
    }
#endif	/* __386BSD__ */
  }
  /* bind to interface */
  if (ioctl(s, BIOCSETIF, (caddr_t)&ifr) < 0) {
    perror("BIOCSETIF");
    return(-1);
  }
  ioarg = 1;
  if (ioctl(s, BIOCIMMEDIATE, &ioarg) < 0) {
    perror("BIOCIMMEDIATE");
    return(-1);
  }
  if (promisc) {
    if (ioctl(s, BIOCPROMISC, (void *)0) < 0) {
      perror("BIOCPROMISC");
      return(-1);
    }
  }
#endif	BPFILT


#ifdef	UPFILT
  if (promisc) {
    ioarg = ENPROMISC;
    if (ioctl(s, EIOCMBIS, &ioarg) < 0) {
      perror("EIOCMBIS/ENPROMISC");
      return(-1);
    }
  }
#ifdef	ENCOPYALL
  ioarg = ENCOPYALL;
  if (ioctl(s, EIOCMBIS, &ioarg) < 0) {
    perror("EIOCMBIS/ENCOPYALL");
    /* non-fatal */
  }
#endif	ENCOPYALL
#ifdef	ENNONEXCL
  ioarg = ENNONEXCL;
  if (ioctl(s, EIOCMBIS, &ioarg) < 0) {
    perror("EIOCMBIS/ENNONEXCL");
    /* non-fatal */
  }
#endif	ENNONEXCL
#endif	UPFILT


#ifdef	DLPIPF
  if (promisc) {
    dl_promiscon_req_t promiscon_req;
    union DL_primitives *dlp;

    promiscon_req.dl_primitive = DL_PROMISCON_REQ;
    promiscon_req.dl_level = DL_PROMISC_PHYS;

    if (doDLPI(s, (char *)&promiscon_req, sizeof(promiscon_req),
      sizeof(dl_ok_ack_t)) < 0)
        return(-1);

    dlp = (union DL_primitives *)dlpictl.buf;
    if (dlp->dl_primitive != (u_long)DL_OK_ACK) {
      fprintf(stderr, "DL_PROMISCON_REQ: bad ack response\n");
      return(-1);
    }
  }
#endif	DLPIPF


#ifdef	sgi
  if (promisc) {
    ioarg = 1;
    if (ioctl(s, SIOCSNOOPING, &ioarg) < 0) {
      perror("SIOCSNOOPING");
      return(-1);
    }
  }
#endif	sgi


#ifdef	hpux
  { struct fis fis;

    fis.reqtype = LOG_READ_CACHE;
    fis.vtype = INTEGERTYPE;
    fis.value.i = 16;
    ioctl(s, NETCTRL, &fis);

    fis.reqtype = LLA_SIGNAL_MASK;
    fis.vtype = INTEGERTYPE;
    fis.value.i = LLA_NO_SIGNAL;
    ioctl(s, NETCTRL, &fis);
  }
  ethtyp = typ;
#endif	hpux


#ifdef	RAWETH
#endif	RAWETH


#ifdef	AIXETH
#endif	AIXETH


#ifdef	linux
  /* record socket interface name */
  strncpy(socklist[s].sa.sa_data, interface, sizeof(socklist[s].sa.sa_data));
#endif	linux

  return(s);
}

/*
 * establish protocol filter
 *
 */

int
setup_pf(s, prot, typ)
int s, typ;
u_short prot;
{
  int ioarg;
  u_short offset;

#ifdef	SNITPF
  struct packetfilt pf;
  register u_short *fwp = pf.Pf_Filter;
  struct strioctl si;
#endif	SNITPF


#ifdef	DLPIPF
  struct packetfilt pf;
  register u_short *fwp = pf.Pf_Filter;
  struct strioctl si;

  /*
   * we get type 1 "for free" with DLPI
   *
   */
  if (typ == 1)
    return(0);

#endif	DLPIPF


#ifdef	USE_ENET
  struct enfilter pf;
  register u_short *fwp = pf.enf_Filter;

#define	Pf_FilterLen enf_FilterLen
#define	Pf_Priority enf_Priority

  ioarg = 16;
  if (ioctl(s, EIOCSETW, &ioarg) < 0) {
    perror("EIOCSETW");
    return(-1);
  }
#endif	USE_ENET


#ifdef	ENETPF
#ifdef	EIOCETHERT
  if (typ == 2)
    ioarg = htons(EF_8023_TYPE);
  else
    ioarg = htons(prot);
  if (ioctl(s, EIOCETHERT, &ioarg) < 0 && errno != EEXIST) {
    perror("EIOCETHERT");
    return(-1);
  }
#endif	EIOCETHERT
#endif	ENETPF


#ifdef	USE_PFILT
#define	s_offset(structp, element) (&(((structp)0)->element))
  bzero((char *)&pf, sizeof(pf));
  pf.Pf_Priority = 128;
#ifdef	UPFILT_FDDI
  if (onfddi[s]) {
    /*
     * fddi is always 802 format so have
     * if_fddi header then if_llc header
     *
     */
    offset = (sizeof(struct fddi_header) + 
      (int)s_offset(struct llc *, llc_un.type_snap.ether_type))/sizeof(u_short);
  } else
#endif	UPFILT_FDDI
  {
    offset = ((int)s_offset(struct ether_header *, ether_type))/sizeof(u_short);
    if (typ == 2)  /* EtherTalk Phase 2 */
      offset += 4; /* shorts: 2 bytes length + 6 bytes of 802.2 and SNAP */
  }
  *fwp++ = ENF_PUSHWORD + offset;
  *fwp++ = ENF_PUSHLIT | ENF_EQ;
  *fwp++ = htons(prot);
  pf.Pf_FilterLen = 3;
  if (typ == 0) { /* IPTalk */
    *fwp++ = ENF_PUSHWORD + offset + 15;
    *fwp++ = ENF_PUSHLIT | ENF_EQ;
    *fwp++ = htons(0xface);
    pf.Pf_FilterLen += 3;
  }
#endif	USE_PFILT


#ifdef	SNITPF
  si.ic_cmd = NIOCSETF;
  si.ic_timout = 10;
  si.ic_len = sizeof(pf);
  si.ic_dp = (char *)&pf;
  if (ioctl(s, I_PUSH, "pf") < 0) {
    perror("I_PUSH");
    return(-1);
  }
  if (ioctl(s, I_STR, (char *)&si) < 0) {
    perror("I_STR");
    return(-1);
  }
#endif	SNITPF


#ifdef	DLPIPF
  si.ic_cmd = PFIOCSETF;
  si.ic_timout = 10;
  si.ic_len = sizeof(pf);
  si.ic_dp = (char *)&pf;
  if (ioctl(s, I_PUSH, "pfmod") < 0) {
    perror("I_PUSH");
    return(-1);
  }
  if (ioctl(s, I_STR, (char *)&si) < 0) {
    perror("I_STR");
    return(-1);
  }
#endif	DLPIPF


#ifdef	USE_ENET
  if (ioctl(s, EIOCSETF, &pf) < 0) {
    perror("EIOCSETF");
    return(-1);
  }
  if (ioctl(s, EIOCFLUSH, 0) < 0) {
    perror("EIOCFLUSH");
    return(-1);
  }
#endif	USE_ENET


#ifdef	BPFILT
  /*
   * Dave Matthews
   * dcjm@dcs.ed.ac.uk
   * dave@prolingua.co.uk
   *
   */
#define	PROG_SIZE	7
  int ppf = 0;
  struct bpf_program pf;
  struct bpf_insn pf_insns[PROG_SIZE];
  extern int errno;

#define	s_offset(structp, element) (&(((structp)0)->element))

  offset = ((int)s_offset(struct ether_header *, ether_type));

  if (typ == 2) /* EtherTalk Phase 2 */
     offset += 8; /* 2 bytes length + 6 bytes of 802.2 and SNAP */

  pf_insns[ppf].code = BPF_LD+BPF_H+BPF_ABS;
  pf_insns[ppf].jt = 0;
  pf_insns[ppf].jf = 0;
  pf_insns[ppf].k = offset;
  ppf++;

  pf_insns[ppf].code = BPF_JMP+BPF_JEQ+BPF_K;
  pf_insns[ppf].jt = 1;
  pf_insns[ppf].jf = 0;
  pf_insns[ppf].k = prot;
  ppf++;

  pf_insns[ppf].code = BPF_RET+BPF_K;       /* Fail */
  pf_insns[ppf].jt = 0;
  pf_insns[ppf].jf = 0;
  pf_insns[ppf].k = 0;
  ppf++;

  if (typ == 0) { /* IPtalk */
    pf_insns[ppf].code = BPF_LD+BPF_H+BPF_ABS;
    pf_insns[ppf].jt = 0;
    pf_insns[ppf].jf = 0;
    pf_insns[ppf].k = offset+30;
    ppf++;

    pf_insns[ppf].code = BPF_JMP+BPF_JEQ+BPF_K;
    pf_insns[ppf].jt = 1;
    pf_insns[ppf].jf = 0;
    pf_insns[ppf].k = 0xface;
    ppf++;

    pf_insns[ppf].code = BPF_RET+BPF_K;       /* Fail */
    pf_insns[ppf].jt = 0;
    pf_insns[ppf].jf = 0;
    pf_insns[ppf].k = 0;
    ppf++;
  }

  pf_insns[ppf].code = BPF_RET+BPF_K;       /* Success */
  pf_insns[ppf].jt = 0;
  pf_insns[ppf].jf = 0;
  pf_insns[ppf].k = (u_int)-1;
  ppf++;

  pf.bf_len = ppf;
  pf.bf_insns = pf_insns;

  if (ioctl(s, BIOCSETF, &pf) < 0) {
    perror("BIOCSETF");
    return(-1);
  }
#endif	BPFILT


#ifdef	sgi
  struct snoopfilter sf;
  struct ether_header *eh;

  if (promisc) {
    bzero((char *)&sf, sizeof(sf));
    eh = RAW_HDR(sf.sf_mask, struct ether_header);
    if (typ == 2) /* EtherTalk Phase 2 */
      eh += 8;
    eh->ether_type = 0xffff;
    if (typ == 0) { /* IPTalk */
      eh += 30;
      eh->ether_type = 0xffff;
    }
    eh = RAW_HDR(sf.sf_match, struct ether_header);
    if (typ == 2) /* EtherTalk Phase 2 */
      eh += 8;
    eh->ether_type = htons(prot);
    if (typ == 0) { /* IPTalk */
      eh += 30;
      eh->ether_type = htons(0xface);
    }
    if (ioctl(s, SIOCADDSNOOP, &sf) < 0) {
      perror("SIOCADDSNOOP");
      return(-1);
    }
  }
#endif	sgi


#ifdef	hpux
  { struct fis fis;

    if (typ == 2) {
      /* Tor Lillqvist <tml@tik.vtt.fi> */
      static unsigned char elap_snap[5] = { 0x08, 0x00, 0x07, 0x80, 0x9b };
      static unsigned char aarp_snap[5] = { 0x00, 0x00, 0x00, 0x80, 0xf3 };
      fis.reqtype = LOG_SNAP_TYPE;
      fis.vtype = 5;
      bcopy((prot == 0x809b) ? elap_snap : aarp_snap, fis.value.s, 5);
      if (ioctl(s, NETCTRL, &fis) < 0) {
	perror("LOG_SNAP_TYPE");
	return(-1);
      }
    } else {
      fis.reqtype = LOG_TYPE_FIELD;
      fis.vtype = INTEGERTYPE;
      fis.value.i = prot;
      if (ioctl(s, NETCTRL, &fis) < 0) {
	perror("LOG_TYPE_FIELD");
	return(-1);
      }
    }
  }
#endif	hpux


#ifdef	RAWETH
  ioarg = ETHER_ALL;
  if (ioctl(s, ETHERIOCSTYPE, &ioarg) < 0) {
    perror("ETHERIOCSTYPE");
    return(-1);
  }
#endif	RAWETH


#ifdef	AIXETH
   { struct session_blk sblk;

     if (typ == 2) {
       sblk.length = 1;
       sblk.netid = 0xaa;
       if (ioctl(s, CIO_START, &sblk) < 0) {
	 perror("CIO_START");
	 return(-1);
       }
     } else {
       sblk.length = 2;
       sblk.netid = prot;
       if (ioctl(s, CIO_START, &sblk) < 0) {
	 perror("CIO_START");
	 return(-1);
       }
     }
   }
#endif	AIXETH


#ifdef	linux
#endif	linux

  return(0);
}

/*
 * get the interface ethernet address
 *
 */

int
eth_addr(s, interface, addr)
int s;
char *interface;
u_char *addr;
{

#ifdef	USE_GIFADDR
  { struct sockaddr *sa;

    strcpy(ifr.ifr_name, interface);
    ifr.ifr_addr.sa_family = AF_UNSPEC;
    if (ioctl(s, SIOCGIFADDR, &ifr) < 0) {
      perror("SIOCGIFADDR");
      return(-1);
    }
#if (defined(sgi) || defined(__386BSD__) || defined(FreeBSD))
    sa = (struct sockaddr *)&ifr.ifr_data;
#else	/* sgi || __386BSD__ || FreeBSD */
    sa = (struct sockaddr *)ifr.ifr_data;
#endif	/* sgi || __386BSD__ || FreeBSD */
    bcopy((char *)sa->sa_data, (char *)addr, 6);
  }
#endif	USE_GIFADDR


#ifdef	USE_SIOCGIFCONF
  { int len, sock;
    struct ifconf ifconf;
    struct sockaddr_dl *sadl;
    struct ifreq ifrbuf[32], *ifra, *ifrb;

    strcpy(ifr.ifr_name, interface);
    ifconf.ifc_len = sizeof(ifrbuf);
    ifconf.ifc_buf = (caddr_t)ifrbuf;

    if ((sock = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
      perror("SOCK_DGRAM");
      return(-1);
    }
    if (ioctl(sock, SIOCGIFCONF, &ifconf) < 0) {
      perror("SIOCGIFCONF");
      close(sock);
      return(-1);
    }
    close(sock);
    ifra = ifrbuf;
    ifrb = (struct ifreq *)((char *)ifra + ifconf.ifc_len);
    while (ifra < ifrb) {
      if (strcmp(ifr.ifr_name, ifra->ifr_name) == 0) {
	if (ifra->ifr_addr.sa_family == AF_LINK) {
	  sadl = (struct sockaddr_dl *)&ifra->ifr_addr;
	  if (sadl->sdl_type == IFT_ETHER) {
	    bcopy((char *)LLADDR(sadl), (char *)addr, 6);
	    return(0);
	  }
#ifdef	bsdi
	  if (sadl->sdl_alen == 0) /* no addr, try other method */
            return(get_eth_addr(s, interface, addr));
#endif	bsdi
	}
      }
      if ((len = ifra->ifr_addr.sa_len+sizeof(ifra->ifr_name)) < sizeof(*ifra))
	len = sizeof(*ifra);
      ifra = (struct ifreq *)((char *)ifra + len);
    }
  }
  return(-1);
#endif	USE_SIOCGIFCONF


#ifdef	USE_ENET
  { struct endevp endev;

    if (ioctl(s, EIOCDEVP, &endev) < 0) {
      perror("EIOCDEVP");
      return(-1);
    }
    bcopy(endev.end_addr, addr, 6);
  }
#endif	USE_ENET


#ifdef	hpux
  { struct fis fis;

    fis.reqtype = LOCAL_ADDRESS;
    if (ioctl(s, NETSTAT, &fis) < 0) {
      perror("LOCAL_ADDRESS");
      return(-1);
    }
    bcopy(fis.value.s, addr, 6);
  }
#endif	hpux


#ifdef	RAWETH
  if (ioctl(s, ETHERIOCGADDR, addr) < 0) {
    perror("ETHERIOCGADDR");
    return(-1);
  }
#endif	RAWETH


#ifdef	AIXETH
  { struct devinfo info;

    if (ioctl(s, IOCINFO, &info) < 0) {
      perror("IOCINFO");
      return(-1);
    }
    bcopy(info.un.ethernet.haddr, addr, 6);
  }
#endif	AIXETH


#ifdef	DLPIPF
  { dl_phys_addr_req_t phys_addr_req;
    union DL_primitives *dlp;

    phys_addr_req.dl_primitive = DL_PHYS_ADDR_REQ;
    phys_addr_req.dl_addr_type = DL_CURR_PHYS_ADDR;

    if (doDLPI(s, (char *)&phys_addr_req, sizeof(phys_addr_req),
      sizeof(dl_phys_addr_ack_t)) < 0)
        return(-1);

    dlp = (union DL_primitives *)dlpictl.buf;
    if (dlp->dl_primitive != (u_long)DL_PHYS_ADDR_ACK) {
      fprintf(stderr, "DL_PHYS_ADDR_REQ: bad ack response\n");
      return(-1);
    }

    bcopy((char *)OFFADDR(dlp, dlp->physaddr_ack.dl_addr_offset), addr, 6);
  }
#endif	DLPIPF


#ifdef	linux
  strcpy(ifr.ifr_name, interface);
  if (ioctl(s, SIOCGIFHWADDR, &ifr) < 0) {
    perror("SIOCGIFHWADDR");
    return(-1);
  }
#ifdef	OLD_SIOCGIFHWADDR
  memcpy((char *)addr, ifr.ifr_hwaddr.sa_data, 6);
#else	OLD_SIOCGIFHWADDR
  memcpy((char *)addr, ifr.ifr_hwaddr, 6);
#endif	OLD_SIOCGIFHWADDR
#endif	linux

  return(0);
}

/*
 * add a multicast address to the interface
 *
 */

int
eth_addmulti(s, interface, addr)
int s;
char *interface;
u_char *addr;
{
  int sock;

#ifdef	USE_SADDMULTI

  strcpy(ifr.ifr_name, interface);

#ifdef	UPFILT
  /* get the real interface name */
  if (ioctl(s, EIOCIFNAME, &ifr) < 0) {
    perror("EIOCIFNAME");
    return(-1);
  }
#endif	UPFILT


#ifdef	ENETPF
  /* get the real interface name */
  if (get_ifname(&ifr) < 0)
    return(-1);
#endif	ENETPF

  ifr.ifr_addr.sa_family = AF_UNSPEC;
  bcopy((char *)addr, ifr.ifr_addr.sa_data, 6);

  /*
   * open a socket, temporarily, to use for SIOC* ioctls
   *
   */
  if ((sock = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
    perror("socket()");
    return(-1);
  }
  if (ioctl(sock, SIOCADDMULTI, (caddr_t)&ifr) < 0) {
    perror("SIOCADDMULTI");
    close(sock);
    return(-1);
  }
  close(sock);
#endif	USE_SADDMULTI


#ifdef	hpux
  { struct fis fis;

    fis.vtype = 6;
    fis.reqtype = ADD_MULTICAST;
    bcopy(addr, fis.value.s, 6);
    if (ioctl(s, NETCTRL, &fis) < 0) {
      perror("ADD_MULTICAST");
      return(-1);
    }
  }
#endif	hpux


#ifdef	sony_phaseII
  if ((sock = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
    perror("socket()");
    return(-1);
  }
  strcpy(ifr.ifr_name, IF_NAME);
  bcopy(addr, ifr.ifr_addr.sa_data, 6);

  if (ioctl(sock, SIOCSMAR, &ifr) < 0) {
    perror("SIOCSMAR");
    close(sock);
    return(-1);
  }
  close(sock);
#endif	/* sony_phaseII */


#ifdef	AIXETH
  { ent_set_multi_t ent;

    ent.opcode = ENT_ADD;
    bcopy(addr, ent.multi_addr, 6);
    if (ioctl(s, ENT_SET_MULTI, &ent) < 0) {
      perror("ENT_SET_MULTI");
      return(-1);
    }
  }
#endif	AIXETH


#ifdef	DLPIPF
  { union DL_primitives *dlp;

    dlp = (union DL_primitives *)dlpibuf;

    dlp->enabmulti_req.dl_primitive = DL_ENABMULTI_REQ;
    dlp->enabmulti_req.dl_addr_length = 6;
    dlp->enabmulti_req.dl_addr_offset = sizeof(dl_enabmulti_req_t);

    bcopy(addr, (char *)OFFADDR(dlpibuf, sizeof(dl_enabmulti_req_t)), 6);

    if (doDLPI(s, (char *)dlpibuf, sizeof(dl_enabmulti_req_t)+6,
      sizeof(dl_ok_ack_t)) < 0)
        return(-1);

    dlp = (union DL_primitives *)dlpictl.buf;
    if (dlp->dl_primitive != (u_long)DL_OK_ACK) {
      fprintf(stderr, "DL_ENABMULTI_REQ: bad ack response\n");
      return(-1);
    }
  }
#endif	DLPIPF

  return(0);
}

/*
 * delete a multicast address from the interface
 *
 */

int
eth_delmulti(s, interface, addr)
int s;
char *interface;
u_char *addr;
{
  int sock;

#ifdef	USE_SADDMULTI

  strcpy(ifr.ifr_name, interface);

#ifdef	UPFILT
  /* get the real interface name */
  if (ioctl(s, EIOCIFNAME, &ifr) < 0) {
    perror("EIOCIFNAME");
    return(-1);
  }
#endif	UPFILT


#ifdef	ENETPF
  /* get the real interface name */
  if (get_ifname(&ifr) < 0)
    return(-1);
#endif	ENETPF

  ifr.ifr_addr.sa_family = AF_UNSPEC;
  bcopy((char *)addr, ifr.ifr_addr.sa_data, 6);

  /*
   * open a socket, temporarily, to use for SIOC* ioctls
   *
   */
  if ((sock = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
    perror("socket()");
    return(-1);
  }
  if (ioctl(sock, SIOCDELMULTI, (caddr_t)&ifr) < 0) {
    perror("SIOCDELMULTI");
    close(sock);
    return(-1);
  }
  close(sock);
#endif	USE_SADDMULTI


#ifdef	hpux
  { struct fis fis;

    fis.vtype = 6;
    fis.reqtype = DELETE_MULTICAST;
    bcopy(addr, fis.value.s, 6);
    if (ioctl(s, NETCTRL, &fis) < 0) {
      perror("DELETE_MULTICAST");
      return(-1);
    }
  }
#endif	hpux


#ifdef	sony_phaseII
  if ((sock = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
    perror("socket()");
    return(-1);
  }
  strcpy(ifr.ifr_name, IF_NAME);
  bcopy(addr, ifr.ifr_addr.sa_data, 6);

  if (ioctl(sock, SIOCSMAR, &ifr) < 0) {		/* ZZ */
    perror("SIOCSMAR");					/* ZZ */
    close(sock);
    return(-1);
  }
  close(sock);
#endif	/* sony_phaseII */


#ifdef	AIXETH
  { ent_set_multi_t ent;

    ent.opcode = ENT_DEL;
    bcopy(addr, ent.multi_addr, 6);
    if (ioctl(s, ENT_SET_MULTI, &ent) < 0) {
      perror("ENT_SET_MULTI");
      return(-1);
    }
  }
#endif	AIXETH


#ifdef	DLPIPF
  { union DL_primitives *dlp;

    dlp = (union DL_primitives *)dlpibuf;

    dlp->disabmulti_req.dl_primitive = DL_DISABMULTI_REQ;
    dlp->disabmulti_req.dl_addr_length = 6;
    dlp->disabmulti_req.dl_addr_offset = sizeof(dl_disabmulti_req_t);

    bcopy(addr, (char *)OFFADDR(dlpibuf, sizeof(dl_disabmulti_req_t)), 6);

    if (doDLPI(s, (char *)dlpibuf, sizeof(dl_disabmulti_req_t)+6,
      sizeof(dl_ok_ack_t)) < 0)
        return(-1);

    dlp = (union DL_primitives *)dlpictl.buf;
    if (dlp->dl_primitive != (u_long)DL_OK_ACK) {
      fprintf(stderr, "DL_DISABMULTI_REQ: bad ack response\n");
      return(-1);
    }
  }
#endif	DLPIPF

  return(0);
}

/*
 * return 1 if ethernet interface capable of multiple opens
 *
 */

int
eth_mopen(phase)
int phase;
{
#ifdef	sony_news
  return(0);
#endif	sony_news


#ifdef	AIXETH
  if (phase == 2)
    return(0);
#endif	AIXETH


#ifdef	sgi
  if (phase == 2)
    return(0);
#endif	sgi


#ifdef	linux
  if (phase == 2)
    return(0);
#endif	linux

  return(1);
}

/*
 * read a packet
 * Read Data Structure describes packet(s) received
 *
 */

int
read_eth(fd, buf, len)
int fd, len;
u_char *buf;
{
  int i, cc;

#ifdef	BPFILT
  u_char *p, *q;
  struct bpf_hdr *bp;

  RDS[0].dataLen = 0;

  /* must read EXACTLY the buffer size */
  if ((cc = read(fd, buf, pf_bufsize)) <= 0) {
#if defined(sun) && !defined(SOLARIS)
    if (errno == EINVAL
     && (long)(tell(fd)+pf_bufsize) < 0) {
      (void)lseek(fd, 0, 0);
      return(0);
    }
#endif	/* SunOS */
    return(cc);
  }

  /*
   * fill in the RDS
   *
   */
  p = buf;
  q = buf+cc;
  for (i = 0; i < (NUMRDS-1) && p < q; i++) {
    bp = (struct bpf_hdr *)p;
    RDS[i].dataLen = bp->bh_caplen;
    RDS[i].dataPtr = p + bp->bh_hdrlen;
    p += BPF_WORDALIGN(bp->bh_hdrlen+bp->bh_caplen);
  }
  RDS[i].dataLen = 0;

  return(RDS[0].dataLen);
#endif	BPFILT


#ifdef	sgi
  int off;
  u_char *p;
  struct snoopheader *sh;

  RDS[0].dataLen = 0;

  if (promisc)
    off = sizeof(struct snoopheader)+RAW_HDRPAD(sizeof(struct ether_header));
  else
    off = RAW_HDRPAD(sizeof(struct ether_header));

  if ((cc = read(fd, buf, len)) <= 0)
    return(cc);

  if ((cc = cc - off) <= 0)
    return(cc);

  if (cc > len)
    cc = len;

  RDS[0].dataLen = cc;
  RDS[0].dataPtr = buf+off;
  RDS[1].dataLen = 0;

  return(cc);
#endif	sgi


#ifdef	hpux
  int off;
  struct fis fis;

  RDS[0].dataLen = 0;

  off = (ethtyp == 1) ? 14 : 22;

  if ((cc = read(fd, buf+off, len-off)) <= 0)
    return(cc);

  fis.reqtype = FRAME_HEADER;
  if (ioctl(fd, NETSTAT, &fis) < 0) {
    perror("FRAME_HEADER");
    return(-1);
  }
  if (fis.vtype != off)
    return(-1);

  bcopy(fis.value.s, (char *)buf, off);

  RDS[0].dataLen = cc+off;
  RDS[0].dataPtr = buf;
  RDS[1].dataLen = 0;

  return(cc+off);
#endif	hpux


#ifdef	linux
  int fromlen;
  struct sockaddr sa;
  /*
   * this is slightly less than optimal due to Linux SOCK_PACKET limitations
   *
   */
  RDS[0].dataLen = 0;
  fromlen = sizeof(struct sockaddr);

#ifdef	USE_MSG_PEEK
  if ((cc = recvfrom(fd, (char *)buf, 22, MSG_PEEK, &sa, &fromlen)) <= 0)
    return(cc);
#else	USE_MSG_PEEK
  if ((cc = recvfrom(fd, (char *)buf, len, 0, &sa, &fromlen)) <= 0)
    return(cc);
#endif	USE_MSG_PEEK

  /* check if right interface */
  if (strcmp(sa.sa_data, socklist[fd].sa.sa_data))
    return(0);

#ifdef	USE_MSG_PEEK
  if ((cc = recvfrom(fd, (char *)buf, len, 0, NULL, NULL)) <= 0)
    return(cc);
#endif	USE_MSG_PEEK

  RDS[0].dataLen = cc;
  RDS[0].dataPtr = buf;
  RDS[1].dataLen = 0;

  return(cc);
#endif	linux


#ifdef	UPFILT
#ifdef	UPFILT_FDDI
  if (onfddi[fd]) {
    struct fddi_header fddibit;
    struct iovec fdiov[3];
    struct llc llcbit;
    int off;                                              

    RDS[0].dataLen = 0;

    /*
     * use readv to get the data we want.  Based on
     * protocol type we need to preserve a length and
     * sap data in the input buffer
     *
     */
    off = (ethtyp[fd] == 2) ? 22 : 14;
    fdiov[0].iov_base = (caddr_t)&fddibit;
    fdiov[0].iov_len = sizeof(struct fddi_header);
    fdiov[1].iov_base = (caddr_t)&llcbit;
    fdiov[1].iov_len = sizeof(struct llc);
    fdiov[2].iov_base = (caddr_t)(buf+off);
    fdiov[2].iov_len = len-off;

    if ((cc = readv(fd, fdiov, 3)) <= 0)
      return(cc);

    bcopy((char *)fddibit.fddi_dhost, (char *)buf, 12);	/* addresses */
    cc = cc - fdiov[0].iov_len - fdiov[1].iov_len + off;/* eth length */

    if (ethtyp[fd] == 2) {
      bcopy((char *)&llcbit, (char *)(buf+14), 8);	/* SAP */
      buf[12] = (cc >> 8) & 0xff;			/* len */
      buf[13] = (cc & 0xff);
    } else {						/* type */
      buf[12] = (llcbit.llc_un.type_snap.ether_type >> 8) & 0xff;
      buf[13] = (llcbit.llc_un.type_snap.ether_type & 0xff);
    }
   
    RDS[0].dataLen = cc;
    RDS[0].dataPtr = buf;
    RDS[1].dataLen = 0;

    return(cc);
  }
#endif	UPFILT_FDDI
#endif	UPFILT


#if (!(defined(BPFILT) || defined(sgi) || defined(hpux) || defined(linux)))
  RDS[0].dataLen = 0;

  if ((cc = read(fd, (char *)buf, len)) <= 0)
    return(cc);

  if (cc > len)
    cc = len;

  RDS[0].dataLen = cc;
  RDS[0].dataPtr = buf;
  RDS[1].dataLen = 0;

  return(cc);
#endif	/* BPFILT || sgi || hpux */

}

/*
 * write a packet
 *
 */

int
write_eth(fd, buf, len)
int fd, len;
u_char *buf;
{

#ifdef	SNITPF
  struct sockaddr sa;
  struct strbuf pbuf, dbuf;

  sa.sa_family = AF_UNSPEC;
  bcopy((char *)buf, (char *)sa.sa_data, sizeof(sa.sa_data));
  pbuf.len = sizeof(struct sockaddr);
  pbuf.buf = (char *) &sa;
  dbuf.len = len-14;
  dbuf.buf = (char *)buf+14;
  if (putmsg(fd, &pbuf, &dbuf, 0) == 0)
    return(len);
#endif	SNITPF


#ifdef	USE_WRITEV
  struct iovec iov[2];

#ifdef	FreeBSD_SWAP_BUG
  u_char swap;

  swap = buf[12];
  buf[12] = buf[13];
  buf[13] = swap;
#endif	FreeBSD_SWAP_BUG


#ifdef	UPFILT
#ifdef	UPFILT_FDDI
  if (onfddi[fd]) {
    int off, tot;
    struct llc llcbit;
    struct iovec fdiov[3];
    struct fddi_header fddibit;
    unsigned char sapSnap[6] = { LLC_SNAP_LSAP, LLC_SNAP_LSAP,
                                 LLC_UI, '\0', '\0', '\0' };

    fddibit.fddi_fc = FDDIFC_LLC_ASYNC;			/* frame control */
    bcopy(buf, fddibit.fddi_dhost, 12);			/* mac addresses */

    if (ethtyp[fd] == 2) {
      bcopy((char *)(buf+14), (char *)&llcbit, 8);	/* SAP */
      off = 22;
    } else {
      bcopy(sapSnap, (char *)&llcbit.llc_dsap, 6);	/* sap snap */
      llcbit.llc_un.type_snap.ether_type = (buf[12] << 8) | buf[13];
      off = 14;
    }

    fdiov[0].iov_base = (caddr_t)&fddibit;
    fdiov[0].iov_len = sizeof(struct fddi_header);
    fdiov[1].iov_base = (caddr_t)&llcbit;
    fdiov[1].iov_len = sizeof(struct llc);
    fdiov[2].iov_base = (caddr_t)(buf+off);
    fdiov[2].iov_len = len-off;

    tot = len-off+sizeof(struct fddi_header)+sizeof(struct llc);

    if (writev(fd, fdiov, 3) == tot)
      return(len);

  }
#endif	UPFILT_FDDI
#endif	UPFILT

  iov[0].iov_base = (caddr_t)buf;
  iov[0].iov_len = 14;
  iov[1].iov_base = (caddr_t)buf+14;
  iov[1].iov_len = len-14;

  if (writev(fd, iov, 2) == len)
    return(len);

#endif	USE_WRITEV


#ifdef	hpux
  struct fis fis;
  int off;

  off = (ethtyp == 1) ? 14 : 22;
  fis.reqtype = LOG_DEST_ADDR;
  fis.vtype = 6;
  bcopy(buf, fis.value.s, 6);
  if (ioctl(fd, NETCTRL, &fis) < 0)
    return(-1);
  if (write(fd, buf+off, len-off) == len-off)
    return(len);
#endif	hpux


#ifdef	RAWETH
  int length;

  length = (len < 60) ? 60 : len;
  if (write(fd, buf, length) == length)
    return(len);
#endif	RAWETH


#ifdef	AIXETH
  int length;

  length = (len < 60) ? 60 : len;
  if (write(fd, buf, length) == length)
    return(len);
#endif	AIXETH


#ifdef	linux
  if (sendto(fd, buf, len, 0, &socklist[fd].sa, sizeof(struct sockaddr)) == len)
    return(len);
#endif	linux

  return(-1);
}


#ifdef	sgi
/*
 * NB: in non-promiscuous mode using DRAIN, we can't see typ==0 (IPTalk)
 *
 */
#define	IEEE8023 1

int
rawopen(interface, protocol, typ)
char *interface;
u_short protocol;
int typ;
{
  int s;
  int sktsiz;
  int rawproto;
  struct drainmap map;
  struct sockaddr_raw sr;

  rawproto = (promisc) ? RAWPROTO_SNOOP : RAWPROTO_DRAIN;
  if ((s = socket(PF_RAW, SOCK_RAW, rawproto)) < 0) {
    perror("socket()");
    return(-1);
  }
  sr.sr_family = AF_RAW;
  if (promisc)
    sr.sr_port = 0;
  else
    sr.sr_port = (typ == 2) ? IEEE8023 : htons(protocol);
  strncpy(sr.sr_ifname, interface, sizeof(sr.sr_ifname));
  if (bind(s, &sr, sizeof(sr)) < 0) {
    perror("bind()");
    return(-1);
  }
  if (rawproto == RAWPROTO_DRAIN) {
    if (typ == 2) {
      map.dm_minport = 10;
      map.dm_maxport = 1498;
      map.dm_toport = IEEE8023;
      if (ioctl(s, SIOCDRAINMAP, &map) < 0) {
        perror("SIOCDRAINMAP");
        return(-1);
      }
    }
  }
  sktsiz = 60000;
  if (setsockopt(s,SOL_SOCKET,SO_RCVBUF,(char *)&sktsiz,sizeof(sktsiz)) < 0) {
    perror("SO_RCVBUF");
    return(-1);
  }
  return(s);
}
#endif	sgi


#ifdef	ENETPF
/*
 * ferret around in /dev/kmem for real interface details
 * (sometime it would be nice to have an ioctl to do this)
 *
 * originally part of enet/etherstat.c
 *
 */

struct enet_info {
	struct ifnet *ifp;
};

int
get_ifname(ifr)
struct ifreq *ifr;
{
  int kmem;
  int dataLen;
  int enUnits;
  int minorDev;
  char *kernelfile;
  char namebuf[64];
  char devName[64];
  struct stat sbuf;
  struct nlist nl[3];
  struct ifnet ifnet;
  struct enet_info *enet_info;

  sprintf(devName, "/dev/%s", ifr->ifr_name);

  kernelfile = "/vmunix";
  if (stat(kernelfile, &sbuf) != 0) {
    perror("stat /vmunix");
    return(-1);
  }
  if ((kmem = open("/dev/kmem", 0)) < 0) {
    perror("open /dev/kmem");
    return(-1);
  }

  nl[0].n_un.n_name = "_enUnits";
  nl[1].n_un.n_name = "_enet_info";
  nl[2].n_un.n_name = "";

  nlist(kernelfile,nl);

  if (nl[0].n_type == 0 || nl[1].n_type == 0) {
    fprintf(stderr, "get_ifname() can't find _enUnits or _enet_info\n");
    close(kmem);
    return(-1);
  }

  (void) lseek(kmem, (nl[0].n_value), 0);
  if (read(kmem, &enUnits, sizeof(enUnits)) != sizeof(enUnits)) {
    perror("read /dev/kmem #1");
    close(kmem);
    return(-1);
  }
  dataLen = enUnits*sizeof(struct enet_info);
  if ((enet_info = (struct enet_info *) malloc(dataLen)) == 0) {
    fprintf(stderr, "get_ifname() can't malloc enet_info\n");
    close(kmem);
    return(-1);
  }
  (void) lseek(kmem, (nl[1].n_value), 0);
  if (read(kmem, enet_info, dataLen) != dataLen) {
    perror("read /dev/kmem #2");
    close(kmem);
    return(-1);
  }

  if (stat(devName, &sbuf) != 0) {
    perror(devName);
    close(kmem);
    return(-1);
  }
  if (!S_ISCHR(sbuf.st_mode)) {
    fprintf(stderr, "%s: not character special device!\n", devName);
    close(kmem);
    return(-1);
  }
  minorDev = sbuf.st_rdev & 0xff;	/* minor device number */

  if (minorDev >= enUnits) {
    fprintf(stderr,"%s: invalid minor device number %d\n", devName, minorDev);
    close(kmem);
    return(-1);
  }
  (void) lseek(kmem, enet_info[minorDev].ifp, 0);
  if (read(kmem, &ifnet, sizeof(ifnet)) != sizeof(ifnet)) {
    perror("read /dev/kmem #3");
    close(kmem);
    return(-1);
  }
  (void) lseek(kmem, ifnet.if_name, 0);
  if (read(kmem, namebuf, sizeof(namebuf)) != sizeof(namebuf)) {
    perror("read /dev/kmem #4");
    close(kmem);
    return(-1);
  }
  sprintf(ifr->ifr_name, "%s%d", namebuf, ifnet.if_unit);
  fprintf(stderr, "ENET %s maps to interface %s", devName, ifr->ifr_name);

  free(enet_info);
  close(kmem);
  return(0);
}
#endif	ENETPF


#ifdef	DLPIPF
/*
 * send message to DLPI STREAMS module and wait for a response
 *
 */

static void
sigalrm()
{
  fprintf(stderr, "sigalrm(): timeout waiting for DLPI response\n");
  exit(1);
}

int
doDLPI(s, req, reqlen, acklen)
int s;
char *req;
int reqlen, acklen;
{
  int flags, rc;

  dlpictl.maxlen = 0;
  dlpictl.len = reqlen;
  dlpictl.buf = req;

  flags = 0;

  if (putmsg(s, &dlpictl, (struct strbuf *)NULL, flags) < 0) {
    perror("putmsg()");
    return(-1);
  }

  dlpictl.maxlen = MAXDLBUF;
  dlpictl.buf = (char *)dlpibuf;
  dlpictl.len = 0;

  flags = 0;

  (void)signal(SIGALRM, sigalrm);

  alarm(10);

  if ((rc = getmsg(s, &dlpictl, (struct strbuf *)NULL, &flags)) < 0) {
    perror("getmsg()");
    exit(1);
  }

  alarm(0);

  if (rc & (MORECTL | MOREDATA)) {
    fprintf(stderr, "doDLPI(): extra CTL/DATA waiting for retrieval\n");
    return(-1);
  }

  if (flags != RS_HIPRI) {
    fprintf(stderr, "doDLPI(): expecting RS_HIPRI response\n");
    return(-1);
  }

  if (dlpictl.len < acklen) {
    fprintf(stderr, "doDLPI(): short response\n");
    return(-1);
  }

  return(0);
}
#endif	DLPIPF

#ifdef	bsdi
struct nlist nlst[] = {
  { "_ifnet" },
  "",
};

/*
 * BSDI
 *
 * Get interface address from the kernel since the SIOCGIFADDR
 * ioctl isn't implemented in a number of ethernet drivers.
 *
 */

int
get_eth_addr(s, interface, addr)
int s;
char *interface;
u_char *addr;
{
  kvm_t *kvmd;
  int err = -1;
  off_t ifnetptr;
  off_t aifnetptr;
  off_t ifaddrptr;
  struct sockaddr sa;
  struct ifnet ifnet;
  struct ifaddr ifaddr;
  struct arpcom arpcom;
  struct sockaddr_dl *sdl;
  char *cp, if_name[16];

  if ((kvmd = kvm_open(NULL, NULL, NULL, O_RDONLY, NULL)) == NULL) {
    fprintf(stderr, "kvm_open: can't open kernel!\n");
    return(-1);
  }
  if (kvm_nlist(kvmd, nlst) < 0 || nlst[0].n_type == 0) {
    fprintf(stderr, "kvm_list: can't find namelist!\n");
    kvm_close(kvmd);
    return(-1);
  }
  if (nlst[0].n_value == 0) {
    fprintf(stderr, "kvm_list: _ifnet symbol not defined!\n");
    kvm_close(kvmd);
    return(-1);
  }
  if (kvm_read(kvmd, nlst[0].n_value, (char *)&ifnetptr,
   sizeof(ifnetptr)) != sizeof(ifnetptr)) {
    fprintf(stderr, "kvm_read: bogus read!\n");
    kvm_close(kvmd);
    return(-1);
  }
  while (ifnetptr) {
    aifnetptr = ifnetptr;
    if (kvm_read(kvmd, ifnetptr, (char *)&ifnet,
     sizeof(ifnet)) == sizeof(ifnet)) {
      if (kvm_read(kvmd, (off_t)ifnet.if_name, if_name,
       sizeof(if_name)) == sizeof(if_name)) {
        ifnetptr = (off_t)ifnet.if_next;
	if_name[15] = '\0';
        cp = (char *)index(if_name, '\0');
	sprintf(cp, "%d", ifnet.if_unit);
	if (strcmp(if_name, interface) != 0)
	  continue;
	ifaddrptr = (off_t)ifnet.if_addrlist;
	while (ifaddrptr) {
	  if (kvm_read(kvmd, ifaddrptr, (char *)&ifaddr,
	   sizeof(ifaddr)) == sizeof(ifaddr)) {
	    ifaddrptr = (off_t)ifaddr.ifa_next;
	    if (kvm_read(kvmd, (off_t)ifaddr.ifa_addr, (char *)&sa,
	     sizeof(sa)) == sizeof(sa)) {
	      if (sa.sa_family == AF_LINK) {
		sdl = (struct sockaddr_dl *)&sa;
		cp = (char *)LLADDR(sdl);
		if (sdl->sdl_alen == 0) {
		/* no address here, try ethernet driver */
		  if (kvm_read(kvmd, aifnetptr, (char *)&arpcom,
		   sizeof(arpcom)) == sizeof(arpcom))
		    cp = (char *)arpcom.ac_enaddr;
		}
		bcopy(cp, (char *)addr, 6);
	        fprintf(stderr, "BSDI: %-5s eth address %x:%x:%x:%x:%x:%x\n",
		  interface, addr[0], addr[1], addr[2],
		  addr[3], addr[4], addr[5]);
		err = 0;
		break;
	      }
	      continue;
	    }
	  }
	  break;
	}
      }
    }
    break;
  }
  kvm_close(kvmd);
  return(err);
}
#endif	bsdi
