/*
 * UAR - UNIX AppleTalk Router
 * 
 * Copyright (c) 1993, The University of Melbourne.
 *
 * All Rights Reserved.
 *
 * UAR may NOT be publicly redistributed (for example via
 * anonymous FTP), sold, or the source used for any other
 * purpose without the permission of the copyright owner.
 * 
 * This software is supplied "as is" without express or
 * implied warranty.
 * 
 * djh@munnari.OZ.AU
 *
 * $Author: djh $
 * $Revision: 1.1.1.5 $
 *
 */

#include "uar.h"

extern int debug;
extern int numintfc;
extern u_long uar_stats[];
extern u_long pkts_dropped;
extern struct iflist iflist[];
extern struct timeval tn;
extern char *stats_msg[];

time_t rtmpModTime;
int rtmp_do_dump = 0;

struct rtmp *rtmpMRU;
struct rtmp *rtmpTab[RTMPTABSIZ];

/*
 * Initialize RTMP data structures
 *
 */

void
rtmp_initialize()
{
	int i;

	rtmpMRU = RTMPNULL;

	for (i = 0; i < RTMPTABSIZ; i++)
	  rtmpTab[i] = RTMPNULL;

	rtmpModTime = tn.tv_sec;

	return;
}

/*
 * seed an interface with network number and zone name
 *
 */

void
rtmp_seed(ifn, net_lo, net_hi, znam)
int ifn;
u_short net_lo, net_hi;
u_char *znam;
{
	short zlen;
	struct zone *zone_new();
	struct zone *zone, *zone_find();
	struct rtmp *rtmp, *rtmp_find();
	void rtmp_intfc(), zone_insert();
	void zone_rtmp_add();

	if (debug & RTMPDEBUG) {
	  fprintf(stderr, "RTMP: %-5s SEED  %3d.%-3d-%3d.%-3d",
	    iflist[ifn].if_name, net_lo >> 8, net_lo & 0xff,
	    net_hi >> 8, net_hi & 0xff);
	  fprintf(stderr, " \"%s\"\n", znam);
	}

	rtmp_intfc(ifn, net_lo, net_hi, (struct pd *)NULL);

	zlen = strlen((char *)znam);
	if ((zone = zone_find(znam, zlen)) == ZONENULL) {
	  if ((zone = zone_new()) != ZONENULL) {
	    bcopy((char *)znam, (char *)(zone->name+1), (int)zlen);
	    zone->name[zlen+1] = '\0';
	    zone->name[0] = zlen;
	    zone->nets = NETNULL;
	    zone->next = ZONENULL;
	    zone_insert(zone);
	  }
	}

	if (zone == ZONENULL)
	  drop(uarMemoryError);

	if ((rtmp = rtmp_find(net_lo, net_hi)) == RTMPNULL)
	  drop(uarRTMPCacheFail);

	zone_rtmp_add(zone, rtmp);

	if (iflist[ifn].phase == PHASE2)
	  iflist[ifn].defzone = zone;

	return;
}

/*
 * Received RTMP packet or GetNetInfo Reply
 * check if interface has been initialized yet
 *
 */

void
rtmp_intfc(ifn, net_lo, net_hi, pd)
int ifn;
u_short net_lo, net_hi;
struct pd *pd;
{
	int new;
	char message[80];
	struct rtmp *rtmp, *rtmp_find(), *rtmp_new();
	void cleanup(), rtmp_insert(), log();
	extern char hostname[];

	if ((iflist[ifn].net_lo == 0 && iflist[ifn].net_hi == 0)
	 || (iflist[ifn].net_lo == 0 && iflist[ifn].net_hi == P2ENDSTUP)) {
	  iflist[ifn].net = net_lo;
	  iflist[ifn].net_lo = net_lo;
	  iflist[ifn].net_hi = net_hi;
	  /*
	   * need to re-probe for new address
	   * and restart interface multinodes
	   *
	   */
	  iflist[ifn].state = ADDR_PENDING;
	  iflist[ifn].count = AARPCount;

	  if (new = ((rtmp = rtmp_find(net_lo, net_hi)) == RTMPNULL))
	    if ((rtmp = rtmp_new()) == RTMPNULL)
	      return;

	  rtmp->net_lo = net_lo;
	  rtmp->net_hi = net_hi;
	  rtmp->dist = 0;
	  rtmp->intfc = ifn;
	  rtmp->rnet = 0;
	  rtmp->rnode = 0;
	  rtmp->state = RTMP_GOOD;
	  rtmp->zquantity = 0;
	  rtmp->zexpected = 1;
	  rtmp->zitlist = ZITNULL;

	  if (iflist[ifn].phase == PHASE2)
	    rtmp->flags |= RTMP_EXTENDED;

	  rtmpModTime = tn.tv_sec;

	  if (new) {
	    rtmp_insert(rtmp);
	    if (debug & RTMPDEBUG)
	      fprintf(stderr, "RTMP: %-5s INIT  %3d.%-3d-%3d.%-3d\n",
	        iflist[ifn].if_name,
	        rtmp->net_lo >> 8, rtmp->net_lo & 0xff,
	        rtmp->net_hi >> 8, rtmp->net_hi & 0xff);
	  }

	  return;
	}

	/*
	 * check net range against current config
	 *
	 */
	if ((net_lo != iflist[ifn].net_lo)
	 || (net_hi != iflist[ifn].net_hi)) {
	  log("PANIC: network range configuration mismatch, giving up.");
	  if (pd != NULL) {
	    sprintf(message,
	      "%s: Netwk %3d.%-3d, Rtr Node %d, Net Range %3d.%-3d-%3d.%-3d",
	      iflist[ifn].if_name, pd->srcNet >> 8, pd->srcNet & 0xff,
	      pd->srcNode,net_lo>>8,net_lo&0xff,net_hi>>8,net_hi&0xff);
	    log(message);
	  }
	  cleanup();
	}

	return;
}

/*
 * Incoming RTMP data packet
 *
 */

void
rtmpd(ifn, pd)
int ifn;
struct pd *pd;
{
	u_char *pkt;
	u_char dist;
	u_char rnode;
	u_char *data;
	u_short rnet;
	u_short net_lo;
	u_short net_hi;
	u_short rnet_lo;
	u_short rnet_hi;
	short tuples = 0;
	short len, tuplelen = 3;
	struct rtmp *rtmp, *rtmp_new();
	struct rtmp *rtmp_find(), *rtmp_delete();
	void rtmp_insert(), rtmp_intfc();

	pkt = pd->data;
	len = pd->dataLen;

	if (len == 0)
	  drop(ddpTooShortErrors);
	  
	if (len > MAXDDPLEN)
	  drop(ddpTooLongErrors);

	if (pd->dstSkt != RTMP)
	  drop(uarNoSocketHandler);

	if (pkt[2] != IDLEN)
	  drop(rtmpInErrors);

	rnet = rnet_lo = rnet_hi = (pkt[0] << 8) | pkt[1];
	rnode = pkt[3];

	data = pkt+4;

	if (iflist[ifn].phase == PHASE1) {
	  if (data[0] == 0x00 && data[1] == 0x00) {
	    if (data[2] != 0x82)
	      drop(rtmpInVersionMismatches);
	    data += 3;
	  }
	}

	if (iflist[ifn].phase == PHASE2) {
	  if (data[2] & 0x80) {
	    if (data[5] != 0x82)
	      drop(rtmpInVersionMismatches);
	    rnet_lo = (data[0] << 8) | data[1];
	    rnet_hi = (data[3] << 8) | data[4];
	    data += 6;
	  }
	}

	rtmp_intfc(ifn, rnet_lo, rnet_hi, pd);

	if (debug & RTMPDEBUG)
	  fprintf(stderr, "RTMP: %-5s DATA  %3d.%-3d   %3d   ",
	      iflist[ifn].if_name, rnet >> 8, rnet & 0xff, rnode);

	/*
	 * process RTMP tuples in packet
	 *
	 */
	while (data < (pkt+len)) {
	  net_lo = (data[0] << 8) | data[1];
	  dist = data[2] & 0x1f;
	  if (data[2] & 0x80) {
	  /* extended tuples */
	    net_hi = (data[3] << 8) | data[4];
	    tuplelen = 6;
	  } else {
	  /* non-extended */
	    net_hi = net_lo;
	    tuplelen = 3;
	  }

	  if ((rtmp = rtmp_find(net_lo, net_hi)) != RTMPNULL) {
	  /* update the entry */
	    if (rtmp->state == RTMP_BAD1 && dist < 15) {
	    /* replace entry */
	      rtmp->dist = dist+1;
	      rtmp->intfc = ifn;
	      rtmp->rnet = rnet;
	      rtmp->rnode = rnode;
	      rtmp->state = RTMP_GOOD;
	      rtmpModTime = tn.tv_sec;
	    } else {
	      if (rtmp->dist >= dist+1 && dist < 15) {
	      /* replace entry */
	        if (rtmp->dist == dist+1)
	          incr(rtmpNextIREqualChanges);
	        else
	          incr(rtmpNextIRLessChanges);
	        rtmp->dist = dist+1;
	        rtmp->intfc = ifn;
	        rtmp->rnet = rnet;
	        rtmp->rnode = rnode;
	        rtmp->state = RTMP_GOOD;
	        rtmpModTime = tn.tv_sec;
	      } else {
	        if (rtmp->rnet == rnet
	         && rtmp->rnode == rnode
	         && rtmp->intfc == ifn) {
	        /* net further away now */
	          rtmpModTime = tn.tv_sec;
	          if (dist != 31) {
	            rtmp->dist = dist+1;
	            if (rtmp->dist < 16)
	              rtmp->state = RTMP_GOOD;
	            else
	              (void)rtmp_delete(rtmp);
	          } else
	            rtmp->state = RTMP_BAD1;
	        }
	      }
	    }
	  } else {
	  /* create a new entry */
	    if ((rtmp = rtmp_new()) != RTMPNULL) {
	      rtmp->net_lo = net_lo;
	      rtmp->net_hi = net_hi;
	      rtmp->dist = dist+1;
	      rtmp->intfc = ifn;
	      rtmp->rnet = rnet;
	      rtmp->rnode = rnode;
	      rtmp->state = (dist == 31) ? RTMP_BAD1 : RTMP_GOOD;
	      rtmp->zquantity = 0;
	      rtmp->zexpected = 1;
	      rtmp->zitlist = ZITNULL;
	      rtmp_insert(rtmp);
	    }
	  }
	  if (data[2] & 0x80)
	    if (rtmp != RTMPNULL)
	      rtmp->flags |= RTMP_EXTENDED;
	  data += tuplelen;
	  tuples++;
	}

	if (debug & RTMPDEBUG)
	  fprintf(stderr, "%d tuples\n", tuples);

	return;
}

/*
 * incoming RTMP Request packet
 *
 */

void
rtmpr(ifn, pd)
int ifn;
struct pd *pd;
{
	short len;
	u_char *pkt;
	u_char buf[64];
	struct pd opd;
	struct rtmp *rtmp_route();
	void ddp_route(), rtmp_send();

	pkt = pd->data;
	len = pd->dataLen;

	if (debug & RTMPDEBUG)
	  fprintf(stderr, "RTMP: %-5s Receive RTMP %sRequest\n",
	    iflist[ifn].if_name, (pkt[0] != 0x01) ? "Route Data " : "");

	if (len == 0)
	  drop(ddpTooShortErrors);
	  
	if (len > MAXDDPLEN)
	  drop(ddpTooLongErrors);

	if (pd->dstSkt != RTMP)
	  drop(uarNoSocketHandler);

	opd.lapType = LDDP;
	opd.ddpType = RTMP;
	opd.dstNet = pd->srcNet;
	opd.srcNet = iflist[ifn].net;
	opd.dstNode = pd->srcNode;
	opd.srcNode = iflist[ifn].node;
	opd.dstSkt = pd->srcSkt;
	opd.srcSkt = RTMP;
	opd.dataLen = 0;
	opd.data = pkt;
	opd.addr = NULL;
	opd.hop  = 0;

	switch (pkt[0]) {
	  case 0x01:		/* RTMP Request			*/
	    break;
	  case 0x02:		/* RTMP Route Data Req (split)	*/
	  case 0x03:		/* RTMP Route Data Req (!split)	*/
	    rtmp_send(ifn, &opd, (pkt[0] == 0x02));
	    return;
	    break;
	  default:
	    drop(rtmpInErrors);
	    break;
	}

	len = 4;
	pkt = buf+HDRSIZ+LDDPHDRSIZ;

	/* RTMP Response */

	pkt[0] = iflist[ifn].net >> 8;		/* this net	*/
	pkt[1] = iflist[ifn].net & 0xff;	/* this net	*/
	pkt[2] = IDLEN;				/* node ID len	*/
	pkt[3] = iflist[ifn].node;		/* this node	*/

	/* Extended RTMP Response */

	if (iflist[ifn].phase == PHASE2) {
	  pkt[4] = iflist[ifn].net_lo >> 8;
	  pkt[5] = iflist[ifn].net_lo & 0xff;
	  pkt[6] = 0x80;
	  pkt[7] = iflist[ifn].net_hi >> 8;
	  pkt[8] = iflist[ifn].net_hi & 0xff;
	  pkt[9] = 0x82;
	  len += 6;
	}

	opd.dataLen = len;
	opd.data = pkt;

	if (debug & RTMPDEBUG)
	  fprintf(stderr, "RTMP: %-5s Sending RTMP Response\n",
	    iflist[ifn].if_name);

	incr(ddpOutRequests);

	ddp_route(ifn, &opd, rtmp_route(opd.dstNet));

	return;
}

/*
 * the RTMP process
 *
 */

void
rtmp_process()
{
	int i;
	void rtmp_send();

	for (i = 0; i < numintfc; i++) {
	  if (iflist[i].rtmps != NULL)
	    (*(iflist[i].rtmps))(i, (struct pd *)NULL, 1);
	  else
	    rtmp_send(i, (struct pd *)NULL, 1);
	}

	return;
}

/*
 * send an RTMP data packet to 'pd' or, if NULL, out interface 'ifn'
 * Use "Split Horizon" if 'split' true
 *
 */

void
rtmp_send(ifn, pd, split)
int ifn;
struct pd *pd;
int split;
{
	struct pd opd;
	u_char *pkt, *data, buf[632];
	short i, len, dlen, maxm, incrm;
	struct rtmp *rtmp, *rtmp_route();
	void rtmp_dosend();

	incr((pd == NULL) ? rtmpOutDataPkts : rtmpOutRequestPkts);

	pkt = buf+HDRSIZ+LDDPHDRSIZ;

	/* RTMP Data */

	pkt[0] = iflist[ifn].net >> 8;		/* this net	*/
	pkt[1] = iflist[ifn].net & 0xff;	/* this net	*/
	pkt[2] = IDLEN;				/* node ID len	*/
	pkt[3] = iflist[ifn].node;		/* this node	*/

	len = 4;

	if (iflist[ifn].phase == PHASE2) {
	  pkt[4] = iflist[ifn].net_lo >> 8;
	  pkt[5] = iflist[ifn].net_lo & 0xff;
	  pkt[6] = 0x80;
	  pkt[7] = iflist[ifn].net_hi >> 8;
	  pkt[8] = iflist[ifn].net_hi & 0xff;
	  pkt[9] = 0x82;
	  len += 6;
	} else {
	  pkt[4] = 0x00;
	  pkt[5] = 0x00;
	  pkt[6] = 0x82;
	  len += 3;
	}

	maxm = 587-len;
	data = pkt+len;
	dlen = 0;

	/*
	 * if not directed reply, send as local broadcast
	 *
	 */
	if (pd == NULL) {
	  opd.lapType = LDDP;
	  opd.ddpType = RTMP;
	  opd.dstNet = iflist[ifn].net;
	  opd.srcNet = iflist[ifn].net;
	  opd.dstNode = BCAST;
	  opd.srcNode = iflist[ifn].node;
	  opd.dstSkt = RTMP;
	  opd.srcSkt = RTMP;
	  opd.data = pkt;
	  opd.dataLen = len;
	  opd.addr = NULL;
	  opd.hop  = 0;
	  pd = &opd;
	}

	/*
	 * copy in RTMP tuples
	 *
	 */
	for (i = 0; i < RTMPTABSIZ; i++) {
	  rtmp = rtmpTab[i];
	  while (rtmp != RTMPNULL) {
	    if (rtmp->intfc != ifn || !split) {
	    /* implement split horizon */
	      if (rtmp->dist < 15) {
	      /* distance less than 15 */
	        data[0] = rtmp->net_lo >> 8;
	        data[1] = rtmp->net_lo & 0xff;
	        data[2] = (rtmp->state < RTMP_BAD1) ? rtmp->dist : 31;
	        incrm = 3;
	        if (rtmp->flags & RTMP_EXTENDED) {
	          data[2] |= 0x80;
	          data[3] = rtmp->net_hi >> 8;
	          data[4] = rtmp->net_hi & 0xff;
	          data[5] = 0x82;
	          incrm = 6;
	        }
	        if ((dlen+incrm) > maxm) {
	          pd->data = pkt;
	          pd->dataLen = len+dlen;
	          rtmp_dosend(ifn, pd);
	          data = pkt+len;
	          dlen = 0;
	          continue;
	        }
	        data += incrm;
	        dlen += incrm;
	      }
	    }
	    rtmp = rtmp->next;
	  }
	}

	pd->data = pkt;
	pd->dataLen = len+dlen;

	rtmp_dosend(ifn, pd);

	return;
}

/*
 * send RTMP data
 *
 */

void
rtmp_dosend(ifn, pd)
int ifn;
struct pd *pd;
{
	struct rtmp *rtmp_route();
	void ddp_route();

	if (debug & RTMPDEBUG)
	  fprintf(stderr, "RTMP: %-5s Sending RTMP data\n",
	    iflist[ifn].if_name);

	incr(ddpOutRequests);

	ddp_route(ifn, pd, rtmp_route(pd->dstNet));

	return;
}

/*
 * find an entry in the RTMP table.
 *
 */

struct rtmp *
rtmp_find(net_lo, net_hi)
u_short net_lo, net_hi;
{
	struct rtmp *rtmp;
	int rtmp_hash();

	rtmp = rtmpTab[rtmp_hash(net_lo, net_hi)];

	while (rtmp != RTMPNULL) {
	  if (rtmp->net_lo == net_lo
	   && rtmp->net_hi == net_hi)
	    return(rtmp);
	  rtmp = rtmp->next;
	}

	return(RTMPNULL);
}

/*
 * find an RTMP route to net
 *
 */

struct rtmp *
rtmp_route(net)
register u_short net;
{
	register int i;
	register struct rtmp *rtmp;

	incr(uarRTMPRoutes);

	if ((rtmp = rtmpMRU) != RTMPNULL) {
	  incr(uarRTMPSearches);
	  if (net >= rtmp->net_lo
	   && net <= rtmp->net_hi)
	    return(rtmp);
	}

	for (i = 0; i < 17; i++) {
	  rtmp = rtmpTab[((net >> i) & RTMPTABMSK)];
	  while (rtmp != RTMPNULL) {
	    incr(uarRTMPSearches);
	    if (net >= rtmp->net_lo
	     && net <= rtmp->net_hi)
	      return((rtmpMRU = rtmp));
	    rtmp = rtmp->next;
	  }
	}

	return(RTMPNULL);
}

/*
 * RTMP validity timer, check & age each entry.
 *
 */

void
rtmp_timer()
{
	int i, nrtmp;
	struct rtmp *rtmp, *rtmp_delete();

	for (i = 0, nrtmp = 0; i < RTMPTABSIZ; i++) {
	  rtmp = rtmpTab[i];
	  while (rtmp != RTMPNULL) {
	    switch (rtmp->state) {
	      case RTMP_GOOD:
	        if (rtmp->dist != 0)
	  	  rtmp->state = RTMP_SUSP;
	        break;
	      case RTMP_SUSP:
	        rtmp->state = RTMP_BAD1;
	        break;
	      case RTMP_BAD1:
	        rtmp->state = RTMP_BAD2;
	        break;
	      case RTMP_BAD2:
	        rtmp = rtmp_delete(rtmp);
	        continue;
	        break;
	    }
	    nrtmp++;
	    rtmp = rtmp->next;
	  }
	}

	if (debug & RTMPDEBUG)
	  fprintf(stderr, "RTMP: VLDT %d entries\n", nrtmp);

	return;
}

/*
 * insert RTMP tuple in linked list
 *
 */

void
rtmp_insert(rtmp)
struct rtmp *rtmp;
{
	int idx, rtmp_hash();
	struct rtmp *p, *q;

	idx = rtmp_hash(rtmp->net_lo, rtmp->net_hi);

	p = rtmpTab[idx];
	q = RTMPNULL;

	while (p != RTMPNULL) {
	  if (p->net_lo >= rtmp->net_lo)
	    break;
	  q = p;
	  p = p->next;
	}

	rtmp->next = p;

	if (q == RTMPNULL)
	  rtmpTab[idx] = rtmp;
	else
	  q->next = rtmp;

	rtmpModTime = tn.tv_sec;

	return;
}

/*
 * delete RTMP tuple from linked list
 *
 */

struct rtmp *
rtmp_delete(rtmp)
struct rtmp *rtmp;
{
	int idx, rtmp_hash();
	struct rtmp *next, *p, *q;
	void rtmp_free();

	idx = rtmp_hash(rtmp->net_lo, rtmp->net_hi);

	p = rtmpTab[idx];
	q = RTMPNULL;

	while (p != RTMPNULL) {
	  if (p == rtmp)
	    break;
	  q = p;
	  p = p->next;
	}

	if (p == RTMPNULL)
	  return(p);

	next = rtmp->next;

	if (q == RTMPNULL)
	  rtmpTab[idx] = next;
	else
	  q->next = next;

	rtmp_free(rtmp);

	rtmpModTime = tn.tv_sec;

	return(next);
}

/*
 * simple hash of high and low net numbers
 *
 */

int
rtmp_hash(net_lo, net_hi)
register u_short net_lo, net_hi;
{
	register int i, idx;

	for (i = 0; i < 17; i++)
	  if ((idx = ((net_lo>>i) & RTMPTABMSK)) == ((net_hi>>i) & RTMPTABMSK))
	    return(idx);
	
	return(0);
}

/*
 * dump RTMP table
 *
 */

char *state[] = { "GOOD", "SUSPECT", "BAD", "BAD" };

void
rtmp_dump()
{
	int count;
	int i, znum = 0;
	FILE *r, *fopen();
	struct rtmp *rtmp;
	struct zit *zit;

	if ((r = fopen(DUMP_RTMP, "w+")) == NULL) {
	  incr(uarBadDumpFile);
	  return;
	}

	fprintf(r, "UAR RTMP Table:\n\n");
	fprintf(r, "low       high     dst ifc brNet  Nd   State Zone\n");

	for (i = 0, count = 0; i < RTMPTABSIZ; i++) {
	  rtmp = rtmpTab[i];
	  while (rtmp != RTMPNULL) {
	    fprintf(r, "%3d.%-3d %c %3d.%-3d %3d %3d %3d.%-3d %2x %7s",
	      rtmp->net_lo >> 8, rtmp->net_lo & 0xff,
	      (rtmp->flags & RTMP_EXTENDED) ? '-' : ' ',
	      rtmp->net_hi >> 8, rtmp->net_hi & 0xff,
	      rtmp->dist, rtmp->intfc, rtmp->rnet >> 8,
	      rtmp->rnet & 0xff, rtmp->rnode, state[rtmp->state]);
	    zit = rtmp->zitlist;
	    while (zit != ZITNULL) {
	      if (znum > 0)
	        fprintf(r, "%42s", "");
	      fprintf(r, " %s\n", zit->zone->name+1);
	      zit = zit->next;
	      znum++;
	    }
	    if (znum == 0)
	      fprintf(r, "\n");
	    znum = 0;
	    rtmp = rtmp->next;
	    count++;
	  }
	}
	fprintf(r, "\n\n%d Networks\n", count);

	(void)fclose(r);

	if (debug & RTMPDEBUG)
	  fprintf(stderr, "RTMP: dumped RTMP table, %d entries\n", count);

	return;
}

/*
 * create a new RTMP element
 *
 */

struct rtmp *
rtmp_new()
{
	struct rtmp *rtmp;

	rtmp = (struct rtmp *)malloc(sizeof(struct rtmp));

	if (rtmp == RTMPNULL) {
	  incr(uarMemoryError);
	  incr(rtmpRoutingTableOverflows);
	  return(RTMPNULL);
	}

	rtmp->flags = 0x00;

	return(rtmp);
}

/*
 * free memory occupied by RTMP element
 *
 */

void
rtmp_free(rtmp)
struct rtmp *rtmp;
{
	incr(rtmpRouteDeletes);

	if (debug & RTMPDEBUG)
	  fprintf(stderr, "RTMP: FREE  %3d.%-3d-%3d.%-3d\n",
	    rtmp->net_lo >> 8, rtmp->net_lo & 0xff,
	    rtmp->net_hi >> 8, rtmp->net_hi & 0xff);

	if (rtmpMRU == rtmp)
	  rtmpMRU = RTMPNULL;

	free((char *)rtmp);

	return;
}
