/*
 * UAR - UNIX AppleTalk Router
 * 
 * Copyright (c) 1993, The University of Melbourne.
 *
 * All Rights Reserved.
 *
 * UAR may NOT be publicly redistributed (for example via
 * anonymous FTP), sold, or the source used for any other
 * purpose without the permission of the copyright owner.
 * 
 * This software is supplied "as is" without express or
 * implied warranty.
 * 
 * djh@munnari.OZ.AU
 *
 * $Author: djh $
 * $Revision: 1.1.1.1 $
 *
 */

#include "uar.h"

extern int debug;
extern u_long uar_stats[];

int stats_do_dump = 0;
u_long pkts_dropped = 0;

char *stats_msg[] = {
	"unknown ethernet type field",
	"unknown LAP packet type field",
	"packet addressed to bad socket",
	"",
	"DDP packet type unknown",
	"packet sent by self, ignore",
	"no AARP table entry for node",
	"no RTMP table entry for address",
	"malloc failed for new list entry",
	"packet not supported for phase",
	"unknown zone name in NBP lookup",
	"no default zone for interface",
	"cannot open file for dumping",
	"packet header bad format",
	"CAP packets transmitted to loopback",
	"CAP packets received from loopback",
	"MultiNode packets transmitted",
	"MultiNode packets received",
	"AARP packets transmitted",
	"AARP packets received",
	"AARP cache searches",
	"AARP cache hits",
	"local DDP requests for transmission",
	"short DDP datagrams transmitted",
	"long  DDP datagrams transmitted",
	"DDP datagrams received",
	"DDP datagrams: for router",
	"DDP datagrams: no handler",
	"DDP datagrams: too short",
	"DDP datagrams: too long",
	"DDP datagrams: not for router",
	"DDP datagrams: bad checksum",
	"DDP datagrams: to be routed",
	"DDP datagrams: dropped, no route",
	"DDP datagrams: broadcast, not for me",
	"DDP datagrams: hop count exceeded",
	"RTMP data packets received",
	"RTMP data packets transmitted",
	"RTMP request packets received",
	"RTMP next IR changes: equal distance",
	"RTMP next IR changes: shorter distance",
	"RTMP routing table deletes",
	"RTMP routing table overflows",
	"RTMP request packets transmitted",
	"RTMP packets rejected: version mismatch",
	"RTMP packets rejected: unspecified error",
	"ZIP queries received",
	"ZIP replies received",
	"ZIP extended replies received",
	"ZIP conflicts detected",
	"ZIP packets rejected: obsolete type",
	"ZIP packets rejected: unspecified error",
	"NBP lookup requests received",
	"NBP lookup replies received",
	"NBP broadcast requests received",
	"NBP forward requests received",
	"NBP lookup replies transmitted",
	"NBP registration failures",
	"NBP packets rejected: unspecified error",
	"AEP requests received",
	"AEP replies transmitted",
	"AEP requests transmitted",
	"AEP requests received",
	"ATP packets received",
	"ATP packets transmitted",
	"", /* RTMP Routes */
	"", /* RTMP Searches */
	"TNNL packets received",
	"TNNL packets transmitted",
	"TNNL packets not for me, dropped",
	"TNNL packets from bogus source addr",
	"", /* ZONE Routes */
	"", /* ZONE Searches */
	""
};

/*
 * initialize the stats structures
 *
 */

void
stats_initialize()
{
	int i;

	for (i = 0; i < maxCounters; i++)
	  uar_stats[i] = 0L;

	return;
}

/*
 * dump the statistics
 *
 */

void
stats_dump()
{
	int i;
	FILE *r, *fopen();

	if ((r = fopen(DUMP_STAT, "w+")) == NULL) {
	  incr(uarBadDumpFile);
	  return;
	}
	
	fprintf(r, "UAR Statistics:\n\n");

	for (i = 0; i < maxCounters; i++)
	  if (*stats_msg[i] != '\0')
	    fprintf(r, "%10d %s\n", uar_stats[i], stats_msg[i]);

	if (uar_stats[uarRTMPRoutes] == 0)
	  uar_stats[uarRTMPRoutes] = 1;

	if (uar_stats[uarZONEFinds] == 0)
	  uar_stats[uarZONEFinds] = 1;

	fprintf(r, "\n");
	fprintf(r, "%10d packets dropped\n", pkts_dropped);
	fprintf(r, "%10.2f average RTMP searches per route request\n",
	  (float)uar_stats[uarRTMPSearches]/(float)uar_stats[uarRTMPRoutes]);
	fprintf(r, "%10.2f average ZONE searches per ZONE lookup\n",
	  (float)uar_stats[uarZONESearches]/(float)uar_stats[uarZONEFinds]);
	fprintf(r, "\n\n");

	(void)fclose(r);

	return;
}
