/*
 * UAR - UNIX AppleTalk Router
 * 
 * Copyright (c) 1993, The University of Melbourne.
 *
 * All Rights Reserved.
 *
 * UAR may NOT be publicly redistributed (for example via
 * anonymous FTP), sold, or the source used for any other
 * purpose without the permission of the copyright owner.
 * 
 * This software is supplied "as is" without express or
 * implied warranty.
 * 
 * djh@munnari.OZ.AU
 *
 * $Author: djh $
 * $Revision: 1.1.1.3 $
 *
 */

#include "uar.h"

extern int debug;
extern int numintfc;
extern struct iflist iflist[];
extern u_long pkts_dropped;
extern u_long uar_stats[];
extern char *stats_msg[];
extern u_char buf[];

static int tnnlfd = -1;
static struct sockaddr_in lsin;

/*
 * open tunnel UDP loopback port and set socket options
 *
 */

int
tnnl_open(interface, mode, protocol, typ)
char *interface;
u_short protocol;
int mode, typ;
{
	int i;
	int len;
	u_long addr = 0;
	struct servent *s, *getservbyname();
	void exit();

	if (tnnlfd != -1)
	  return(tnnlfd);

	if ((tnnlfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
	  perror("socket()");
	  exit(1);
	}

	/*
	 * find the interface number & get the source ip address
	 *
	 */
	for (i = 0; i < numintfc; i++) {
	  if (strcmp(interface, iflist[i].if_name) == 0) {
	    if (iflist[i].mnode[iflist[i].node].client == MNODE_TNNL
	     && iflist[i].mnode[iflist[i].node].state == MNODE_VALID)
	      addr = iflist[i].mnode[iflist[i].node].addr;
	    break;
	  }
	}

	s = getservbyname("tnnl", "udp");

	lsin.sin_family = AF_INET;
	lsin.sin_addr.s_addr = (addr) ? addr : INADDR_ANY;
	lsin.sin_port = (s == NULL) ? htons(TNNL_PORT) : s->s_port;

	if (bind(tnnlfd, (struct sockaddr *)&lsin, sizeof(lsin)) < 0) {
	  perror("bind()");
	  exit(1);
	}

	len = SKTBUFSIZ;
	if (setsockopt(tnnlfd, SOL_SOCKET, SO_RCVBUF, (char *)&len,
	 sizeof(int)) < 0) {
	  perror("SO_RCVBUF");
	  exit(1);
	}

	len = SKTBUFSIZ;
	if (setsockopt(tnnlfd, SOL_SOCKET, SO_SNDBUF, (char *)&len,
	 sizeof(int)) < 0) {
	  perror("SO_SNDBUF");
	  exit(1);
	}

	return(tnnlfd);
}

/*
 * read from the tunnel port
 *
 */

void
tnnl_read(ifn, fd)
int ifn, fd;
{
	u_char *pkt;
	u_char dstNode;
	u_char srcNode;
	u_char lapType;
	int cc, fromlen;
	struct sockaddr_in from;
	void ddp(), dumppacket();
	void tnnl_sendconf();

	fromlen = sizeof(struct sockaddr);

	if ((cc = recvfrom(fd, (char *)(buf+HDRSIZ), BUFSIZE-HDRSIZ, 0,
	 (struct sockaddr *)&from, &fromlen)) < 0) {
	  if (debug) perror("readtnnl()");
	  return;
	}

	if (cc == 0)
	  return;

	pkt = buf+HDRSIZ;

	incr(uarTNNLPacketsIn);

	if (debug & TNNLDEBUG) {
	  fprintf(stderr, " UAR: %-5s READ  %3d bytes\n",
	    iflist[ifn].if_name, cc);
	  if (debug & PKTDEBUG)
	    dumppacket(pkt, cc);
	}

	dstNode = pkt[1];
	srcNode = pkt[2];
	lapType = pkt[3];

	/*
	 * check if configuration request
	 *
	 */
	if (pkt[0] == UAR_CONF) {
	  if (dstNode == 0x00) {
	    switch (lapType) {
	      case UAR_CONF_REQ:
		tnnl_sendconf(ifn, &from);
		break;
	      case UAR_CONF_DATA:
		/*
		 * ignore it for now
		 *
		 */
		break;
	    }
	  }
	  return;
	}

	/*
	 * check destination
	 *
	 */
	if (dstNode != iflist[ifn].node && dstNode != BCAST)
	  drop(uarTNNLNotForMe);

	/*
	 * check source address
	 *
	 */
	if (iflist[ifn].mnode[srcNode].client != MNODE_TNNL
	 || iflist[ifn].mnode[srcNode].state != MNODE_VALID
	 || iflist[ifn].mnode[srcNode].addr != from.sin_addr.s_addr)
	  drop(uarTNNLBogusSrc);


	/*
	 * ddp() expects an ELAP header, fake it
	 *
	 */
	switch (iflist[ifn].phase) {
	  case PHASE1:
	    cc += 13;
	    pkt -= 13;
	    bzero((char *)pkt, 17);
	    pkt[12] = 0x80;
	    pkt[13] = 0x9b;
	    pkt[14] = dstNode;
	    pkt[15] = srcNode;
	    pkt[16] = lapType;
	    break;
	  case PHASE2:
	    cc += 18;
	    pkt -= 18;
	    bzero((char *)pkt, 22);
	    pkt[20] = 0x80;
	    pkt[21] = 0x9b;
	    break;
	  default:
	    return;
	    break;
	}

	ddp(ifn, pkt, cc);

	return;
}

/*
 * send packets to tunnel clients via UDP/IP
 *
 */

int
tnnl_write(ifn, pkt, len)
register u_char *pkt;
int ifn, len;
{
	int i;
	u_char dstNode;
	u_char srcNode;
	char *inet_ntoa();
	struct sockaddr_in to;
	void dumppacket();

	dstNode = pkt[1];
	srcNode = pkt[2];

	if (dstNode == BCAST) {
	  for (i = 1; i < MAXNODES-1; i++) {
	    if (i != srcNode) {
	      if (iflist[ifn].mnode[i].client == MNODE_TNNL
	       && iflist[ifn].mnode[i].state == MNODE_VALID) {
	        pkt[1] = i; /* not BCAST */
	        (void)tnnl_write(ifn, pkt, len);
	       }
	    }
	  }
	  return(0);
	}

	if (iflist[ifn].mnode[dstNode].client != MNODE_TNNL
	 || iflist[ifn].mnode[dstNode].state != MNODE_VALID)
	  return(0);

	to.sin_family = AF_INET;
	to.sin_port = lsin.sin_port;
	to.sin_addr.s_addr = iflist[ifn].mnode[dstNode].addr;

	if (debug & TNNLDEBUG) {
	  fprintf(stderr, " UAR: %-5s WRITE %3d bytes to %d @ %s\n",
	    iflist[ifn].if_name, len, dstNode, inet_ntoa(to.sin_addr));
	  if (debug & PKTDEBUG)
	    dumppacket(pkt, len);
	}

	incr(uarTNNLPacketsOut);

	if (sendto(iflist[ifn].etfd, (char *)pkt, len, 0,
	 (struct sockaddr *)&to, sizeof(struct sockaddr)) < 0)
	  if (debug) perror("sendto()");

	return(len);
}

/*
 * send the TNNL configuration data to requestor
 * (silently ignored if not a configured client)
 *
 */

void
tnnl_sendconf(ifn, host)
int ifn;
struct sockaddr_in *host;
{
	int i, len;
	u_char *pkt;
	u_long *addr;
	char *inet_ntoa();
	void dumppacket();

	if (host->sin_addr.s_addr == 0)
	  return;

	/*
	 * check if requestor is known peer
	 *
	 */
	for (i = 1; i < MAXNODES-1; i++)
	  if (iflist[ifn].mnode[i].addr == (u_long)host->sin_addr.s_addr)
	    break;

	if (i == MAXNODES-1)
	  return;

	/*
	 * build packet
	 *
	 */
	pkt = buf+HDRSIZ;
	
	pkt[0] = UAR_CONF;
	pkt[1] = 0x00;
	pkt[2] = 0x00;
	pkt[3] = UAR_CONF_DATA;
	pkt[4] = iflist[ifn].net >> 8;
	pkt[5] = iflist[ifn].net & 0xff;
	pkt[6] = 0x00;	/* must be zero */
	pkt[7] = i & 0xff; /* requestor node */

	/*
	 * copy client address data
	 *
	 */
	addr = (u_long *)(pkt+8);

	for (i = 0; i < MAXNODES; i++)
	  *addr++ = iflist[ifn].mnode[i].addr;

	len = 1032;

	if (debug & TNNLDEBUG) {
	  fprintf(stderr, " UAR: %-5s WRITE %3d config bytes to %d @ %s\n",
	    iflist[ifn].if_name, len, pkt[7], inet_ntoa(host->sin_addr));
	  if (debug & PKTDEBUG)
	    dumppacket(pkt, len);
	}

	incr(uarTNNLPacketsOut);

	if (sendto(iflist[ifn].etfd, (char *)pkt, len, 0,
	 (struct sockaddr *)host, sizeof(struct sockaddr)) < 0)
	  if (debug) perror("sendto()");

	return;
}
