/*
 * UAR - UNIX AppleTalk Router
 * 
 * Copyright (c) 1993, The University of Melbourne.
 *
 * All Rights Reserved.
 *
 * UAR may NOT be publicly redistributed (for example via
 * anonymous FTP), sold, or the source used for any other
 * purpose without the permission of the copyright owner.
 * 
 * This software is supplied "as is" without express or
 * implied warranty.
 * 
 * djh@munnari.OZ.AU
 *
 * $Author: djh $
 * $Revision: 1.1.1.7 $
 *
 */

#include "uar.h"

int etfd;				/* EtherTalk ELAP socket	*/
int aarpfd;				/* EtherTalk AARP socket	*/
int capfd = -1;				/* CAP service loopback socket	*/
int docap = 0;				/* Attach CAP to first i/face	*/

int phase = 2;				/* AppleTalk Phase 1 or 2	*/
int debug = 0;				/* See uar.h for debug flags	*/
int promisc = 0;			/* Interface non-promiscuous	*/
int numintfc = 0;			/* Number of interfaces in use	*/
int checksum = 1;			/* Check/Send DDP checksums	*/

u_long uar_stats[maxCounters];		/* UAR statistics		*/

char *conffile = CONFIG_FILE;		/* Default configuration file	*/
char *logfile = LOGFILE;		/* Default logfile		*/
char *interface = NULL;			/* Network interface		*/
char *progname = NULL;			/* Saved argv[0]		*/
char *thisZone = NULL;			/* server zone name		*/
char *timesrvr = NULL;			/* time server name		*/
char *linep;				/* line pointer for configure()	*/

u_char baddr[] = { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff }; /* broadcast	*/
u_char maddr[] = { 0x09, 0x00, 0x07, 0xff, 0xff, 0xff }; /* multicast	*/

struct timeval tp;			/* time to run protocol driver	*/
struct timeval tn;			/* time packet read from net	*/

struct iflist iflist[MAXIF];		/* interface list		*/

u_char buf[BUFSIZE];			/* receive buffer		*/
char hostname[MAXHOST];			/* machine name			*/


main(argc, argv)
int argc;
char *argv[];
{
	char *p;
	int c, i;
	fd_set fdset;
	int nfds, res;
	struct timeval to;
	extern char *optarg;
	extern int optind, opterr;
	void cap_initialize(), readcap();
	void rtmp_initialize(), rtmp_seed();
	void zip_initialize(), zip_multicast();
	void aarp_initialize(), nbp_initialize(), stats_initialize();
	void readeth(), readaarp(), readdev();
	void configure(), process(), cleanup();
	void uar_dump_sig();
	void usage(), log();
	void strcpycp();
	int write_eth();
	int init_eth();
	long strtol();
	void exit();

	opterr = 0;
	progname = *argv;

	/*
	 * module initialization
	 *
	 */
	stats_initialize();
	aarp_initialize();
	rtmp_initialize();
	nbp_initialize();
	zip_initialize();

	/*
	 * set up signal handlers
	 *
	 */
	(void)signal(SIGINT,  cleanup);
	(void)signal(SIGTERM, cleanup);
	(void)signal(SIGQUIT, uar_dump_sig);
	(void)signal(SIGHUP,  SIG_IGN);
	(void)signal(SIGUSR1, SIG_IGN);
	(void)signal(SIGUSR2, SIG_IGN);

	/*
	 * process command line options
	 *
	 */
	while ((c = getopt(argc, argv, "d:f:l:t:z:c12Cv")) != -1) {
	  switch (c) {
	    case 'd':
	      debug = (int)strtol(optarg, (char **)NULL, 0);
	      break;
	    case 'f':
	      conffile = optarg;
	      break;
	    case 'l':
	      logfile = optarg;
	      break;
	    case 't':
	      timesrvr = optarg;
	      break;
	    case 'v':
	      fprintf(stderr, "UAR 1.0.16 September 1995\n");
	      exit(0);
	      break;
	    case 'z':
	      thisZone = optarg;
	      break;
	    case 'c':
	      checksum = 0;
	      break;
	    case 'C':
	      docap = 1;
	      break;
	    case '1':
	      phase = 1;
	      break;
	    case '2':
	      phase = 2;
	      break;
	    case '?':
	      usage();
	      exit(1);
	      break;
	  }
	}

#ifdef sony_news
#ifndef sony_phaseII
	if (phase == PHASE2) {
	  fprintf(stderr, "Phase 2 is not supported.\n");
	  exit(1);
	}
#endif /* sony_phaseII */
#endif /* sony_news */

	/*
	 * find host name
	 *
	 */
	hostname[0] = '\0';
	if (gethostname(hostname, MAXHOST) == 0) {
	  if ((p = (char *)index(hostname, '.')) != NULL)
	    *p = '\0'; /* kill host domain names */
	}

	(void)gettimeofday(&tn, (struct timezone *)NULL);

	log("UAR starting");

	/*
	 * basic interface initialization
	 *
	 */
	for (numintfc = 0;  optind < argc; optind++, numintfc++) {
	  iflist[numintfc].etfd = -1;
	  iflist[numintfc].aarpfd = -1;
	  iflist[numintfc].net = 0;
	  iflist[numintfc].net_lo = 0;
	  iflist[numintfc].net_hi = 0;
	  iflist[numintfc].node = 0xfe;		/* initial node number	*/
	  iflist[numintfc].itype = ITYPE_ETHERTALK;
	  iflist[numintfc].state = ADDR_PENDING;
	  iflist[numintfc].phase = phase;
	  iflist[numintfc].count = AARPCount;
	  iflist[numintfc].if_name = argv[optind];
	  iflist[numintfc].if_timesrvr[0] = '\0';
	  iflist[numintfc].defzone = ZONENULL;
	  iflist[numintfc].ourzone = ZONENULL;
	  iflist[numintfc].clients = 0;
	  iflist[numintfc].iopen = init_eth;
	  iflist[numintfc].write = write_eth;
	  iflist[numintfc].reade = readeth;
	  iflist[numintfc].reada = readaarp;
	  iflist[numintfc].rtmps = NULL;
	  bzero((char *)iflist[numintfc].eaddr,
	    sizeof(iflist[numintfc].eaddr));
	  iflist[numintfc].if_nbpname[1] = '\0';
	  p = (char *)iflist[numintfc].if_nbpname;
	  strncat(p+1, hostname, 25);
	  strncat(p+1, "-", 1);
	  strncat(p+1, iflist[numintfc].if_name, 5);
	  p[0] = (char)strlen(p+1);
	  bzero((char *)iflist[numintfc].mnode,
	    sizeof(iflist[numintfc].mnode));
	}

	if (numintfc == 0) {
	  fprintf(stderr, " UAR: no interfaces specified\n");
	  exit(1);
	}

	/*
	 * read interface configuration file
	 *
	 */
	configure(conffile);

	/*
	 * attach CAP to first interface, if requested
	 *
	 */
	if (docap && capfd == -1)
	  cap_initialize(0);

	/*
	 * open & configure each of the interfaces
	 *
	 */
	for (i = 0; i < numintfc; i++) {

	  interface = iflist[i].if_name;

	  /* ELAP */
	  if (iflist[i].iopen != NULL) {
	    if ((etfd = (*(iflist[i].iopen))(interface,
	      O_RDWR, 0x809b, iflist[i].phase)) < 0) {
	        fprintf(stderr, " UAR: failed to open %s\n", interface);
	        exit(1);
	    }
	  }
	  iflist[i].etfd = etfd;

	  /* AARP */
	  iflist[i].aarpfd = -1;
	  if (iflist[i].itype == ITYPE_ETHERTALK) {
	    if (eth_mopen(iflist[i].phase)) {
	    /* interface handles multiple opens */
	      if ((aarpfd = (*(iflist[i].iopen))(interface,
	        O_RDWR, 0x80f3, iflist[i].phase)) < 0) {
	          fprintf(stderr, " UAR: failed to open %s\n", interface);
	          exit(1);
	      }
	      iflist[i].aarpfd = aarpfd;
	    } else {
	    /* must demultiplex reads */
	      iflist[i].reade = readdev;
	      iflist[i].reada = NULL;
	    }
	  }

	  if (iflist[i].itype == ITYPE_ETHERTALK)
	  /* get interface ethernet address */
	    eth_addr(etfd, iflist[i].if_name, iflist[i].eaddr);

	  if (iflist[i].phase == PHASE2) {
	    if (iflist[i].node == 0xfe)		/* invalid node number	*/
	      iflist[i].node = 0xfd;		/*     for Phase 2 use	*/
	    if (iflist[i].net_lo == 0) {	/* not initialised yet	*/
	      iflist[i].net = P2STARTUP;	/* startup range 255.0	*/
	      iflist[i].net_hi = P2ENDSTUP;	/* 0 through to 255.254	*/
	    }
	  }

	  /*
	   * keep startup node number
	   *
	   */
	  iflist[i].hint = iflist[i].node;

	  /*
	   * timeserver required ?
	   *
	   */
	  if (i == 0 && timesrvr != NULL
	   && iflist[0].if_timesrvr[0] == '\0')
	    strcpycp(timesrvr, iflist[0].if_timesrvr);
	}

	if (debug)
	  fprintf(stderr, " UAR: initialised %d interface%s\n",
	     numintfc, (numintfc == 1) ? "" : "s");

	nfds = MAXIF*2;
	FD_ZERO(&fdset);

	/*
	 * disassociate
	 *
	 */
	if (!debug) {
	  if (fork())
	    exit(0);
#ifndef POSIX
#ifdef TIOCNOTTY
	  { int f;
	    if ((f = open("/dev/tty", O_RDWR, 0644)) >= 0) {
	      (void)ioctl(f, TIOCNOTTY, (char *)0);
	      (void)close(f);
	    }
	  }
#endif /* TIOCNOTTY */
#else /* POSIX */
	  (void)setsid();
#endif /* POSIX */
	}
	
	/*
	 * add multicast addresses to interfaces after disassociation
	 * (some operating systems associate mcasts with process ids,
	 * and some associate mcasts with descriptors ie: Solaris 2.N)
	 *
	 */
	for (i = 0; i < numintfc; i++) {
	  if (iflist[i].itype == ITYPE_ETHERTALK) {
	    if (iflist[i].phase == PHASE2) {
	      if (iflist[i].etfd >= 0)
	        eth_addmulti(iflist[i].etfd, iflist[i].if_name, maddr);
#ifndef SINGLE_MCAST
	      if (iflist[i].aarpfd >= 0)
	        eth_addmulti(iflist[i].aarpfd, iflist[i].if_name, maddr);
#endif /* SINGLE_MCAST */
	      if (iflist[i].net_lo != 0) /* interface already seeded */
	        zip_multicast(i, ADDMULTI);
	    }
	  }
	}

	/*
	 * run ...
	 *
	 */
	for ( ; ; ) {
	  for (i = 0; i < numintfc; i++) {
	    if (iflist[i].etfd >= 0)
	      FD_SET(iflist[i].etfd, &fdset);
	    if (iflist[i].aarpfd >= 0)
	      FD_SET(iflist[i].aarpfd, &fdset);
	  }
	  if (capfd >= 0)
	    FD_SET(capfd, &fdset);

	  to.tv_sec  = 0;
	  to.tv_usec = 200000;	/* 200 ms */

	  if ((res = select(nfds, &fdset, (fd_set *)0, (fd_set *)0, &to))>=0){
	    (void)gettimeofday(&tn, (struct timezone *)NULL);
	    for (i = 0; i < numintfc; i++) {
	      if (iflist[i].etfd >= 0)
	        if (FD_ISSET(iflist[i].etfd, &fdset))
	          if (iflist[i].reade != NULL)
	            (*(iflist[i].reade))(i, iflist[i].etfd);
	      if (iflist[i].aarpfd >= 0)
	        if (FD_ISSET(iflist[i].aarpfd, &fdset))
	          if (iflist[i].reada != NULL)
	            (*(iflist[i].reada))(i, iflist[i].aarpfd);
	    }
	    if (capfd >= 0)
	      if (FD_ISSET(capfd, &fdset))
	        readcap(capfd);
	    if (res == 0 || protTimeout())
	      process();
	  }
	}
	/*NOTREACHED*/
}

void
usage()
{
	fprintf(stderr, "usage: %s [-d flags] [-f uar.conf]\n", progname);
	fprintf(stderr, "\t[-c] [-C] [-l logfile] [-t timeserver]\n");
	fprintf(stderr, "\t[-z zone] [-1] [-2] interfaces\n");

	return;
}

/*
 * check if the protocol process needs running
 * (only called if select() timeout didn't happen)
 *
 */

int
protTimeout()
{
	if (tn.tv_sec > tp.tv_sec)
	  return(1);

	if (tn.tv_sec == tp.tv_sec)
	  if (tn.tv_usec >= tp.tv_usec)
	    return(1);

	return(0);
}

/*
 * do the protocol housekeeping
 * (needs to run every 200 ms for AARP)
 *
 */

void
process()
{
	static int rtmpValTimer = 0;
	static int rtmpTimer = 0;
	static int nbpTimer = 0;
	static int zipTimer = 0;
	extern int rtmp_do_dump;
	extern int nbp_do_dump;
	extern int aarp_do_dump;
	extern int zone_do_dump;
	extern int stats_do_dump;
	void rtmp_process(), rtmp_timer();
	void aarp_process(), aarp_dump();
	void zip_process(), zone_dump();
	void nbp_process(), nbp_dump();
	void rtmp_dump(), stats_dump();

	/*
	 * set the time the protocol process has to run next
	 *
	 */
	tp.tv_sec = tn.tv_sec;
	if ((tp.tv_usec = tp.tv_usec + 200000) >= 1000000) {
	  tp.tv_usec -= 1000000;
	  tp.tv_sec++;
	}

	/*
	 * RTMP validity timer
	 *
	 */
	if (rtmpValTimer++ >= RTMPValTimer) {
	  rtmpValTimer = 0;
	  rtmp_timer();
	}

	/*
	 * RTMP process
	 *
	 */
	if (rtmpTimer++ >= RTMPTimer) {
	  rtmpTimer = 0;
	  rtmp_process();
	}

	/*
	 * NBP process
	 *
	 */
	if (nbpTimer++ >= NBPTimer) {
	  nbpTimer = 0;
	  nbp_process();
	}

	/*
	 * ZIP process
	 *
	 */
	if (zipTimer++ >= ZIPTimer) {
	   zipTimer = 0;
	   zip_process();
	 }

	/*
	 * AARP process
	 *
	 */
	aarp_process();

	/*
	 * handle dump requests
	 *
	 */
	 if (rtmp_do_dump) {
	   rtmp_do_dump = 0;
	   rtmp_dump();
	 }
	 if (nbp_do_dump) {
	   nbp_do_dump = 0;
	   nbp_dump();
	 }
	 if (aarp_do_dump) {
	   aarp_do_dump = 0;
	   aarp_dump();
	 }
	 if (zone_do_dump) {
	   zone_do_dump = 0;
	   zone_dump();
	 }
	 if (stats_do_dump) {
	   stats_do_dump = 0;
	   stats_dump();
	 }
	
	return;
}

/*
 * UAR dump signal handler
 *
 */

void
uar_dump_sig()
{
	extern int rtmp_do_dump;
	extern int nbp_do_dump;
	extern int aarp_do_dump;
	extern int zone_do_dump;
	extern int stats_do_dump;

	rtmp_do_dump  = 1;
	nbp_do_dump   = 1;
	aarp_do_dump  = 1;
	zone_do_dump  = 1;
	stats_do_dump = 1;

	(void)signal(SIGQUIT, uar_dump_sig);

	return;
}

/*
 * display packet content in hex and ascii
 *
 */

void
dumppacket(pkt, len)
u_char *pkt;
int len;
{
	int i, j;
	u_char line[16];

	for (i = 0, j = 0; i < len; i++) {
	  fprintf(stderr, "%02x ", (line[j++] = pkt[i]));
	  if ((i+1) % 16 == 0 || (i+1) == len) {
	    while (j < 16) {
	      fprintf(stderr, "   ");
	      line[j++] = ' ';
	    }
	    fprintf(stderr, "      ");
	    for (j = 0; j < 16; j++)
	      fprintf(stderr, "%c", (isprint(line[j])) ? line[j] : '.');
	    fprintf(stderr, "\n");
	    j = 0;
	  }
	}

	return;
}

/*
 * cleanup on termination
 *
 */

void
cleanup()
{
	int i;
	void log(), exit();
	void zip_multicast();

	for (i = 0; i < numintfc; i++) {
	  if (iflist[i].itype == ITYPE_ETHERTALK) {
	    if (iflist[i].phase == PHASE2) {
	      if (iflist[i].etfd >= 0)
	        eth_delmulti(iflist[i].etfd, iflist[i].if_name, maddr);
#ifndef SINGLE_MCAST
	      if (iflist[i].aarpfd >= 0)
	        eth_delmulti(iflist[i].aarpfd, iflist[i].if_name, maddr);
#endif /* SINGLE_MCAST */
	      if (iflist[i].net_lo != 0) /* interface already seeded */
	        zip_multicast(i, DELMULTI);
	    }
	  }
	}

	log("UAR exiting");

	exit(0);

	/*NOTREACHED*/
}

/*
 * write message to the logfile
 *
 */

void
log(msg)
char *msg;
{
	FILE *log, *fopen();
	struct tm *tm, *localtime();

	if (debug)
	  fprintf(stderr, " UAR: %s\n", msg);

	if (logfile == NULL)
	  return;

	if ((log = fopen(logfile, "a")) == NULL) {
	  incr(uarBadDumpFile);
	  return;
	}

	tm = localtime(&tn.tv_sec);

	fprintf(log, "%2d/%02d/%02d %02d:%02d:%02d ", tm->tm_year,
	  tm->tm_mon+1,tm->tm_mday,tm->tm_hour,tm->tm_min,tm->tm_sec);

	fprintf(log, "%s\n", msg);

	(void)fclose(log);

	return;
}

/*
 * register NBP name of interface
 * and any optional NBP services
 *
 */

void
uar_register(ifn)
{
	struct ataddr ataddr;
	void nbp_register();

	ataddr.net = iflist[ifn].net;
	ataddr.node = iflist[ifn].node;
	ataddr.skt = AEP;

	/*
	 * interface name
	 *
	 */
	nbp_register(ifn, iflist[ifn].if_nbpname,
	  NBPTYPE, NBPZONE, &ataddr, 8, 5);

	/*
	 * Time server, if specified
	 *
	 */
	if (iflist[ifn].if_timesrvr[0] != '\0') {
	  ataddr.skt = TMLDSKT;
	  nbp_register(ifn, iflist[ifn].if_timesrvr,
	    TMLDTYPE, NBPZONE, &ataddr, 8, 5);
	}

	return;
}

/*
 * read the UAR configuration file
 *
 */

void
configure(filename)
char *filename;
{
	int peer;
	u_char node;
	FILE *r, *fopen();
	int ifn, phase, zlen, zones;
	u_short net, net_lo, net_hi, atnetshort();
	char interface[64], name[64], value[64], line[256], znam[NBPSTR], *p;
	struct zone *zone, *zonelist[258], *zone_make();
	unsigned long inet_addr();
	void cap_initialize();
	void log(), exit();
	void do_config();
	void strcpycp();
	void getfield();

	if ((r = fopen(filename, "r")) == NULL) {
	  fprintf(stderr, " UAR: No configuration file (%s)\n", filename);
	  fprintf(stderr, " UAR: Running as non-seed appletalk router.\n");
	  log("No configuration file, running as non-seed router");
	  return;
	}
	
	ifn = -1;
	znam[0] = '\0';
	phase = node = 0;
	net = net_lo = net_hi = 0;

	zones = 0;
	zonelist[0] = ZONENULL;

	while (fgets(line, sizeof(line), r) != NULL) {
	  if (line[0] == '#' || line[0] == '\n')
	    continue;
	  if ((p = (char *)index(line, '\n')) != NULL)
	    *p = '\0';
	  linep = line;

	  getfield(name, sizeof(name), 0);
	  getfield(value, sizeof(value), 0);

	  if (strcmp(name, "interface") == 0) {
	    if (ifn >= 0)
	      do_config(ifn, net_lo, net_hi, net, znam, zonelist);

	    net = net_lo = net_hi = 0;
	    strcpy(interface, value);
	    zonelist[0] = ZONENULL;
	    znam[0] = '\0';
	    zones = 0;

	    for (ifn = 0; ifn < numintfc; ifn++)
	      if (strcmp(value, iflist[ifn].if_name) == 0)
	        break;

	    if (ifn == numintfc)
	      ifn = -1;

	    continue;
	  }

	  if (strcmp(name, "network") == 0) {
	    net = atnetshort(value);
	    continue;
	  }

	  if (strcmp(name, "networklo") == 0) {
	    net_lo = atnetshort(value);
	    continue;
	  }

	  if (strcmp(name, "networkhi") == 0) {
	    net_hi = atnetshort(value);
	    continue;
	  }

	  /*
	   * default zone for network
	   *
	   */
	  if (strcmp(name, "zone") == 0) {
	    if ((zlen = strlen(value)) > MAXNBP) {
	      fprintf(stderr, " UAR: Zonename \"%s\" too long (%d)\n",
	        value, zlen);
	      exit(1);
	    }
	    bcopy((char *)value, (char *)znam, zlen);
	    znam[zlen] = '\0';
	    continue;
	  }

	  /*
	   * zone name for interface services, ie: CAP
	   *
	   */
	  if (strcmp(name, "zonename") == 0) {
	    if (ifn >= 0)
	      iflist[ifn].ourzone = zone_make((u_char *)value);
	    continue;
	  }

	  /*
	   * Phase 2 zone list
	   *
	   */
	  if (strcmp(name, "zonelist") == 0) {
	    if (ifn >= 0) {
	      if ((zone = zone_make((u_char *)value)) != ZONENULL) {
	        if (zones < 256) {
	          zonelist[zones++] = zone;
	          zonelist[zones] = ZONENULL;
	        } else {
	          fprintf(stderr, " UAR: %-5s Too many zones in zonelist\n",
	            iflist[ifn].if_name);
	          exit(1);
	        }
	      }
	    }
	    continue;
	  }

	  if (ifn >= 0) {
	    if (strcmp(name, "node") == 0) {
	      node = atoi(value) & 0xff;
	      if (node > 0 && node < 0xff)
	        iflist[ifn].node = node;
	      continue;
	    }

	    if (strcmp(name, "phase") == 0) {
	      phase = atoi(value);
	      if (phase == 1 || phase == 2)
	        iflist[ifn].phase = atoi(value);
	      continue;
	    }

	    if (strcmp(name, "timesrvr") == 0) {
	      strcpycp(value, iflist[ifn].if_timesrvr);
	      continue;
	    }

	    if (strcmp(name, "cap") == 0) {
	      if (strcmp(interface, "tnnl") == 0) {
	        fprintf(stderr, " UAR: cannot attach CAP to %s interface\n",
	          interface);
	        exit(1);
	      }
	      if (strcmp(value, "on") == 0)
	        cap_initialize(ifn);
	      continue;
	    }

	    if (strcmp(name, "peer") == 0) {
	      if (strcmp(interface, "tnnl") == 0) {
	        if ((p = (char *)index(value, '@')) == NULL) {
	          fprintf(stderr,
	            " UAR: bad format peer entry \"%s %s\" in %s\n",
	            name, value, filename);
	          exit(1);
	        }
	        *p++ = '\0';
	        peer = atoi(value);
	        if (peer < 0 || peer > 253) {
	          fprintf(stderr,
	            " UAR: tnnl peer nodes must be in the range 1-253\n");
	          exit(1);
	        }

	        /* skip white space */
	        while (isspace(*p) && *p != '\0')
	          *p++ = '\0';

	        if (!isdigit(*p)) {
	          fprintf(stderr,
	            " UAR: tnnl peer addresses must be numeric IP numbers\n");
	          exit(1);
	        }

	        iflist[ifn].mnode[peer].client = MNODE_TNNL;
	        iflist[ifn].mnode[peer].state = MNODE_VALID;
	        iflist[ifn].mnode[peer].addr = inet_addr(p);

	        continue;
	      }
	    }

	    fprintf(stderr, " UAR: I don't understand \"%s %s\" in %s\n",
	      name, value, filename);

	    exit(1);
	  }
	}

	(void)fclose(r);

	if (ifn >= 0)
	  do_config(ifn, net_lo, net_hi, net, znam, zonelist);

	return;
}

/*
 * Configure the interface 'ifn'
 *
 */

void
do_config(ifn, net_lo, net_hi, net, znam, zonelist)
int ifn;
u_short net_lo, net_hi, net;
char *znam;
struct zone *zonelist[];
{
	int i = 0;
	struct rtmp *rtmp, *rtmp_route();
	void rtmp_seed(), zone_rtmp_add();
	void tnnl_read();
	int tnnl_write();
	int tnnl_open();
	void exit();

	if (net_lo != 0 && net_hi != 0 && znam[0] != '\0') {
	  if (net_lo > net_hi) {
	    fprintf(stderr, " UAR: %-5s Bogus network range (lo > hi)\n",
	      iflist[ifn].if_name);
	    exit(1);
	  }
	  if (iflist[ifn].phase == PHASE1 && net_lo != net_hi) {
	    fprintf(stderr, " UAR: %-5s Network range illegal for Phase 1\n",
	      iflist[ifn].if_name);
	    exit(1);
	  }
	  rtmp_seed(ifn, net_lo, net_hi, (u_char *)znam);
	  if (net != 0) {
	    if (net >= net_lo && net <= net_hi)
	      iflist[ifn].net = net;
	    else {
	      fprintf(stderr, " UAR: %-5s Net number not within lo/hi range\n",
	        iflist[ifn].if_name);
	      exit(1);
	    }
	  } else net = net_lo;
	} else {
	  if (net == 0 || znam[0] == '\0') {
	    fprintf(stderr, " UAR: %-5s Configuration data incomplete\n",
	      iflist[ifn].if_name);
	    exit(1);
	  }
	  rtmp_seed(ifn, net, net, (u_char *)znam);
	}

	if (zonelist[0] != ZONENULL) {
	  if (iflist[ifn].phase != PHASE2) {
	    fprintf(stderr, " UAR: %-5s Zonelist entries illegal for Phase 1\n",
	      iflist[ifn].if_name);
	    exit(1);
	  }
	  if ((rtmp = rtmp_route(net)) != RTMPNULL) {
	    while (zonelist[i] != ZONENULL) {
	      if (debug & ZIPDEBUG)
	      fprintf(stderr, " UAR: %-5s Adding Zone Name      \"%s\"\n",
	        iflist[ifn].if_name, zonelist[i]->name+1);
	      zone_rtmp_add(zonelist[i++], rtmp);
	    }
	  }
	}

	/*
	 * handle special internal interfaces
	 *
	 */
	if (strcmp(iflist[ifn].if_name, "arns") == 0) {
	  iflist[ifn].itype = ITYPE_ARNS;
	  iflist[ifn].state = ADDR_VALID;
	  iflist[ifn].iopen = NULL;
	  iflist[ifn].write = NULL;
	  iflist[ifn].reade = NULL;
	  iflist[ifn].reada = NULL;
	  iflist[ifn].rtmps = NULL;
	}
	if (strcmp(iflist[ifn].if_name, "tnnl") == 0) {
	  iflist[ifn].itype = ITYPE_TNNL;
	  iflist[ifn].state = ADDR_VALID;
	  iflist[ifn].iopen = tnnl_open;
	  iflist[ifn].write = tnnl_write;
	  iflist[ifn].reade = tnnl_read;
	  iflist[ifn].reada = NULL;
	  iflist[ifn].rtmps = NULL;
	}

	return;
}

/*
 * get a short network number from string str
 *
 */

u_short
atnetshort(str)
register char *str;
{
	register char *cp;

	if (*str == '\0')
	  return(0);

	if ((cp = (char *)index(str, '.')) == 0)
	  return(atoi(str));

	*cp++ = 0;

	return((atoi(str) << 8) | atoi(cp));
}

/*
 * Get next field from 'line' buffer into 'str'.  'linep' is the 
 * pointer to current position.
 *
 * Fields are white space separated, except within quoted strings.
 * If 'quote' is true the quotes of such a string are retained, otherwise
 * they are stripped.  Quotes are included in strings by doubling them or
 * escaping with '\'.
 *
 */

void
getfield(str, len, quote)
char *str;
int len, quote;
{
	register char *lp = linep;
	register char *cp = str;

	while (*lp == ' ' || *lp == '\t')
	  lp++;	/* skip spaces/tabs */

	if (*lp == 0 || *lp == '#') {
	  *cp = 0;
	  return;
	}
	len--;	/* save a spot for a null */

	if (*lp == '"' || *lp == '\'') {		/* quoted string */
	  register term = *lp;

	  if (quote) {
	    *cp++ = term;
	    len -= 2;	/* one for now, one for later */
	  }
	  lp++;
	  while (*lp) {
	    if (*lp == term) {
	      if (lp[1] == term)
	        lp++;
	      else
	        break;
	    }
	    /* check for \', \", \\ only */
	    if (*lp == '\\'
	      && (lp[1] == '\'' || lp[1] == '"' || lp[1] == '\\'))
	      lp++;
	    *cp++ = *lp++;
	    if (--len <= 0) {
	      fprintf(stderr, "string truncated: %s", str);
	      if (quote)
	        *cp++ = term;
	      *cp = 0;
	      linep = lp;
	      return;
	    }
	  }
	  if (!*lp)
	    fprintf(stderr,"unterminated string: %s", str);
	  else {
	    lp++;	/* skip the terminator */

	    if (*lp && *lp != ' ' && *lp != '\t' && *lp != '#') {
	      fprintf(stderr, "garbage after string: %s", str);
	      while (*lp && *lp != ' ' && *lp != '\t' && *lp != '#')
	        lp++;
	    }
	  }
	  if (quote)
	    *cp++ = term;
	} else {
	  while (*lp && *lp != ' ' && *lp != '\t' && *lp != '#') {
	    *cp++ = *lp++;
	    if (--len <= 0) {
	      fprintf(stderr, "string truncated: %s", str);
	      break;
	    }
	  }
	}
	*cp = 0;
	linep = lp;

	return;
}
