/*
 * UAR - UNIX AppleTalk Router
 * 
 * Copyright (c) 1993, The University of Melbourne.
 *
 * All Rights Reserved.
 *
 * UAR may NOT be publicly redistributed (for example via
 * anonymous FTP), sold, or the source used for any other
 * purpose without the permission of the copyright owner.
 * 
 * This software is supplied "as is" without express or
 * implied warranty.
 * 
 * djh@munnari.OZ.AU
 *
 * $Author: djh $
 * $Revision: 1.1.1.7 $
 *
 */

#include <stdio.h>
#include <ctype.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#if defined (ultrix) || defined (AIX)
#include <sys/file.h>
#else	/* ultrix || AIX */
#include <sys/fcntl.h>
#endif	/* ultrix || AIX */
#ifdef	AIX
#include <sys/select.h>
#endif	/* AIX */
#include <sys/ioctl.h>
#include <netinet/in.h>
#if (!(defined(sony_news) || defined(__386BSD__)))
#include <netinet/tcp.h>
#endif	/* sony_news || __386BSD__ */
#include <signal.h>
#include <netdb.h>
#include <time.h>
#include <pwd.h>

/* map compatible functions */

#ifdef	SOLARIS
#define MAPFUNCS
#endif	/* SOLARIS */

#ifdef	MAPFUNCS
#define bcopy(a,b,s)	memcpy((char *)(b),(char *)(a),(s))
#define bcmp(a,b,s)	memcmp((char *)(a),(char *)(b),(s))
#define bzero(a,s)	memset((char *)(a),0,(s))
#define rindex(s,c)	strrchr((char *)(s),(c))
#define index(s,c)	strchr((char *)(s),(c))
#endif	/* MAPFUNCS */

#ifdef	hpux
#define POSIX
#define SINGLE_MCAST
#endif	/* hpux */

/* misc numbers		*/

#define MAXIF		10		/* maximum number of interfaces	*/
#define MAXNODES	256		/* maxm. nodes per interface	*/
#define MAXHOST		64		/* maximum host name length	*/

#define SKTBUFSIZ	32768		/* SO_RCVBUF buffer size	*/
#define HDRSIZ		24		/* room for Phase 2 header	*/
#define BUFSIZE		4096+HDRSIZ	/* main receive buffer size	*/
#define NBPSTR		34		/* NBP string lengths		*/
#define MAXNBP		32
#define MAXDDPLEN	586

#define INFOTimer	5		/* server ZIP/Info proc (1 sec)	*/
#define NBPTimer	5		/* server NBP process (1 sec)	*/
#define RTMPTimer	50		/* RTMP sendRTMP timer (10 sec)	*/
#define RTMPValTimer	100		/* RTMP validity timer (20 sec)	*/
#define ZIPTimer	RTMPTimer	/* server ZIP process (10 sec)	*/

#define AARPCount	10		/* # aarp packets to send	*/
#define NBPCACHETIMEOUT	300		/* lives in local cache (5 min)	*/

#define ADDMULTI	0
#define DELMULTI	1

#define AARPTABSIZ	256
#define RTMPTABSIZ	256
#define RTMPTABMSK	RTMPTABSIZ-1

#define SDDPHDRSIZ	5
#define LDDPHDRSIZ	13
#define NBPHDRSIZ	2
#define ENTHDRSIZ	5
#define ZIPQHDRSIZ	2
#define ZIPATPHDRSIZ	8

#define CAP_PORT	903
#define TNNL_PORT	9115

#define P2STARTUP	65280
#define P2ENDSTUP	65534

#define AARPReq		0x01		/* Address Resolution Request	*/
#define AARPResp	0x02		/* Address Resolution Response	*/
#define AARPProbe	0x03		/* Address Resolution Probe	*/

#define NBPBRRQ		0x01		/* NBP bridge request packet	*/
#define NBPLKUPREQ	0x02		/* NBP lookup request packet	*/
#define NBPLKUPREPLY	0x03		/* NBP lookup reply packet	*/
#define NBPFWDREQ	0x04		/* NBP forward request		*/

#define nbpEquals	'='
#define nbpApprox	0xc5

#define ECHOMSG		"[Are You There ?]"
#define NETINFOZONE	"UAR GetNetInfo Zone"
#define TMLDTYPE	"\010Timelord"
#define NBPTYPE		"\003UAR"
#define NBPZONE		"\001*"
#define TMLDSKT		253

#define PHASE1		0x01
#define PHASE2		0x02

#define SDDP		0x01
#define LDDP		0x02

#define BCAST		0xff

#define RTMP		0x01
#define NBP		0x02
#define ATP		0x03
#define AEP		0x04
#define RTMPR		0x05
#define ZIP		0x06
#define ADSP		0x07
#define SNMP		0x08

#define TREQ		0x40
#define TRESP		0x80
#define TREL		0xc0
#define FNMSK		0xc0

#define XO		0x20
#define EOM		0x10
#define STS		0x08

#define TREL030		0x00
#define TREL060		0x01
#define TREL120		0x02
#define TREL240		0x03
#define TREL480		0x04
#define TRELMSK		0x07

#define IDLEN		0x08

#define ZIPQUERY	0x01
#define ZIPREPLY	0x02
#define ZIPTAKEDOWN	0x03
#define ZIPBRINGUP	0x04
#define ZIPINFOREQ	0x05
#define ZIPINFOREP	0x06
#define ZIPNOTIFY	0x07
#define ZIPEREPLY	0x08

#define ZIPGETMYZONE	0x07
#define ZIPGETZONELIST	0x08
#define ZIPGETLOCALZON	0x09

#define MAXZIPQUERY	255
#define MAXZIPREPLYLEN	550

#define TMLDOFFSET	0x7c25b080	/* Mac Time 00:00:00 GMT Jan 1, 1970 */
#define TMLDGETTIME	0
#define TMLDENOPERM	10
#define TMLDENONE	12

#define UAR_CONF	1
#define UAR_CONF_REQ	1
#define UAR_CONF_DATA	2

/* UAR debug flags	*/

#define RTMPDEBUG	0x0001
#define NBPDEBUG	0x0002
#define ECHODEBUG	0x0004
#define ZIPDEBUG	0x0008
#define IPDEBUG		0x0010
#define ETHDEBUG	0x0020
#define ROUTEDEBUG	0x0040
#define PKTDEBUG	0x0080
#define TIMERDEBUG	0x0100
#define AARPDEBUG	0x0200
#define CAPDEBUG	0x0400
#define ARNSDEBUG	0x0800
#define TNNLDEBUG	0x1000
#define TMLDDEBUG	0x2000

/* structure defns	*/

/*
 * multinode/special interface support
 *
 */

struct mnode {
	u_char client;
#define MNODE_CAP	0x01
#define MNODE_ARNS	0x80
#define MNODE_TNNL	0x81
	u_char state;
#define MNODE_UNUSED	0
#define MNODE_ROUTER	1
#define MNODE_PENDING	2
#define MNODE_VALID	3
	u_long addr;
};

#define MNODE_MIN	MNODE_CAP
#define MNODE_MAX	MNODE_CAP

/*
 * structure per interface
 *
 */

struct iflist {
	int etfd;
	int aarpfd;
	u_short net;
	u_short net_lo;
	u_short net_hi;
	u_char node;
	u_char hint;
	u_char state;
#define ADDR_PENDING	0
#define ADDR_VALID	1
	u_char itype;
#define ITYPE_ETHERTALK	0
#define ITYPE_ARNS	1
#define ITYPE_TNNL	2
	short phase;
	short count;
	u_char eaddr[6];
	char *if_name;
	char *if_nbpname[NBPSTR];	/* NBP name for UAR interface	*/
	char *if_timesrvr[NBPSTR];	/* NBP name for Time Server	*/
	struct zone *defzone;		/* default zone for interface	*/
	struct zone *ourzone;		/* default zone for services	*/
	struct mnode mnode[MAXNODES];	/* multinodes for interface	*/
	int clients;			/* services supported on ifc	*/
	int (*iopen)();			/* interface open routine	*/
	int (*write)();			/* interface write routine	*/
	void (*reade)();		/* ethertalk read routine	*/
	void (*reada)();		/* aarp read routine		*/
	void (*rtmps)();		/* rtmp send routine		*/
};

/*
 * packet descriptor
 *
 */

struct pd {
	u_char lapType;
	u_char ddpType;
	u_short dstNet;
	u_short srcNet;
	u_char dstNode;
	u_char srcNode;
	u_char dstSkt;
	u_char srcSkt;
	short dataLen;
	u_char *data;
	u_char *addr;
	short hop;
};

/*
 * Read Data Structure
 *
 */

#define NUMRDS	32

struct RDS {
	u_short dataLen;
	u_char *dataPtr;
};

/*
 * AppleTalk address
 *
 */

struct ataddr {
	u_short net;
	u_char node;
	u_char skt;
};

/*
 * NET list per zone name
 *
 */

struct net {
	u_short net_lo;			/* net number entry		*/
	u_short net_hi;			/* net number entry		*/
	struct net *next;		/* next in linked list		*/
};

/*
 * ZIP zone list
 *
 */

struct zone {
	u_char name[NBPSTR];		/* zone name entry (pascal str)	*/
	struct net *nets;		/* list of nets for this zone	*/
	struct zone *next;		/* next in linked list		*/
};

/*
 * ZIP zone table
 *
 */

struct zit {
	struct zone *zone;
	struct zit *next;
};

/*
 * NBP queue
 *
 */

struct nbpq {
	int intfc;
	short count;
	short timer;
	short period;
	short zonlen;
	struct pd opd;
	u_char data[160];
	short dataLen;
	u_char *modifier;
	struct nbpq *next;
};

/*
 * NBP entity list ordered by objects
 *
 */

struct nbpobj {
	u_char nbpent[ENTHDRSIZ];
	u_char nbpobj[NBPSTR];
	u_char zone[NBPSTR];
	time_t when;
	int intfc;
	struct nbpobj *next;
};

/*
 * NBP cache is a list sorted by type
 *
 */

struct nbptyp {
	u_char nbptyp[NBPSTR];
	struct nbpobj *list;
	struct nbptyp *next;
};

/*
 * RTMP table
 *
 */

struct rtmp {
	u_short net_lo;			/* network range start		*/
	u_short net_hi;			/* network range end		*/
	u_short dist;			/* distance to destination	*/
	u_short rnet;			/* router network address	*/
	int intfc;			/* interface to send on		*/
	u_char rnode;			/* router node address		*/
	u_char state;			/* entry state			*/
#define RTMP_GOOD	0
#define RTMP_SUSP	1
#define RTMP_BAD1	2
#define RTMP_BAD2	3
	u_char flags;			/* flag bits for this entry	*/
#define RTMP_EXTENDED	0x01
	u_char dummy;			/* padding to longword		*/
	u_short zquantity;		/* number of zones in list	*/
	u_short zexpected;		/* expected zones for network	*/
	struct zit *zitlist;		/* zone list per network	*/
	struct rtmp *next;		/* next in linked list		*/
};

/*
 * AARP Table
 *
 */

struct aarp {
	u_short net;			/* network number		*/
	u_char node;			/* node number			*/
	u_char eaddr[6];		/* hardware address		*/
	struct aarp *next;		/* next in linked list		*/
};

struct aarpq {
	u_short net;			/* network number		*/
	u_char node;			/* node number			*/
	u_short function;		/* probe or lookup		*/
	int intfc;			/* interface to send on		*/
	short count;			/* count of packets sent	*/
	struct aarpq *next;		/* next in linked list		*/
};

/*
 * statistics collection
 *
 */

#define incr(x) uar_stats[x]++
#define decr(x) uar_stats[x]--
#define drop(x) \
	{ \
	  if (debug) fprintf(stderr, "DROP: %s\n", stats_msg[x]); \
	  uar_stats[x]++; \
	  pkts_dropped++; \
	  return; \
	}

/*
 * UAR specific counters
 *
 */

#define uarBadEtherType			0
#define uarBadLapType			1
#define uarNoSocketHandler		2
#define uarBadDDPType			4
#define uarIgnoreSelfSend		5
#define uarAARPCacheFail		6
#define uarRTMPCacheFail		7
#define uarMemoryError			8
#define uarUnknownPhase			9
#define uarUnknownZone			10
#define uarNoDefaultZone		11
#define uarBadDumpFile			12
#define uarBadPacketHdr			13
#define uarCAPPacketsOut		14
#define uarCAPPacketsIn			15
#define uarMNodePacketsOut		16
#define uarMNodePacketsIn		17
#define uarAARPPacketsOut		18
#define uarAARPPacketsIn		19

/*
 * AppleTalk MIB Counters (per gateway)
 *
 */

#define aarpLookups			20
#define aarpHits			21
#define ddpOutRequests			22
#define ddpOutShorts			23
#define ddpOutLongs			24
#define ddpInReceives			25
#define ddpInLocalDatagrams		26
#define ddpNoProtocolHandlers		27
#define ddpTooShortErrors		28
#define ddpTooLongErrors		29
#define ddpShortDDPErrors		30
#define ddpChecksumErrors		31
#define ddpForwRequests			32
#define ddpOutNoRoutes			33
#define ddpBroadcastErrors		34
#define ddpHopCountErrors		35
#define rtmpInDataPkts			36
#define rtmpOutDataPkts			37
#define rtmpInRequestPkts		38
#define rtmpNextIREqualChanges		39
#define rtmpNextIRLessChanges		40
#define rtmpRouteDeletes		41
#define rtmpRoutingTableOverflows	42
#define rtmpOutRequestPkts		43
#define rtmpInVersionMismatches		44
#define rtmpInErrors			45
#define zipInZipQueries			46
#define zipInZipReplies			47
#define zipInZipExtendedReplies		48
#define zipZoneConflictErrors		49
#define zipInObsoletes			50
#define zipInErrors			51
#define nbpInLookUpRequests		52
#define nbpInLookUpReplies		53
#define nbpInBroadcastRequests		54
#define nbpInForwardRequests		55
#define nbpOutLookUpReplies		56
#define nbpRegistrationFailures		57
#define nbpInErrors			58
#define atechoRequests			59
#define atechoReplies			60
#define atechoOutRequests		61
#define atechoInReplies			62
#define atpInPkts			63
#define atpOutPkts			64

#define uarRTMPRoutes			65
#define uarRTMPSearches			66

#define uarTNNLPacketsIn		67
#define uarTNNLPacketsOut		68
#define uarTNNLNotForMe			69
#define uarTNNLBogusSrc			70

#define uarZONEFinds			71
#define uarZONESearches			72

#define maxCounters			73

/*
 * AppleTalk MIB Counters (per interface)
 * - later
 *
 */

/*
 * linked list terminators
 *
 */

#define ZITNULL (struct zit *)NULL
#define NETNULL (struct net *)NULL
#define RTMPNULL (struct rtmp *)NULL
#define ZONENULL (struct zone *)NULL
#define AARPNULL (struct aarp *)NULL
#define AARPQNULL (struct aarpq *)NULL
#define NBPQNULL (struct nbpq *)NULL
#define NBPOBJNULL (struct nbpobj *)NULL
#define NBPTYPNULL (struct nbptyp *)NULL

/*
 * strings
 *
 */

#define ETH_THIS_NET	"thisNet"
#define ETH_THIS_NODE	"thisNode"
#define ETH_THIS_ZONE	"thisZone"
#define ETH_BRIDGE_NET	"bridgeNet"
#define ETH_BRIDGE_NODE	"bridgeNode"
#define ETH_BRIDGE_IP	"bridgeIP"
#define ETH_NIS_NET	"nisNet"
#define ETH_NIS_NODE	"nisNode"
#define ETH_ASYNC_NET	"asyncNet"
#define ETH_ASYNC_ZONE	"asyncZone"
#define ETH_INTERFACE	"interface"
#define ETH_NET_START	"netRangeStart"
#define ETH_NET_END	"netRangeEnd"

#ifndef ETH_DATABASE
#define ETH_DATABASE	"/usr/local/cap/etc/etalk.local"
#endif  /* ETH_DATABASE */

#ifndef CONFIG_FILE
#define CONFIG_FILE	"/usr/local/cap/lib/uar.conf"
#endif  /* CONFIG_FILE */

#ifndef DUMP_AARP
#define DUMP_AARP	"/usr/tmp/uar.aarp"
#endif  /* DUMP_AARP */

#ifndef DUMP_RTMP
#define DUMP_RTMP	"/usr/tmp/uar.rtmp"
#endif  /* DUMP_RTMP */

#ifndef DUMP_NBP
#define DUMP_NBP	"/usr/tmp/uar.nbp"
#endif  /* DUMP_NBP */

#ifndef DUMP_ZONE
#define DUMP_ZONE	"/usr/tmp/uar.zone"
#endif  /* DUMP_ZONE */

#ifndef DUMP_STAT
#define DUMP_STAT	"/usr/tmp/uar.stats"
#endif  /* DUMP_STAT */

#ifndef LOGFILE
#define LOGFILE		"/usr/tmp/uar.log"
#endif  /* LOGFILE */
