/*
 * UAR - UNIX AppleTalk Router
 * 
 * Copyright (c) 1993, The University of Melbourne.
 *
 * All Rights Reserved.
 *
 * UAR may NOT be publicly redistributed (for example via
 * anonymous FTP), sold, or the source used for any other
 * purpose without the permission of the copyright owner.
 * 
 * This software is supplied "as is" without express or
 * implied warranty.
 * 
 * djh@munnari.OZ.AU
 *
 * $Author: djh $
 * $Revision: 1.1.1.6 $
 *
 */

#include "uar.h"

extern int debug;
extern u_long uar_stats[];
extern u_long pkts_dropped;
extern struct iflist iflist[];
extern char *stats_msg[];

int zone_do_dump = 0;

struct zone *zoneMRU;
struct zone *zoneTab;			/* Ordered zone list	*/

static u_char zbuf[632];		/* ZIP Query/Reply data	*/
static u_char zrnode = 0;		/* ZIP Query address	*/
static u_short zrnet = 0;		/* ZIP Query address	*/
static short zcount = 0;		/* ZIP Query net count	*/
static short zoffset = 0;		/* ZIP Reply offset	*/
static int zintfc = -1;			/* ZIP Query interface	*/

/*
 * Initialize ZIP data structures
 *
 * zoneTab is a sorted list of individual zone names. Each rtmpTab
 * entry has a linked list of pointers to each of the zoneTab zone
 * names applicable to that network. Each zoneTab entry also has a
 * linked list of all of the network numbers belonging to that zone.
 * These lists are maintained by the ZIP process.
 *
 */

void
zip_initialize()
{
	zoneTab = ZONENULL;
	zoneMRU = ZONENULL;

	return;
}

/*
 * Check zones list for each network in the RTMP table. If no entries
 * yet or less zones than expected then build and send a ZIP query.
 *
 */

void
zip_process()
{
	int i;
	struct zit *zit;
	struct rtmp *rtmp;
	extern int numintfc;
	extern struct rtmp *rtmpTab[];
	void zip_net_list(), zone_getnetinfo();
	void zip_query_build(), zip_query_send();

	zcount = zrnet = zrnode = 0;
	zintfc = -1;

	for (i = 0; i < RTMPTABSIZ; i++) {
	  rtmp = rtmpTab[i];
	  while (rtmp != RTMPNULL) {
	    rtmp->zquantity = 0;
	    zit = rtmp->zitlist;
	    while (zit != ZITNULL) {
	      zip_net_list(rtmp, zit);
	      rtmp->zquantity++;
	      zit = zit->next;
	    }
	    if (rtmp->zitlist == ZITNULL
	     || rtmp->zquantity < rtmp->zexpected)
	      zip_query_build(rtmp);
	    rtmp = rtmp->next;
	  }
	}
	zip_query_send();

	for (i = 0; i < numintfc; i++)
	  if (iflist[i].phase == PHASE2)
	    if (iflist[i].defzone == ZONENULL)
	      zone_getnetinfo(i);
	  
	return;
}

/*
 * build a ZIP query packet. If a new entry is destined for a different
 * "address" or is to be sent on a different interface or the packet is
 * full, send it.
 *
 */

void
zip_query_build(rtmp)
struct rtmp *rtmp;
{
	u_char *pkt;
	void zip_query_send();

	if (rtmp->rnet  != zrnet
	 || rtmp->rnode != zrnode
	 || rtmp->intfc != zintfc)
	  zip_query_send();

	zrnet  = rtmp->rnet;
	zrnode = rtmp->rnode;
	zintfc = rtmp->intfc;

	pkt = zbuf+HDRSIZ+LDDPHDRSIZ+ZIPQHDRSIZ+(zcount<<1);

	pkt[0] = rtmp->net_lo >> 8;
	pkt[1] = rtmp->net_lo & 0xff;

	if (++zcount >= MAXZIPQUERY)
	  zip_query_send();

	return;
}

/*
 * send the accumulated ZIP query data
 *
 */

void
zip_query_send()
{
	short len;
	u_char *pkt;
	struct pd opd;
	struct rtmp *rtmp_route();
	void ddp_route();

	if (zintfc == -1)
	  return;

	if (zcount == 0)
	  return;

	pkt = zbuf+HDRSIZ+LDDPHDRSIZ;

	/* ZIP Query Data */

	pkt[0] = ZIPQUERY;		/* ZIP Query		*/
	pkt[1] = zcount & 0xff;		/* ZIP Network Count	*/

	len = ZIPQHDRSIZ+(zcount<<1);	/* ZIP Query pkt length	*/

	opd.lapType = LDDP;
	opd.ddpType = ZIP;
	opd.dstNet  = (zrnet == 0)  ? iflist[zintfc].net : zrnet;
	opd.srcNet  = iflist[zintfc].net;
	opd.dstNode = (zrnode == 0) ? BCAST : zrnode;
	opd.srcNode = iflist[zintfc].node;
	opd.dstSkt  = ZIP;
	opd.srcSkt  = ZIP;
	opd.dataLen = len;
	opd.data = pkt;
	opd.addr = NULL;
	opd.hop  = 0;

	if (debug & ZIPDEBUG)
	  fprintf(stderr, " ZIP: %-5s Sending Query (%d net%s)\n",
	    iflist[zintfc].if_name, zcount, (zcount == 1) ? "" : "s");

	incr(ddpOutRequests);

	ddp_route(zintfc, &opd, rtmp_route(opd.dstNet));

	zcount = zrnet = zrnode = 0;
	zintfc = -1;

	return;
}

/*
 * maintain the net list for each zone name entry
 *
 */

void
zip_net_list(rtmp, zit)
struct rtmp *rtmp;
struct zit *zit;
{
	struct zone *zone;
	struct net *p, *q, *net;

	if (zit == ZITNULL || rtmp == RTMPNULL)
	  return;

	if ((zone = zit->zone) == ZONENULL)
	  return;

	p = zone->nets;
	q = NETNULL;

	while (p != NETNULL) {
	  if (p->net_lo == rtmp->net_lo
	   && p->net_hi == rtmp->net_hi)
	    return;
	  if (p->net_lo >= rtmp->net_lo)
	    break;
	  q = p;
	  p = p->next;
	}

	if ((net = (struct net *)malloc(sizeof(struct net))) == NETNULL)
	  drop(uarMemoryError);

	net->net_lo = rtmp->net_lo;
	net->net_hi = rtmp->net_hi;

	net->next = p;

	if (q == NETNULL)
	  zone->nets = net;
	else
	  q->next = net;

	return;
}

/*
 * Incoming ZIP packet
 *
 */

void
zip(ifn, pd)
int ifn;
struct pd *pd;
{
	void zip_query(), zip_reply(), zip_notify();
	void zip_inforeq(), zip_inforep();

	switch (pd->data[0]) {
	  case ZIPQUERY:
	    incr(zipInZipQueries);
	    zip_query(ifn, pd);
	    break;
	  case ZIPREPLY:
	    incr(zipInZipReplies);
	    zip_reply(ifn, pd);
	    break;
	  case ZIPEREPLY:
	    incr(zipInZipExtendedReplies);
	    zip_reply(ifn, pd);
	    break;
	  case ZIPINFOREQ:
	    zip_inforeq(ifn, pd);
	    break;
	  case ZIPINFOREP:
	    zip_inforep(ifn, pd);
	    break;
	  case ZIPNOTIFY:
	    zip_notify(ifn, pd);
	    break;
	  case ZIPTAKEDOWN:
	  case ZIPBRINGUP:
	    drop(zipInObsoletes);
	    break;
	  default:
	    drop(zipInErrors);
	    break;
	}

	return;
}

/*
 * ZIP queries contain n network numbers for which zone lists
 * are being sought. These network numbers indicate the start
 * of the range associated with the desired networks.
 *
 * A router that receives a ZIP query should respond to the
 * requesting socket with a series of ZIP Replies indicating
 * the requested zones lists. Routers that are unaware of the
 * the requested zones list or do not yet have complete zones
 * lists should ignore the ZIP query and need not respond.
 *
 */

void
zip_query(ifn, pd)
int ifn;
struct pd *pd;
{
	u_short net;
	u_char *pkt;
	short len, netcount;
	struct rtmp *rtmp, *rtmp_route();
	void zip_reply_build(), zip_reply_send();

	pkt = pd->data;
	len = pd->dataLen;

	netcount = pkt[1];

	if (debug & ZIPDEBUG)
	  fprintf(stderr, " ZIP: %-5s Receive Query (%d net%s)\n",
	    iflist[ifn].if_name, netcount, (netcount == 1) ? "" : "s");

	if (len == 0)
	  drop(ddpTooShortErrors);
	  
	if (len > MAXDDPLEN)
	  drop(ddpTooLongErrors);

	if (pd->dstSkt != ZIP)
	  drop(uarNoSocketHandler);

	pkt += 2;
	len -= 2;

	zcount = 0;
	zoffset = ZIPQHDRSIZ;

	while (netcount > 0 && len > 0) {
	  net = (pkt[0] << 8) | pkt[1];
	  if ((rtmp = rtmp_route(net)) != RTMPNULL) {
	    if (rtmp->zquantity >= rtmp->zexpected)
	      zip_reply_build(ifn, pd, net, rtmp);
	  }
	  netcount--;
	  pkt += 2;
	  len -= 2;
	}

	zip_reply_send(ifn, pd, ZIPREPLY);

	return;
}

/*
 * ZIP replies consist of a series of network-number/zone-name
 * pairs, with zone names preceded by a length byte. The zones
 * list for a given network must be contiguous in the packet.
 *
 */

void
zip_reply_build(ifn, pd, net, rtmp)
int ifn;
u_short net;
struct pd *pd;
struct rtmp *rtmp;
{
	short zlen;
	u_char *data;
	struct zit *zit;
	void zip_reply_send();

	if ((zit = rtmp->zitlist) == ZITNULL)
	  return;

	data = zbuf+HDRSIZ+LDDPHDRSIZ+zoffset;

	if (rtmp->zquantity > 1) {
	/* multiple zones for 'net' */
	  if (zcount > 0) {
	  /* send already accumulated data */
	    zip_reply_send(ifn, pd, ZIPREPLY);
	    data = zbuf+HDRSIZ+LDDPHDRSIZ+zoffset;
	  }
	  while (zit != ZITNULL) {
	    data[0] = net >> 8;
	    data[1] = net & 0xff;
	    zcount = rtmp->zquantity;
	    zlen = zit->zone->name[0]+1;
	    bcopy((char *)zit->zone->name, (char *)(data+2), zlen);
	    zoffset += (zlen+2);
	    if (zoffset >= MAXZIPREPLYLEN)
	      zip_reply_send(ifn, pd, ZIPEREPLY);
	    data = zbuf+HDRSIZ+LDDPHDRSIZ+zoffset;
	    zit = zit->next;
	  }
	  zcount = rtmp->zquantity;
	  /* use extended reply for 'net' zones */
	  zip_reply_send(ifn, pd, ZIPEREPLY);
	  return;
	}

	data[0] = net >> 8;
	data[1] = net & 0xff;

	zlen = zit->zone->name[0]+1;
	bcopy((char *)zit->zone->name, (char *)(data+2), zlen);

	zoffset += (zlen+2);
	zcount++;

	if (zoffset >= MAXZIPREPLYLEN)
	  zip_reply_send(ifn, pd, ZIPREPLY);

	return;
}


/*
 * send the accumulated ZIP reply data
 *
 */

void
zip_reply_send(ifn, pd, function)
int ifn, function;
struct pd *pd;
{
	short len;
	u_char *pkt;
	struct pd opd;
	struct rtmp *rtmp_route();
	void ddp_route();

	if (zcount == 0)
	  return;

	if (zoffset == ZIPQHDRSIZ)
	  return;

	pkt = zbuf+HDRSIZ+LDDPHDRSIZ;

	/* ZIP Reply Data */

	pkt[0] = function;		/* ZIP Reply		*/
	pkt[1] = zcount & 0xff;		/* ZIP Network Count	*/

	len = zoffset;			/* ZIP Reply pkt length	*/

	opd.lapType = LDDP;
	opd.ddpType = ZIP;
	opd.dstNet  = pd->srcNet;
	opd.srcNet  = iflist[ifn].net;
	opd.dstNode = pd->srcNode;
	opd.srcNode = iflist[ifn].node;
	opd.dstSkt  = pd->srcSkt;
	opd.srcSkt  = ZIP;
	opd.dataLen = len;
	opd.data = pkt;
	opd.addr = NULL;
	opd.hop  = 0;

	if (debug & ZIPDEBUG)
	  fprintf(stderr, " ZIP: %-5s Sending Reply (%d net%s)\n",
	    iflist[ifn].if_name, zcount, (zcount == 1) ? "" : "s");

	incr(ddpOutRequests);

	ddp_route(ifn, &opd, rtmp_route(opd.dstNet));

	zcount = 0;
	zoffset = ZIPQHDRSIZ;

	return;
}

/*
 * ZIP replies consist of a series of network-number/zone-name
 * pairs, with zone names preceded by a length byte. The zones
 * list for a given network must be contiguous in the packet.
 *
 */

void
zip_reply(ifn, pd)
int ifn;
struct pd *pd;
{
	u_short net;
	short len, zlen, netcount;
	u_char *pkt, *data, *znam;
	struct rtmp *rtmp, *rtmp_route();
	struct zone *zone, *zone_find(), *zone_new();
	void zone_insert(), zone_rtmp_add();

	pkt = pd->data;
	len = pd->dataLen;

	netcount = pkt[1];

	if (debug & ZIPDEBUG)
	  fprintf(stderr, " ZIP: %-5s Receive %sReply (%d net%s)\n",
	    iflist[ifn].if_name, (pkt[0] == ZIPEREPLY) ? "Extended " : "",
	    netcount, (netcount == 1) ? "" : "s");

	if (len == 0)
	  drop(ddpTooShortErrors);
	  
	if (len > MAXDDPLEN)
	  drop(ddpTooLongErrors);

	if (pd->dstSkt != ZIP)
	  drop(uarNoSocketHandler);

	data = pkt+2;
	len -= 2;

	while (len > 0) {
	  net = (data[0] << 8) | data[1];
	  zlen = data[2];
	  znam = data+3;
#define MAXZONELEN 32
	  if (zlen > MAXZONELEN)
	    drop(zipInErrors);
	  if ((zone = zone_find(znam, zlen)) == ZONENULL) {
	    if ((zone = zone_new()) != ZONENULL) {
	      bcopy((char *)znam, (char *)(zone->name+1), zlen);
	      zone->name[zlen+1] = '\0';
	      zone->name[0] = zlen;
	      zone->nets = NETNULL;
	      zone->next = ZONENULL;
	      zone_insert(zone);
	    }
	  }
	  if (zone == ZONENULL)
	    drop(uarMemoryError);

	  if ((rtmp = rtmp_route(net)) == RTMPNULL)
	    drop(uarRTMPCacheFail);

	  if (pkt[0] == ZIPEREPLY)
	    rtmp->zexpected = pkt[1];

	  zone_rtmp_add(zone, rtmp);

	  data += (zlen+3);
	  len  -= (zlen+3);
	}

	return;
}

/*
 * add a zone name entry to the RTMP table zones list
 * (called from GetNetInfo reply, ZIP query reply and
 * interface configuration, although during configuration
 * the filter descriptors are not valid (not yet opened))
 *
 */

void
zone_rtmp_add(zone, rtmp)
struct zone *zone;
struct rtmp *rtmp;
{
	struct zit *zit, *p, *q;
	u_char *zaddr, *zone_mcast();

	p = rtmp->zitlist;
	q = ZITNULL;

	while (p != ZITNULL) {
	  if (p->zone == zone)
	    return;
	  q = p;
	  p = p->next;
	}

	if (rtmp->dist == 0 && iflist[rtmp->intfc].phase == PHASE2) {
	  if (iflist[rtmp->intfc].etfd >= 0) {
	    if ((zaddr = zone_mcast(zone->name+1,(short)zone->name[0]))!=NULL) {
	      eth_addmulti(iflist[rtmp->intfc].etfd, 
	        iflist[rtmp->intfc].if_name, zaddr);
	      if (debug & ZIPDEBUG) {
	        fprintf(stderr, " ZIP: %-5s Adding Zone Multicast",
	          iflist[rtmp->intfc].if_name);
	        fprintf(stderr, " %02x:%02x:%02x:%02x:%02x:%02x  \"%s\"\n",
	          zaddr[0], zaddr[1], zaddr[2], zaddr[3], zaddr[4], zaddr[5],
	          zone->name+1);
	      }
	    }
	  }
	}

	if ((zit = (struct zit *)malloc(sizeof(struct zit))) == ZITNULL)
	  drop(uarMemoryError);

	zit->zone = zone;
	zit->next = ZITNULL;

	if (q == ZITNULL)
	  rtmp->zitlist = zit;
	else
	  q->next = zit;

	return;
}

/*
 * add/delete all the zone multicast addresses for the interface zone
 * list (called after interface has been configured and filters opened).
 *
 */

void
zip_multicast(ifn, fn)
int ifn, fn;
{
	struct zit *zit;
	struct rtmp *rtmp, *rtmp_find();
	u_char *zaddr, *zone_mcast();

	if (iflist[ifn].net_lo == 0)
	  return;

	rtmp = rtmp_find(iflist[ifn].net_lo, iflist[ifn].net_hi);

	if (rtmp == RTMPNULL)
	  return;
	
	if (rtmp->dist != 0)
	  return;

	zit = rtmp->zitlist;

	while (zit != ZITNULL) {
	  zaddr = zone_mcast(zit->zone->name+1, (short)zit->zone->name[0]);
	  if (zaddr != NULL) {
	    if (fn == ADDMULTI) {
	      if (iflist[ifn].etfd >= 0)
	        eth_addmulti(iflist[ifn].etfd, iflist[ifn].if_name, zaddr);
	    } else { /* DELMULTI */
	      if (iflist[ifn].etfd >= 0)
	        eth_delmulti(iflist[ifn].etfd, iflist[ifn].if_name, zaddr);
	    }
	    if (debug & ZIPDEBUG) {
	      fprintf(stderr, " ZIP: %-5s %s Zone Multicast",
	        iflist[rtmp->intfc].if_name,
	        (fn == ADDMULTI) ? "Adding" : "Deleting");
	      fprintf(stderr, " %02x:%02x:%02x:%02x:%02x:%02x  \"%s\"\n",
	        zaddr[0], zaddr[1], zaddr[2], zaddr[3], zaddr[4], zaddr[5],
	        zit->zone->name+1);
	    }
	  }
	  zit = zit->next;
	}

	return;
}

/*
 * The GetNetInfo request data consists of 4 bytes of zeroes
 * followed by the zone name for which the information is being
 * requested.
 *
 */

void
zip_inforeq(ifn, pd)
int ifn;
struct pd *pd;
{
	short len, zlen, zvalid, zcount;
	u_char *pkt, *znam, *mcast, *zaddr, *zone_mcast();
	struct rtmp *rtmp, *rtmp_find(), *rtmp_route();
	struct zit *zit;
	struct pd opd;
	void ddp_route();

	pkt = pd->data;
	len = pd->dataLen;

	if (debug & ZIPDEBUG)
	  fprintf(stderr, " ZIP: %-5s Receive GetNetInfo\n",
	    iflist[ifn].if_name);

	if (iflist[ifn].phase == PHASE1)
	  drop(zipInErrors);

	if (len == 0)
	  drop(ddpTooShortErrors);
	  
	if (len > MAXDDPLEN)
	  drop(ddpTooLongErrors);

	if (pd->dstSkt != ZIP)
	  drop(uarNoSocketHandler);

	if (pkt[1]|pkt[2]|pkt[3]|pkt[4]|pkt[5])
	  drop(zipInErrors);

	zlen = pkt[6];
	znam = pkt+7;

	rtmp = rtmp_find(iflist[ifn].net_lo, iflist[ifn].net_hi);

	if (rtmp == RTMPNULL)
	  drop(uarRTMPCacheFail);

	zvalid = 0;
	zcount = 0;
	zit = rtmp->zitlist;
	while (zit != ZITNULL) {
	  if (zlen > 0 && zstrncmpci(zit->zone->name+1, znam, zlen) == 0)
	    zvalid = 1;
	  zcount++;
	  zit = zit->next;
	}

	len = zlen+7+7;

	pkt[0] = ZIPINFOREP;
	pkt[1] = 0x00;

	if (!zvalid)
	  pkt[1] |= 0x80;

	if (zcount == 1)
	  pkt[1] |= 0x20;

	pkt[2] = iflist[ifn].net_lo >> 8;
	pkt[3] = iflist[ifn].net_lo & 0xff;
	pkt[4] = iflist[ifn].net_hi >> 8;
	pkt[5] = iflist[ifn].net_hi & 0xff;

	mcast = znam+zlen;

	if (!zvalid) {
	  if (iflist[ifn].defzone == ZONENULL)
	    drop(uarNoDefaultZone);
	  znam = iflist[ifn].defzone->name+1;
	  zlen = iflist[ifn].defzone->name[0];
	  bcopy((char *)znam, (char *)(mcast+8), zlen);
	  *(mcast+7) = zlen;
	  len += (zlen+1);
	}

	if ((zaddr = zone_mcast(znam, zlen)) == NULL)
	  return;

	*mcast++ = 0x06;
	bcopy((char *)zaddr, (char *)mcast, 6);

	opd.lapType = LDDP;
	opd.ddpType = ZIP;
	opd.dstNet  = pd->srcNet;
	opd.srcNet  = iflist[ifn].net;
	opd.dstNode = pd->srcNode;
	opd.srcNode = iflist[ifn].node;
	opd.dstSkt  = pd->srcSkt;
	opd.srcSkt  = ZIP;
	opd.dataLen = len;
	opd.data = pkt;
	opd.addr = NULL;
	opd.hop  = 0;

	/*
	 * handle broken GetNetInfo requests
	 *
	 */
	if (opd.dstNet < P2STARTUP
	 && (opd.dstNet < iflist[ifn].net_lo
	  || opd.dstNet > iflist[ifn].net_hi)) {
	  opd.dstNet = iflist[ifn].net;
	  opd.dstNode = BCAST;
	}

	if (debug & ZIPDEBUG)
	  fprintf(stderr, " ZIP: %-5s Sending GetNetInfo Reply\n",
	    iflist[ifn].if_name);

	incr(ddpOutRequests);

	ddp_route(ifn, &opd, rtmp_route(opd.dstNet));

	return;
}

/*
 * return pointer to zone multicast address
 *
 */

u_char *
zone_mcast(znam, zlen)
u_char *znam;
short zlen;
{
	u_char zone[NBPSTR];
	u_short i, chkSum();
	u_char mactoupper();
	static u_char zmcaddr[6];

	if (zlen > sizeof(zone))
	  return(NULL);

	for (i = 0; i < zlen; i++)
	  zone[i] = mactoupper(znam[i]);

	zmcaddr[0] = 0x09;
	zmcaddr[1] = 0x00;
	zmcaddr[2] = 0x07;
	zmcaddr[3] = 0x00;
	zmcaddr[4] = 0x00;

	zmcaddr[5] = (u_char)(chkSum(zone, zlen) % 253);

	return(zmcaddr);
}

/*
 * send a GetNetInfo request to determine the default zone
 *
 * The GetNetInfo request data consists of 4 bytes of zeroes
 * (reserved) followed by the zone name for which information
 * is being requested. The length byte should be set to zero
 * if no zone name is being provided in the packet.
 *
 */

void
zone_getnetinfo(ifn)
int ifn;
{
	short len;
	struct pd opd;
	u_char *pkt, buf[80];
	struct rtmp *rtmp_route();
	void ddp_route();

	pkt = buf+HDRSIZ+LDDPHDRSIZ;

	len = strlen(NETINFOZONE);

	pkt[0] = ZIPINFOREQ;
	pkt[1] = 0x00;			/* must be zero	*/
	pkt[2] = 0x00;			/* must be zero	*/
	pkt[3] = 0x00;			/* must be zero	*/
	pkt[4] = 0x00;			/* must be zero	*/
	pkt[5] = 0x00;			/* must be zero	*/
	pkt[6] = len;			/* zone length	*/

	bcopy(NETINFOZONE, (char *)(pkt+7), len);

	len += 0x07;

	opd.lapType = LDDP;
	opd.ddpType = ZIP;
	opd.dstNet  = iflist[ifn].net;
	opd.srcNet  = iflist[ifn].net;
	opd.dstNode = BCAST;
	opd.srcNode = iflist[ifn].node;
	opd.dstSkt  = ZIP;
	opd.srcSkt  = ZIP;
	opd.dataLen = len;
	opd.data = pkt;
	opd.addr = NULL;
	opd.hop  = 0;

	if (debug & ZIPDEBUG)
	  fprintf(stderr, " ZIP: %-5s Sending GetNetInfo\n",
	    iflist[ifn].if_name);

	incr(ddpOutRequests);

	ddp_route(ifn, &opd, rtmp_route(opd.dstNet));

	return;
}

/*
 * The GetNetInfo Reply data begins with the starting and ending
 * network numbers of the network on which the request was made.
 * This range is always followed by a copy of the zone name from
 * the request. The zone multicast address to be used by the node
 * follows the copy of the requested zone name. It is preceded by
 * its length in bytes.
 *
 */

void
zip_inforep(ifn, pd)
int ifn;
struct pd *pd;
{
	u_char flags;
	u_short net_lo;
	u_short net_hi;
	char message[80];
	short len, zlen, mlen;
	u_char *pkt, *znam, *mcast;
	struct zone *zone, *zone_find(), *zone_new();
	void log(), cleanup();
	void zone_insert();
	void rtmp_intfc();

	pkt = pd->data;
	len = pd->dataLen;

	if (debug & ZIPDEBUG)
	  fprintf(stderr, " ZIP: %-5s Receive GetNetInfo Reply ...\n",
	    iflist[ifn].if_name);

	if (iflist[ifn].phase == PHASE1)
	  drop(zipInErrors);

	if (len == 0)
	  drop(ddpTooShortErrors);
	  
	if (len > MAXDDPLEN)
	  drop(ddpTooLongErrors);

	if (pd->dstSkt != ZIP)
	  drop(uarNoSocketHandler);

	flags = pkt[1];

	net_lo = (pkt[2] << 8) | pkt[3];
	net_hi = (pkt[4] << 8) | pkt[5];

	zlen = pkt[6];
	znam = pkt+7;

	if (zlen != strlen(NETINFOZONE)
	 || bcmp((char *)znam, NETINFOZONE, zlen) != 0)
	/* reply not ours */
	  drop(zipInErrors);

	rtmp_intfc(ifn, net_lo, net_hi, pd);

	mcast = znam+zlen;

	if ((mlen = *mcast++) != 0x06)
	  drop(zipInErrors);

	if (flags & 0x80) {
	/* lkup zone invalid */
	  zlen = mcast[mlen];
	  znam = mcast+mlen+1;
	  if (zlen == 0)
	    drop(zipInErrors);
	}

	if ((zone = zone_find(znam, zlen)) == ZONENULL) {
	  if ((zone = zone_new()) != ZONENULL) {
	    bcopy((char *)znam, (char *)(zone->name+1), zlen);
	    zone->name[zlen+1] = '\0';
	    zone->name[0] = zlen;
	    zone->nets = NETNULL;
	    zone->next = ZONENULL;
	    zone_insert(zone);
	  }
	}
	if (zone == ZONENULL)
	  drop(uarMemoryError);

	if (iflist[ifn].defzone != NULL) {
	  if (iflist[ifn].defzone != zone) {
	    log("PANIC: default zone configuration mismatch, giving up.");
	    sprintf(message, "%s: Config %s, Real %s", iflist[ifn].if_name,
	      iflist[ifn].defzone->name+1, zone->name+1);
	    log(message);
	    cleanup();
	  }
	}

	iflist[ifn].defzone = zone;

	if (debug & ZIPDEBUG) {
	  fprintf(stderr, "\tLoRange: %3d.%-3d\n",
	    net_lo >> 8, net_lo & 0xff);
	  fprintf(stderr, "\tHiRange: %3d.%-3d\n",
	    net_hi >> 8, net_hi & 0xff);
	  fprintf(stderr, "\tDefZone: \"%s\"\n", zone->name+1);
	  fprintf(stderr, "\tMCZaddr: %02x:%02x:%02x:%02x:%02x:%02x\n",
	    mcast[0],mcast[1],mcast[2],mcast[3],mcast[4],mcast[5]);
	}

	return;
}

/*
 *
 *
 */

void
zip_notify(ifn, pd)
int ifn;
struct pd *pd;
{
	short len;
	u_char *pkt;

	pkt = pd->data;
	len = pd->dataLen;

	if (debug & ZIPDEBUG)
	  fprintf(stderr, " ZIP: %-5s Receive Notify\n",
	    iflist[ifn].if_name);

	if (iflist[ifn].phase == PHASE1)
	  drop(zipInErrors);

	if (len == 0)
	  drop(ddpTooShortErrors);
	  
	if (len > MAXDDPLEN)
	 drop(ddpTooLongErrors);

	if (pd->dstSkt != ZIP)
	  drop(uarNoSocketHandler);

	/* NYI */

	return;
}

/*
 * incoming ZIP ATP packet
 *
 */

void
zip_atp(ifn, pd)
int ifn;
struct pd *pd;
{
	short len;
	u_short tid, idx;
	char *zip_atp_type();
	u_char *pkt, function;
	void zip_myzone(), zip_zonelist(), zip_localzones();

	pkt = pd->data;
	len = pd->dataLen;

	function = pkt[4];

	if (debug & ZIPDEBUG)
	  fprintf(stderr, " ZIP: %-5s Receive ATP %s\n",
	    iflist[ifn].if_name, zip_atp_type(function));

	if (len == 0)
	  drop(ddpTooShortErrors);
	  
	if (len > MAXDDPLEN)
	  drop(ddpTooLongErrors);

	if (pd->dstSkt != ZIP)
	  drop(uarNoSocketHandler);

	if (pkt[0] != 0x40 || pkt[1] != 0x01)
	  drop(zipInErrors);

	tid = (pkt[2] << 8) | pkt[3];
	idx = (pkt[6] << 8) | pkt[7];

	switch (function) {
	  case ZIPGETMYZONE:
	    zip_myzone(ifn, pd, tid, idx);
	    break;
	  case ZIPGETZONELIST:
	    zip_zonelist(ifn, pd, tid, idx);
	    break;
	  case ZIPGETLOCALZON:
	    zip_localzones(ifn, pd, tid, idx);
	    break;
	  default:
	    drop(zipInErrors);
	    break;
	}

	return;
}

/*
 * return (default) zone for interface
 *
 */

void
zip_myzone(ifn, pd, tid, idx)
int ifn;
struct pd *pd;
u_short tid, idx;
{
	short len, zlen;
	u_char *pkt, *znam;
	struct rtmp *rtmp_route();
	struct zone *zone, *zone_default();
	struct pd opd;
	void ddp_route();

	if (idx != 0)
	  drop(uarBadPacketHdr);

	pkt = zbuf+HDRSIZ+LDDPHDRSIZ;
	len = 8;

	pkt[0] = TRESP | EOM;
	pkt[1] = 0x00;
	pkt[2] = tid >> 8;
	pkt[3] = tid & 0xff;

	pkt[4] = 0x00;			/* must be zero	*/
	pkt[5] = 0x00;			/* must be zero	*/
	pkt[6] = 0x00;			/* num of zones	*/
	pkt[7] = 0x01;			/* num of zones	*/

	if ((zone = zone_default(ifn)) == ZONENULL)
	  drop(uarNoDefaultZone);

	zlen = zone->name[0];
	znam = zone->name+1;

	pkt[8] = zlen;
	bcopy((char *)znam, (char *)(pkt+9), zlen);

	len += (zlen+1);

	opd.lapType = LDDP;
	opd.ddpType = ATP;
	opd.dstNet  = pd->srcNet;
	opd.srcNet  = iflist[ifn].net;
	opd.dstNode = pd->srcNode;
	opd.srcNode = iflist[ifn].node;
	opd.dstSkt  = pd->srcSkt;
	opd.srcSkt  = ZIP;
	opd.dataLen = len;
	opd.data = pkt;
	opd.addr = NULL;
	opd.hop  = 0;

	if (debug & ZIPDEBUG)
	  fprintf(stderr, " ZIP: %-5s Sending GetMyZoneReply\n",
	    iflist[ifn].if_name);

	incr(ddpOutRequests);

	ddp_route(ifn, &opd, rtmp_route(opd.dstNet));

	return;
}

/*
 * return complete zone list
 *
 */

void
zip_zonelist(ifn, pd, tid, idx)
int ifn;
struct pd *pd;
u_short tid, idx;
{
	short zlen;
	u_char *data;
	struct zone *zone;
	void zip_zone_send();

	data = zbuf+HDRSIZ+LDDPHDRSIZ+ZIPATPHDRSIZ;
	zoffset = ZIPATPHDRSIZ;
	zcount = 0;

	zone = zoneTab;

	while (zone != ZONENULL) {
	  if (--idx == 0)
	    break;
	  zone = zone->next;
	}

	while (zone != ZONENULL) {
	  zlen = zone->name[0]+1;
	  bcopy((char *)zone->name, (char *)data, zlen);
	  zoffset += zlen;
	  data += zlen;
	  zcount++;
	  if (zoffset > MAXZIPREPLYLEN) {
	    zip_zone_send(ifn, pd, tid, (zone->next == ZONENULL));
	    return;
	  }
	  zone = zone->next;
	}

	zip_zone_send(ifn, pd, tid, 1);

	return;
}

/*
 * return zone list for interface
 *
 */

void
zip_localzones(ifn, pd, tid, idx)
int ifn;
struct pd *pd;
u_short tid, idx;
{
	short zlen;
	u_char *data;
	struct zit *zit;
	struct zone *zone;
	struct rtmp *rtmp, *rtmp_find();
	void zip_zone_send();

	data = zbuf+HDRSIZ+LDDPHDRSIZ+ZIPATPHDRSIZ;
	zoffset = ZIPATPHDRSIZ;
	zcount = 0;

	rtmp = rtmp_find(iflist[ifn].net_lo, iflist[ifn].net_hi);

	if (rtmp == RTMPNULL)
	  drop(uarRTMPCacheFail);

	if ((zit = rtmp->zitlist) == ZITNULL)
	  return; /* drop it */

	while (zit != ZITNULL) {
	  if (--idx == 0)
	    break;
	  zit = zit->next;
	}

	while (zit != ZITNULL) {
	  if ((zone = zit->zone) != ZONENULL) {
	    zlen = zone->name[0]+1;
	    bcopy((char *)zone->name, (char *)data, zlen);
	    zoffset += zlen;
	    data += zlen;
	    zcount++;
	    if (zoffset > MAXZIPREPLYLEN) {
	      zip_zone_send(ifn, pd, tid, (zit->next == ZITNULL));
	      return;
	    }
	  }
	  zit = zit->next;
	}

	zip_zone_send(ifn, pd, tid, 1);

	return;
}

/*
 * send the ATP zone list or local zone list reply
 *
 */

void
zip_zone_send(ifn, pd, tid, last)
int ifn;
struct pd *pd;
u_short tid;
int last;
{
	short len;
	u_char *pkt;
	struct pd opd;
	struct rtmp *rtmp_route();
	void ddp_route();

	pkt = zbuf+HDRSIZ+LDDPHDRSIZ;

	/* ZIP ATP Reply Data */

	pkt[0] = TRESP | EOM;
	pkt[1] = 0x00;				/* zero seq. #		*/
	pkt[2] = tid >> 8;
	pkt[3] = tid & 0xff;

	pkt[4] = (zcount == 0) ? 1 : last;
	pkt[5] = 0x00;				/* must be zero		*/
	pkt[6] = zcount >> 8;
	pkt[7] = zcount & 0xff;

	len = zoffset;				/* ZIP Reply pkt length	*/

	opd.lapType = LDDP;
	opd.ddpType = ATP;
	opd.dstNet  = pd->srcNet;
	opd.srcNet  = iflist[ifn].net;
	opd.dstNode = pd->srcNode;
	opd.srcNode = iflist[ifn].node;
	opd.dstSkt  = pd->srcSkt;
	opd.srcSkt  = ZIP;
	opd.dataLen = len;
	opd.data = pkt;
	opd.addr = NULL;
	opd.hop  = 0;

	if (debug & ZIPDEBUG)
	  fprintf(stderr, " ZIP: %-5s Sending Zones Reply (%d zones)\n",
	    iflist[ifn].if_name, zcount);

	incr(ddpOutRequests);

	ddp_route(ifn, &opd, rtmp_route(opd.dstNet));

	zcount = 0;
	zoffset = ZIPATPHDRSIZ;

	return;
}

/*
 * return a string describing the ZIP ATP packet type
 *
 */

char *
zip_atp_type(fn)
u_char fn;
{
	switch (fn) {
	case ZIPGETMYZONE:
	  return("GetMyZone");
	  break;
	case ZIPGETZONELIST:
	  return("GetZoneList");
	  break;
	case ZIPGETLOCALZON:
	  return("GetLocalZones");
	  break;
	 default:
	  break;
	}

	return("Unknown");
}

/*
 * find a zone in the (sorted) zone table
 *
 */

struct zone *
zone_find(znam, zlen)
u_char *znam;
short zlen;
{
	int scmp, ifn;
	struct zone *zone;
	extern int numintfc;

	incr(uarZONEFinds);

	/*
	 * check if most-recently-used
	 *
	 */
	if ((zone = zoneMRU) != ZONENULL) {
	  incr(uarZONESearches);
	  if (zstrncmpci(zone->name+1, znam, zlen) == 0)
	    return(zone);
	}

	/*
	 * check interface default zones
	 *
	 */
	for (ifn = 0; ifn < numintfc; ifn++) {
	  if ((zone = iflist[ifn].defzone) != ZONENULL) {
	    incr(uarZONESearches);
	    if (zstrncmpci(zone->name+1, znam, zlen) == 0)
	      return((zoneMRU = zone));
	  }
	}

	zone = zoneTab;

	while (zone != ZONENULL) {
	  incr(uarZONESearches);
	  if ((scmp = zstrncmpci(zone->name+1, znam, zlen)) == 0)
	    return((zoneMRU = zone));
	  if (scmp > 0)
	    break;
	  zone = zone->next;
	}

	return(ZONENULL);
}

/*
 * insert a zone name in the zone table
 *
 */

void
zone_insert(zone)
struct zone *zone;
{
	struct zone *p, *q;
	int zstrncmpci();

	p = zoneTab;
	q = ZONENULL;

	while (p != ZONENULL) {
	  if (zstrncmpci(p->name+1, zone->name+1, (short)*zone->name) > 0)
	    break;
	  q = p;
	  p = p->next;
	}

	zone->next = p;

	if (q == ZONENULL)
	  zoneTab = zone;
	else
	  q->next = zone;

	return;
}

/*
 * find or make a zone entry
 *
 */

struct zone *
zone_make(znam)
u_char *znam;
{
	short zlen;
	void zone_insert(), exit();
	struct zone *zone, *zone_find(), *zone_new();

	if ((zlen = strlen((char *)znam)) > MAXNBP) {
	  fprintf(stderr, " UAR: Zonename \"%s\" too long (%d)\n", znam, zlen);
	  exit(1);
	}

	if ((zone = zone_find(znam, zlen)) == ZONENULL) {
	  if ((zone = zone_new()) != ZONENULL) {
	    bcopy((char *)znam, (char *)(zone->name+1), zlen);
	    zone->name[zlen+1] = '\0';
	    zone->name[0] = zlen;
	    zone->nets = NETNULL;
	    zone->next = ZONENULL;
	    zone_insert(zone);
	  }
	}

	return(zone);
}

/*
 * return the default zone for interface 'ifn'
 *
 */

struct zone *
zone_default(ifn)
int ifn;
{
	struct zit *zit;
	struct rtmp *rtmp, *rtmp_find();

	if (iflist[ifn].phase == PHASE2)
	  return(iflist[ifn].defzone);

	rtmp = rtmp_find(iflist[ifn].net_lo, iflist[ifn].net_hi);

	if (rtmp == RTMPNULL)
	  return(ZONENULL);

	if ((zit = rtmp->zitlist) == ZITNULL)
	  return(ZONENULL);

	if (zit->zone == ZONENULL)
	  return(ZONENULL);

	if (zit->zone->next != ZONENULL)
	  return(ZONENULL);

	return(zit->zone);
}

/*
 * return the services zone name (C string) for interface 'ifn'
 *
 */

char *
zone_name(ifn)
int ifn;
{
	short zlen;
	u_char *znam;
	static u_char name[NBPSTR];
	struct zone *zone, *zone_default();

	name[0] = '\0';

	if ((zone = iflist[ifn].ourzone) == ZONENULL)
	  if ((zone = zone_default(ifn)) == ZONENULL)
	    return((char *)name);

	znam = zone->name+1;
	zlen = zone->name[0];

	if (zlen >= sizeof(name))
	  return((char *)name);

	bcopy((char *)znam, (char *)name, zlen);
	name[zlen] = '\0';

	return((char *)name);
}

/*
 * dump the zone table
 *
 */

void
zone_dump()
{
	FILE *r, *fopen();
	struct zone *zone;
	struct net *net;
	short znum = 0;
	short lnum = 0;

	if ((r = fopen("/usr/tmp/uar.zone", "w+")) == NULL) {
	  incr(uarBadDumpFile);
	  return;
	}

	fprintf(r, "UAR ZONE Table:\n\n");

	zone = zoneTab;
	while (zone != ZONENULL) {
	  fprintf(r, "%24s", zone->name+1);
	  net = zone->nets;
	  while (net != NETNULL) {
	    if (net->net_lo == net->net_hi)
	      fprintf(r, " %3d.%-3d  ",
	        net->net_lo >> 8, net->net_lo & 0xff);
	    else
	      fprintf(r, " %3d.%-3d - %3d.%-3d",
	        net->net_lo >> 8, net->net_lo & 0xff,
	        net->net_hi >> 8, net->net_hi & 0xff);
	    net = net->next;
	    if (net != NETNULL && (++lnum%0x04) == 0)
	      fprintf(r, "\n%24s", "");
	  }
	  znum++;
	  lnum = 0;
	  fprintf(r, "\n");
	  zone = zone->next;
	}
	fprintf(r, "\n\n%d Zones\n", znum);

	(void)fclose(r);

	if (debug & ZIPDEBUG)
	  fprintf(stderr, " ZIP: dumped ZIP table, %d zones\n", znum);

	return;
}

/*
 * make a new zone table entry
 *
 */

struct zone *
zone_new()
{
	struct zone *zone;

	zone = (struct zone *)malloc(sizeof(struct zone));

	if (zone == ZONENULL)
	  incr(uarMemoryError);

	return(zone);
}

/*
 * Case insensitive version of strncmp() for zone names
 *
 * (first arg is null terminated, second arg is of
 * length n and is NOT necessarily null terminated)
 *
 */

int
zstrncmpci(s, t, n)
register u_char *s, *t;
register short n;
{
	register u_char c, d;
	u_char mactolower();

	for ( ; n > 0 ; n--) {
	  c = mactolower(*s++);
	  d = mactolower(*t++);
	  if (c != d)
	    return(c-d);
	}

	return((*s == '\0') ? 0 : *s);
}
